/***********************************************************************\
*
* $Revision: 642 $
* $Date: 2018-01-18 16:38:10 +0100 (Thu, 18 Jan 2018) $
* $Author: torsten $
* Contents: Redmine backend implementation
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.lang.ref.SoftReference;
import java.io.BufferedReader;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.StringWriter;

import java.net.HttpURLConnection;
import java.net.ProtocolException;
import java.net.UnknownHostException;
import java.net.URL;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.List;

// SSL
import javax.net.ssl.HttpsURLConnection;

// graphics
import org.eclipse.swt.graphics.Image;

// XML
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

//import org.w3c.dom.NamedNodeMap;

// SQLite database
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

/****************************** Classes ********************************/

/** Redmine backend driver functions
 */
public class BackendDriverRedmine extends BackendDriver
{
  /** element handler
   */
  private abstract class ElementHandler<T>
  {
    // --------------------------- constants --------------------------------

    // --------------------------- variables --------------------------------
    private int     index;
    private int     totalCount;
    private Object  result;
    private boolean done;

    // ------------------------ native functions ----------------------------

    // ---------------------------- methods ---------------------------------

    /** create element handler
     * @param defaultResult default result value
     */
    ElementHandler(Object defaultResult)
    {
      this.index      = 0;
      this.totalCount = 0;
      this.result     = defaultResult;
      this.done       = false;
    }

    /** create element handler
     */
    ElementHandler()
    {
      this(null);
    }

    /** done iteration
     * @param result result value
     */
    public void done(Object result)
    {
      this.result = result;
      this.done   = true;
    }

    /** done iteration
     */
    public void done()
    {
      done(null);
    }

    /** set element index
     */
    protected void setIndex(int index)
    {
      this.index = index;
    }

    /** increment element index
     */
    protected void incrementIndex()
    {
      this.index++;
    }

    /** set total count
     */
    protected void setTotalCount(int totalCount)
    {
      this.totalCount = totalCount;
    }

    /** process element
     * @param element element
     * @param index index [0..n-1]
     * @param totalCount total count
     */
    protected T process(Element element, int index, int totalCount)
      throws BackendException
    {
      return process(element);
    }

    /** process element
     * @param element element
     */
    protected T process(Element element)
      throws BackendException
    {
      T entity = parse(element);
      if (entity != null)
      {
        store(entity,index,totalCount);
      }

      return entity;
    }

    /** parse element
     * @param element data element
     * @return entity
     */
    protected T parse(Element element)
    {
//Dprintf.dprintf("-------");
//new Throwable().printStackTrace();
//throw new Error();
      return null;
    }

    /** process root element handler
     * @param element header element
     */
    protected void root(Element element)
    {
    }

    /** store entity handler
     * @param entity entity to store
     * @param index index [0..n-1]
     * @param totalCount total count
     */
    protected void store(T entity, int index, int totalCount)
      throws BackendException
    {
//Dprintf.dprintf("-------");
//new Throwable().printStackTrace();
//throw new Error();
      store(entity);
    }

    /** store entity handler
     * @param index index [0..n-1]
     * @param totalCount total count
     * @param entity entity to store
     */
    protected void store(T entity)
      throws BackendException
    {
//Dprintf.dprintf("-------");
//new Throwable().printStackTrace();
//throw new Error();
    }

    /** check if done
     * @return true iff done
     */
    protected boolean isDone()
    {
      return done;
    }

    /** get result
     * @return result or null
     */
    protected Object getResult()
    {
      return this.result;
    }

    /** handle error
     * @param throwable throwable
     */
    protected void error(Exception exception)
      throws BackendException
    {
      Timemine.printStackTrace(exception);
      throw new BackendException(exception);
    }
  }

  /** create element handler
   */
  private abstract class CreateHandler<T extends Entity>
  {
    // --------------------------- constants --------------------------------

    // --------------------------- variables --------------------------------

    // ------------------------ native functions ----------------------------

    // ---------------------------- methods ---------------------------------

    /** create XML document handler
     */
    CreateHandler()
    {
    }

    /** create XML data element
     * @param document XML document
     * @param rootElement root element
     * @param entity entity
     */
    abstract public void data(Document document, Element rootElement, T entity);

    // ----------------------------------------------------------------------
  }

  // --------------------------- constants --------------------------------
  private final static int               ENTRY_LIMIT         = 100; // max. 100 (see Redmine REST API)
  private final static SimpleDateFormat  DATE_FORMAT         = new SimpleDateFormat("yyyy-MM-dd");
  private final static SimpleDateFormat  DATETIME_FORMAT     = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
  private final static SimpleDateFormat  ISO_DATETIME_FORMAT = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");

  private final SoftReference<TimeEntry> TIME_ENTRY_NULL     = new SoftReference<TimeEntry>(null);

  private final String FILTER_EQ                             = "%3D";
  private final String FILTER_GE                             = "%3E%3D";
  private final String FILTER_LE                             = "%3C%3D";

  // --------------------------- variables --------------------------------
  private static SimpleDateFormat dateFormats[];

  private static User             currentUser   = null;
  private static Member           currentMember = null;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  static
  {
    final String[] DATE_FORMATS = new String[]
    {
      "yyyy-MM-dd'T'HH:mm:ss'Z'",    // 2011-01-25T15:19:06Z
      "yyyy-MM-dd HH:mm:ss Z",       // 2011-01-25 15:19:06 +0900
      "yyyy-MM-dd HH:mm Z",          // 2011-01-25 15:19 +0900
      "yyyy/MM/dd HH:mm:ss",         // 2010/05/13 07:09:36
      "EEE MMM dd HH:mm:ss yyyy Z",  // Wed Dec 17 15:41:19 2008 +0100
      "yyyy-MM-dd",                  // 2011-01-25
      "dd-MMM-yy",                   // 25-Jan-11
    };

    dateFormats = new SimpleDateFormat[DATE_FORMATS.length];
    for (int i = 0; i < DATE_FORMATS.length; i++)
    {
      dateFormats[i] = new SimpleDateFormat(DATE_FORMATS[i]);
    }
  }

  /** initialize Redmine backend
   * @param serverName server name
   * @param serverPort server port
   * @param serverUseSSL true iff server use SSL
   * @param Name login name
   * @param loginPasssword login password
   * @param  certificateType certificate type
   * @param certificateFileName certificate file (.cert, can be null)
   * @param keyFileName key file (.pem, can be null)
   * @param p12FileName p12 file name (.p12, can be null)
   * @param keyPassword key password
   * @param logo logo image
   */
  public BackendDriverRedmine(String                    serverName,
                              int                       serverPort,
                              boolean                   serverUseSSL,
                              final String              loginName,
                              String                    loginPassword,
                              Settings.CertificateTypes certificateType,
                              String                    certificateFileName,
                              String                    keyFileName,
                              String                    p12FileName,
                              String                    keyPassword,
                              Image                     logo
                             )
    throws BackendException
  {
    super(serverName,
          serverPort,
          serverUseSSL,
          loginName,
          loginPassword,
          certificateType,
          certificateFileName,
          keyFileName,
          p12FileName,
          keyPassword,
          logo
         );
  }

  /** get backend type
   * @return REDMINE
   */
  @Override
  public Backend.Types getType()
  {
    return Backend.Types.REDMINE;
  }

  /** check iff activity in time entries is supported
   * @return true iff activity in time entries is supported
   */
  public boolean isActivitySupported()
  {
    return true;
  }

  /** check iff edit progress on issues is supported
   * @return true iff setting progress on issues is supported
   */
  public boolean isEditProgressSupported()
  {
    return true;
  }

  //--------------------------------------------------------------------

  /** load issue from Redmine server
   * @param projectId project Redmine id
   * @param issueId issue Redmine id
   * @param id Redmine id
   * @param clazz Issue class
   */
  @Override
  public Issue loadIssue(final Id issueId)
    throws BackendException
  {
    Issue issue = null;

    assert(issueId.hasBackend());

    try
    {
      issue = getData("/issues/"+issueId.backend,"issue",new ElementHandler<Issue>()
      {
        @Override
        public Issue parse(Element element)
        {
          return new Issue(Backend.Types.REDMINE,
                           issueId,
                           getIdAttribute(element,"parent","id"),
                           getIdAttribute(element,"project","id"),
                           getIdAttribute(element,"tracker","id"),
                           getIdAttribute(element,"status","id"),
                           getIdAttribute(element,"priority","id"),
                           getIdAttribute(element,"author","id"),
                           getValue(element,"subject",""),
                           getValue(element,"description",""),
                           getDateValue(element,"start_date"),
                           getDateValue(element,"due_date"),
                           getIdAttribute(element,"assigned_to","id",Id.NONE),
                           getIntValue(element,"done_ratio"),
                           getIdAttribute(element,"fixed_version","id",Id.ANY),
                           (long)(getDoubleValue(element,"estimated_hours",Issue.ESTIMATED_NONE)*60.0),
                           getDateValue(element,"created_on"),
                           getDateValue(element,"updated_on"),
                           getDateValue(element,"closed_on")
                          );
        }
      });
    }
    catch (InterruptedException exception)
    {
      throw new BackendException(exception);
    }

    return issue;
  }

  /** store issue on Redmine server
   * @param issue issue to store
   */
  @Override
  public void storeIssue(final Issue issue)
    throws BackendException
  {
    assert(issue.backendType == Backend.Types.REDMINE);

    postData("/issues","issue",issue,new CreateHandler<Issue>()
    {
      public void data(Document document, Element rootElement, Issue issue)
      {
        Element element;

        if (issue.parentId.backend != Id.NONE)
        {
          element = document.createElement("parent_issue_id");
          element.appendChild(document.createTextNode(Integer.toString(issue.parentId.backend)));
          rootElement.appendChild(element);
        }

        element = document.createElement("project_id");
        element.appendChild(document.createTextNode(Integer.toString(issue.projectId.backend)));
        rootElement.appendChild(element);

        if (issue.trackerId.backend != Id.NONE)
        {
          element = document.createElement("tracker_id");
          element.appendChild(document.createTextNode(Integer.toString(issue.trackerId.backend)));
          rootElement.appendChild(element);
        }

        if (issue.statusTransitionId.backend != Id.NONE)
        {
          element = document.createElement("status_id");
          element.appendChild(document.createTextNode(Integer.toString(issue.statusTransitionId.backend)));
          rootElement.appendChild(element);
        }

        if (issue.priorityId.backend != Id.NONE)
        {
          element = document.createElement("priority_id");
          element.appendChild(document.createTextNode(Integer.toString(issue.priorityId.backend)));
          rootElement.appendChild(element);
        }

        element = document.createElement("subject");
        element.appendChild(document.createTextNode(issue.summary));
        rootElement.appendChild(element);

        element = document.createElement("description");
        element.appendChild(document.createTextNode(issue.description));
        rootElement.appendChild(element);

        if (!issue.targetVersionId.isNone())
        {
          element = document.createElement("fixed_version_id");
          element.appendChild(document.createTextNode(Integer.toString(issue.targetVersionId.backend)));
          rootElement.appendChild(element);
        }

        if (!issue.assignedToMemberId.isAny())
        {
          element = document.createElement("assigned_to_id");
          element.appendChild(document.createTextNode(Integer.toString(issue.assignedToMemberId.backend)));
          rootElement.appendChild(element);
        }

        element = document.createElement("estimated_hours");
        element.appendChild(document.createTextNode(Double.toString((double)issue.estimatedDuration/60.0)));
        rootElement.appendChild(element);
      }
    });

//TODO
Dprintf.dprintf("");
/*
    // store id
    try
    {
      updateLocalDatabaseId(issue);
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
*/
  }

  /** update issue on Redmine server
   * @param issue issue to update
   * @param originIssue origin issue
   */
  @Override
  public void updateIssue(final Issue issue, final Issue originIssue)
    throws BackendException
  {
/*
    category_id
    parent_issue_id - ID of the parent issue
    custom_fields - See Custom fields
    estimated_hours - Number of hours estimated for issue
*/
    assert(issue.backendType == Backend.Types.REDMINE);
    assert(issue.id.hasBackend());

    putData("/issues/"+issue.id.backend,"issue",issue,new CreateHandler<Issue>()
    {
      public void data(Document document, Element rootElement, Issue issue)
      {
        Element element;

        if (issue.parentId.backend != Id.NONE)
        {
          element = document.createElement("parent_issue_id");
          element.appendChild(document.createTextNode(Integer.toString(issue.parentId.backend)));
          rootElement.appendChild(element);
        }

        element = document.createElement("tracker_id");
        element.appendChild(document.createTextNode(Integer.toString(issue.trackerId.backend)));
        rootElement.appendChild(element);

        if (!issue.statusTransitionId.isNone())
        {
          element = document.createElement("status_id");
          element.appendChild(document.createTextNode(Integer.toString(issue.statusTransitionId.backend)));
          rootElement.appendChild(element);
        }

        element = document.createElement("priority_id");
        element.appendChild(document.createTextNode(Integer.toString(issue.priorityId.backend)));
        rootElement.appendChild(element);

        element = document.createElement("author_id");
        element.appendChild(document.createTextNode(Integer.toString(issue.authorId.backend)));
        rootElement.appendChild(element);

        if (!issue.assignedToMemberId.isAny())
        {
          element = document.createElement("assigned_to_id");
          element.appendChild(document.createTextNode(Integer.toString(issue.assignedToMemberId.backend)));
          rootElement.appendChild(element);
        }

        element = document.createElement("subject");
        element.appendChild(document.createTextNode(issue.summary));
        rootElement.appendChild(element);

        element = document.createElement("description");
        element.appendChild(document.createTextNode(issue.description));
        rootElement.appendChild(element);

        element = document.createElement("estimated_hours");
        if (issue.estimatedDuration > Issue.ESTIMATED_NONE)
        {
          element.appendChild(document.createTextNode(Double.toString((double)issue.estimatedDuration/60.0)));
        }
        else
        {
          element.appendChild(document.createTextNode(""));
        }
        rootElement.appendChild(element);

        element = document.createElement("done_ratio");
        element.appendChild(document.createTextNode(Integer.toString(issue.doneRatio)));
        rootElement.appendChild(element);
      }
    });

//TODO
Dprintf.dprintf("");
/*
    // store id
    try
    {
      updateLocalDatabaseId(issue);
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
*/
  }

  /** delete issue on Redmine server
   * @param issue issue to delete
   */
  public void deleteIssue(final Issue issue)
    throws BackendException
  {
    assert(issue.backendType == Backend.Types.REDMINE);
    assert(issue.id.hasBackend());

throw new Error("still not implemented");
  }

  //--------------------------------------------------------------------

  /** load time entry from Redmine server
   * @param projectId project Redmine id
   * @param issueId issue Redmine id
   * @param timeEntryId time entry Redmine id
   * @param clazz Issue class
   */
  @Override
  public TimeEntry loadTimeEntry(Id projectId, Id issueId, Id timeEntryId)
    throws BackendException
  {
    TimeEntry timeEntry = null;

    assert(timeEntryId.hasBackend());

    try
    {
      timeEntry = getData("/time_entries/"+timeEntryId.backend,"time_entry",new ElementHandler<TimeEntry>()
      {
        @Override
        public TimeEntry parse(Element element)
        {
          return new TimeEntry(Backend.Types.REDMINE,
                               getIdValue(element,"id"),
                               getIdAttribute(element,"project","id"),
                               getIdAttribute(element,"issue","id"),
                               getIdAttribute(element,"user","id"),
                               getIdAttribute(element,"activity","id"),
                               (long)(getDoubleValue(element,"hours")*60.0),
                               getValue(element,"comments"),
                               new SpentOn(getDateValue(element,"spent_on")),
                               getDateValue(element,"created_on"),
                               getDateValue(element,"updated_on")
                              );
        }
      });
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }

    return timeEntry;
  }

  /** load time entry from Redmine server
   * @param projectId project Redmine id
   * @param issueId issue Redmine id
   * @param timeEntryId time entry Redmine id
   * @param clazz TimeEntry class
   */
// TODO:: remove
/*
  @Override
  public TimeEntry load(Id projectId, Id issueId, Id timeEntryId, Class<TimeEntry> clazz)
    throws BackendException
  {
    return getData("/time_entries/"+timeEntryId.backend,"time_entry",new ElementHandler<TimeEntry>()
    {
      @Override
      public TimeEntry parse(Element element)
      {
        return new TimeEntry(Backend.Types.REDMINE,
                             getIdValue(element,"id"),
                             getIdAttribute(element,"project","id"),
                             getIdAttribute(element,"issue","id"),
                             getIdAttribute(element,"user","id"),
                             getIdAttribute(element,"activity","id"),
                             (long)(getLongValue(element,"hours",0)*60.0),
                             getValue(element,"comments"),
                             getSpentOn(element,"spent_on"),
                             getDateValue(element,"created_on"),
                             getDateValue(element,"updated_on")
                            );
      }
    });
  }
*/

  /** store time entry on Redmine server
   * @param timeEntry time entry to store
   */
  @Override
  public void storeTimeEntry(final TimeEntry timeEntry)
    throws BackendException
  {
    assert(timeEntry.backendType == Backend.Types.REDMINE);

Dprintf.dprintf("---------------------- timeEntry=%s",timeEntry);
    postData("/time_entries","time_entry",timeEntry,new CreateHandler<TimeEntry>()
    {
      public void data(Document document, Element rootElement, TimeEntry timeEntry)
      {
        Element element;

        element = document.createElement("issue_id");
        element.appendChild(document.createTextNode(Integer.toString(timeEntry.issueId.backend)));
        rootElement.appendChild(element);

        element = document.createElement("spent_on");
        element.appendChild(document.createTextNode(timeEntry.spentOn.toString(DATE_FORMAT)));
        rootElement.appendChild(element);

        element = document.createElement("hours");
        element.appendChild(document.createTextNode(Double.toString((double)timeEntry.duration/60.0)));
        rootElement.appendChild(element);

        element = document.createElement("activity_id");
        element.appendChild(document.createTextNode(Integer.toString(timeEntry.activityId.backend)));
        rootElement.appendChild(element);

        element = document.createElement("comments");
        element.appendChild(document.createTextNode(timeEntry.comments));
        rootElement.appendChild(element);
      }
    });

//TODO
Dprintf.dprintf("");
/*
    // store id
    try
    {
      updateLocalDatabaseId(timeEntry);
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
*/
  }

  /** update time entry on Redmine server
   * @param timeEntry time entry to update
   * @param originTimeEntry origin time entry
   */
  @Override
  public void updateTimeEntry(final TimeEntry timeEntry, final TimeEntry originTimeEntry)
    throws BackendException
  {
    assert(timeEntry.backendType == Backend.Types.REDMINE);
    assert(timeEntry.id.hasBackend());

    if      (timeEntry.id.isNone())
    {
      // add time entry on Redmine server
      storeTimeEntry(timeEntry);
    }
// TODO remove
    /*
    else if (!timeEntry.newIssueId.isNone())
    {
      // issue changed => delete+add time entry on Redmine server
      delete(timeEntry);
      store(timeEntry);
    }*/
    else
    {
      // update time entry on Redmine server
      putData("/time_entries/"+timeEntry.id.backend,"time_entry",timeEntry,new CreateHandler<TimeEntry>()
      {
        public void data(Document document, Element rootElement, TimeEntry timeEntry)
        {
          Element element;

          element = document.createElement("issue_id");
          element.appendChild(document.createTextNode(Integer.toString(timeEntry.issueId.backend)));
          rootElement.appendChild(element);

          element = document.createElement("spent_on");
          element.appendChild(document.createTextNode(timeEntry.spentOn.toString(DATE_FORMAT)));
          rootElement.appendChild(element);

          element = document.createElement("hours");
          element.appendChild(document.createTextNode(Double.toString((double)timeEntry.duration/60.0)));
          rootElement.appendChild(element);

          element = document.createElement("activity_id");
          element.appendChild(document.createTextNode(Integer.toString(timeEntry.activityId.backend)));
          rootElement.appendChild(element);

          element = document.createElement("comments");
          element.appendChild(document.createTextNode(timeEntry.comments));
          rootElement.appendChild(element);
        }
      });
    }

//TODO
Dprintf.dprintf("");
/*
    // store id
    try
    {
      updateLocalDatabaseId(timeEntry);
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
*/
  }

  /** delete time entry on Redmine server
   * @param timeEntry time entry to delete
   */
  @Override
  public void deleteTimeEntry(final TimeEntry timeEntry)
    throws BackendException
  {
    assert(timeEntry.backendType == Backend.Types.REDMINE);
    assert(timeEntry.id.hasBackend());

    if (!timeEntry.id.isNone())
    {
      // delete time entry from Redmine server
      deleteData("/time_entries/"+timeEntry.id.backend);
    }
  }

  //--------------------------------------------------------------------

  /** get users from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public void getUsers(final BackendProgress backendProgress, final EntityHandler<User> entityHandler)
    throws BackendException
  {
    /* Note: bug in Redmine: non-admin user cannot get list of users
       http://www.redmine.org/issues/7773
       And it seems the Redmine team will not fix this issue :-(

       Workaround: first get members and then a list of user id from
                   members. Get info for each user id and store in
                   local database.
    */

/*
    getData("/users/"+userId.backend,"user",new ElementHandler<User>()
    {
      @Override
      public User parse(Element element)
      {
        return new User(getIdValue(element,"id"),
                        getValue(element,"firstname"),
                        getValue(element,"lastname"),
                        getIdAttribute(element,"role","id"),
                        getValue(element,"mail"),
                        getDateValue(element,"created_on"),
                        getDateValue(element,"last_login_on")
                       );
      }
      @Override
      public void store(User user)
        throws BackendException
      {
        entityHandler.handle(user);
        entityHandler.done(user);
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    });
*/
Dprintf.dprintf("");
  }

  /** get current user from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public User getCurrentUser(final BackendProgress backendProgress, final EntityHandler<User> entityHandler)
    throws BackendException
  {
    if (currentUser == null)
    {
      // try to get current user
      try
      {
        currentUser = getData("/users/current","user",new ElementHandler<User>()
        {
          @Override
          public User process(Element element, int index, int totalCount)
            throws BackendException
          {
            return new User(Backend.Types.REDMINE,
                            getIdValue(element,"id"),
                            getValue(element,"firstname"),
                            getValue(element,"lastname"),
                            "",
                            getIdAttribute(element,"role","id"),
                            getValue(element,"mail"),
                            getDateValue(element,"created_on"),
                            getDateValue(element,"last_login_on")
                           );
          }
          @Override
          protected void error(Exception exception)
          {
            // ignored
          }
        });
      }
      catch (InterruptedException exception)
      {
        // ignored
      }
    }

    entityHandler.handle(currentUser);
    entityHandler.done(currentUser);

    return currentUser;
  }

  /** get roles from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getRoles(final BackendProgress backendProgress, final EntityHandler<Role> entityHandler)
    throws BackendException
  {
    try
    {
      getData("/roles","role",new ElementHandler<Role>()
      {
        @Override
        public Role process(Element element, int index, int totalCount)
          throws BackendException
        {
          assert(totalCount > 0);

          Role role = new Role(Backend.Types.REDMINE,
                               getIdValue(element,"id"),
                               new Id(Id.ANY),  // projectId
                               getValue(element,"name",""),
                               ""  // description
                              );
          entityHandler.handle(role);
          entityHandler.done(role);

          backendProgress.setSelection(index,totalCount);

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }
  }

  /** get trackers from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public void getTrackers(final BackendProgress backendProgress, final EntityHandler<Tracker> entityHandler)
    throws BackendException
  {
    try
    {
      getData("/trackers","tracker",new ElementHandler<Tracker>()
      {
        @Override
        public Tracker process(Element element, int index, int totalCount)
          throws BackendException
        {
          assert totalCount > 0;

          Tracker tracker = new Tracker(Backend.Types.REDMINE,
                                        getIdValue(element,"id"),
                                        new Id(Id.ANY),  // projectId
                                        getValue(element,"name",""),
                                        "" // description
                                       );
          entityHandler.handle(tracker);
          entityHandler.done(tracker);

          backendProgress.setSelection(index,totalCount);

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }
  }

  /** get epic names from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public void getEpicNames(final BackendProgress backendProgress, final EntityHandler<EpicName> entityHandler)
    throws BackendException
  {
    // nothing to do
  }

  /** get status from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getStatus(final BackendProgress backendProgress, final EntityHandler<Status> entityHandler)
    throws BackendException
  {
    try
    {
      getData("/issue_statuses","issue_status",new ElementHandler<Status>()
      {
        @Override
        public Status process(Element element, int index, int totalCount)
          throws BackendException
        {
          assert totalCount > 0;

          Status status = new Status(Backend.Types.REDMINE,
                                     getIdValue(element,"id"),
                                     new Id(Id.ANY),  // projectId
                                     getValue(element,"name",""),
                                     "",  // description
                                     getBooleanValue(element,"is_default")
                                    );
          entityHandler.handle(status);
          entityHandler.done(status);

          backendProgress.setSelection(index,totalCount);

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }
  }

  /** get priorities from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getPriorities(final BackendProgress backendProgress, final EntityHandler<Priority> entityHandler)
    throws BackendException
  {
    try
    {
      getData("/enumerations/issue_priorities","issue_priority",new ElementHandler<Priority>()
      {
        @Override
        public Priority process(Element element, int index, int totalCount)
          throws BackendException
        {
          assert totalCount > 0;

          Priority priority = new Priority(Backend.Types.REDMINE,
                                           getIdValue(element,"id"),
                                           new Id(Id.ANY),  // projectId
                                           getValue(element,"name",""),
                                           "",  // description
                                           getBooleanValue(element,"is_default")
                                          );
          entityHandler.handle(priority);
          entityHandler.done(priority);

          backendProgress.setSelection(index,totalCount);

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }
  }

  /** get activities from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getActivities(final BackendProgress backendProgress, final EntityHandler<Activity> entityHandler)
    throws BackendException
  {
    try
    {
      getData("/enumerations/time_entry_activities","time_entry_activity",new ElementHandler<Activity>()
      {
        @Override
        public Activity process(Element element, int index, int totalCount)
          throws BackendException
        {
          assert totalCount > 0;

          Activity activity = new Activity(Backend.Types.REDMINE,
                                           getIdValue(element,"id"),
                                           new Id(Id.ANY),  // projectId
                                           getValue(element,"name",""),
                                           "",  // description
                                           getBooleanValue(element,"is_default")
                                          );
          entityHandler.handle(activity);
          entityHandler.done(activity);

          backendProgress.setSelection(index,totalCount);

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }
  }

  /** get projects from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public void getProjects(final BackendProgress backendProgress, final EntityHandler<Project> entityHandler)
    throws BackendException
  {
    try
    {
      getData("/projects","project",new ElementHandler<Project>()
      {
        @Override
        public Project process(Element element, int index, int totalCount)
          throws BackendException
        {
          assert totalCount > 0;

          Project project = new Project(Backend.Types.REDMINE,
                                        getIdValue(element,"id"),
                                        getIdAttribute(element,"parent","id"),
                                        getValue(element,"name",""),
                                        ProjectStates.toEnum(getIntValue(element,"status")),
                                        getValue(element,"identifier"),
                                        getValue(element,"description",""),
                                        getDateValue(element,"created_on"),
                                        getDateValue(element,"updated_on")
                                       );
          entityHandler.handle(project);
          entityHandler.done(project);

          backendProgress.setSelection(index,totalCount);

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }
  }

  /** get members from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getMembers(final BackendProgress backendProgress, final EntityHandler<Member> entityHandler)
    throws BackendException
  {
    {
      backendProgress.beginSteps(2);
    }
    try
    {
      // get project id list
      backendProgress.setStep(0);
      ArrayList<Integer> projectIdList = getProjectIdList(backendProgress);

      // get members of all projects
      backendProgress.setStep(1);
      for (int i = 0; i < projectIdList.size(); i++)
      {
        try
        {
          final int projectId = projectIdList.get(i);

          getData("/projects/"+projectId+"/memberships","membership",new ElementHandler<User>()
          {
            @Override
            public User process(Element element)
              throws BackendException
            {
              // get first/last name
              String[] names     = getAttribute(element,"user","name","").split(" ",2);
              String   firstName = (names.length >= 1) ? names[0] : "";
              String   lastName  = (names.length >= 2) ? names[1] : "";

              // add member
              if (!firstName.isEmpty() && !lastName.isEmpty())
              {
                Member member = new Member(Backend.Types.REDMINE,
                                           new Id(),
                                           new Id(projectId),
                                           getIdAttribute(element,"user","id"),
                                           firstName,
                                           lastName,
                                           getAttribute(element,"user","name",""),
                                           getIdAttribute(element,"role","id")
                                          );
                entityHandler.handle(member);
                entityHandler.done(member);
              }

              return null;
            }
            @Override
            protected void error(Exception exception)
            {
              // ignored
            }
          });
        }
        catch (BackendException exception)
        {
          // ignored
        }
        backendProgress.setSelection(i,projectIdList.size());

        // throttle execution
        try { Thread.sleep(Settings.throttleTime); } catch (InterruptedException exception) { /* ignored */ }
      }
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
    }
  }

  /** get current member from backend
   * @param backendProgress progress handler or null
   * @param projectId project id
   * @param entityHandler entity handler
   * @return current member or Member.NONE
   */
  public Member getCurrentMember(final BackendProgress backendProgress, final Id projectId, final EntityHandler<Member> entityHandler)
    throws BackendException
  {
    if (currentMember == null)
    {
      // try to get current member
      try
      {
        currentMember = getData("/users/current","user",new ElementHandler<Member>()
        {
          @Override
          public Member process(Element element, int index, int totalCount)
            throws BackendException
          {
            return new Member(Backend.Types.REDMINE,
                              new Id(),
                              projectId,
                              getIdValue(element,"id"),
                              getValue(element,"firstname"),
                              getValue(element,"lastname"),
                              "",
                              new Id()
                             );
          }
          @Override
          protected void error(Exception exception)
          {
            // ignored
          }
        });
      }
      catch (InterruptedException exception)
      {
        // nothing to do
      }
    }

    entityHandler.handle(currentMember);
    entityHandler.done(currentMember);

    return (currentMember != null) ? currentMember : null;
  }

  /** get versions from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getVersions(final BackendProgress backendProgress, final EntityHandler<Version> entityHandler)
    throws BackendException
  {
    {
      backendProgress.beginSteps(2);
    }
    try
    {
      // get all project ids
      backendProgress.setStep(0);
      final ArrayList<Id> projectIdList = new ArrayList<Id>();
      getData("/projects","project",new ElementHandler<Project>()
      {
        @Override
        public Project process(Element element, int index, int totalCount)
          throws BackendException
        {
          assert totalCount > 0;

          Project project = new Project(Backend.Types.REDMINE,
                                        getIdValue(element,"id"),
                                        getIdValue(element,"parent_id"),
                                        getValue(element,"name",""),
                                        ProjectStates.toEnum(getIntValue(element,"status")),
                                        getValue(element,"identifier"),
                                        getValue(element,"description",""),
                                        getDateValue(element,"created_on"),
                                        getDateValue(element,"updated_on")
                                       );

          projectIdList.add(project.id);

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });

      // update versions for projects
      backendProgress.setStep(1);
      for (int i = 0; i < projectIdList.size(); i++)
      {
        try
        {
          getData("/projects/"+projectIdList.get(i).backend+"/versions","version",new ElementHandler<Version>()
          {
            @Override
            public Version process(Element element, int index, int totalCount)
              throws BackendException
            {
              assert totalCount > 0;

              Version version = new Version(Backend.Types.REDMINE,
                                            getIdValue(element,"id"),
                                            getIdAttribute(element,"project","id"),
                                            getValue(element,"name",""),
                                            getValue(element,"description",""),
                                            getValue(element,"status"),
                                            getDateValue(element,"due_date"),
                                            getDateValue(element,"created_on"),
                                            getDateValue(element,"updated_on")
                                           );
              entityHandler.handle(version);
              entityHandler.done(version);

              backendProgress.setSelection(index,totalCount);

              return null;
            }
            @Override
            public void error (Exception exception)
            {
              // ignored, e. g. no access to project
            }
          });

        }
        catch (BackendException exception)
        {
          // ignored
        }
        backendProgress.setSelection(i,projectIdList.size());

        // throttle execution
        try { Thread.sleep(Settings.throttleTime); } catch (InterruptedException exception) { /* ignored */ }
      }
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
    }
  }

  /** get issues from backend
   * @param backendProgress progress handler or null
   * @param lastUpdatedOnTimestamp last updated time stamp for incremental update or 0
   * @param entityHandler entity handler
   */
  @Override
  public void getIssues(final BackendProgress backendProgress, long lastUpdatedOnTimestamp, final EntityHandler<Issue> entityHandler)
    throws BackendException
  {
    try
    {
      getData("/issues","issue",new ElementHandler<Issue>()
      {
        @Override
        public Issue process(Element element, int index, int totalCount)
          throws BackendException
        {
          assert totalCount > 0;

          Issue issue = new Issue(Backend.Types.REDMINE,
                                  getIdValue(element,"id"),
                                  getIdAttribute(element,"parent","id"),
                                  getIdAttribute(element,"project","id"),
                                  getIdAttribute(element,"tracker","id"),
                                  getIdAttribute(element,"status","id"),
                                  getIdAttribute(element,"priority","id"),
                                  getIdAttribute(element,"author","id"),
                                  getValue(element,"subject",""),
                                  getValue(element,"description",""),
                                  getDateValue(element,"start_date"),
                                  getDateValue(element,"due_date"),
                                  getIdAttribute(element,"assigned_to","id",Id.NONE),
                                  getIntValue(element,"done_ratio"),
                                  getIdAttribute(element,"fixed_version","id",Id.ANY),
                                  (long)(getDoubleValue(element,"estimated_hours",Issue.ESTIMATED_NONE)*60.0),
                                  getDateValue(element,"created_on"),
                                  getDateValue(element,"updated_on"),
                                  getDateValue(element,"closed_on")
                                 );
          assert(!issue.id.isNone());
          assert(!issue.projectId.isNone());
          entityHandler.handle(issue);
          entityHandler.done(issue);

          backendProgress.setSelection(index,totalCount);

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      },
      "status_id=*",
      (lastUpdatedOnTimestamp > 0) ? "updated_on="+FILTER_GE+ISO_DATETIME_FORMAT.format(lastUpdatedOnTimestamp*1000) : null
      );
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }
  }

  /** get time entry from backend
   * @param backendProgress progress handler or null
   * @param timeEntryId time entry id
   * @return time entry
   */
  public TimeEntry getTimeEntry(final BackendProgress backendProgress, final long createdOnTimestamp, final EntityHandler<TimeEntry> entityHandler)
    throws BackendException, InterruptedException
  {
// TODO: implement
return null;
  }

  /** get time entries from backend
   * @param backendProgress progress handler or null
   * @param createdOnTimestamp created-on timestamp
   * @param entityHandler entity handler
   */
  public void getTimeEntries(final BackendProgress backendProgress, final long createdOnTimestamp, final EntityHandler<TimeEntry> entityHandler)
    throws BackendException
  {
//TODO: use currentUser?
    try
    {
      // get own user id
      final Id ownUserId = getData("/users/current","user",new ElementHandler<Id>()
      {
        @Override
        public Id parse(Element element)
        {
          return getIdValue(element,"id");
        }
      });

      // get time entry data and store into local database
      getData("/time_entries","time_entry",new ElementHandler<TimeEntry>()
      {
        @Override
        public TimeEntry process(Element element, int index, int totalCount)
          throws BackendException
        {
          assert totalCount > 0;

  //Dprintf.dprintf("element=%s projectId=%d issueId=%d userId=%d spentOn=%s",element,getIntAttribute(element,"project","id"),getIntAttribute(element,"issue","id"),getIntAttribute(element,"user","id"),getDateValue(element,"spent_on"),getDoubleValue(element,"hours"));
          TimeEntry timeEntry = new TimeEntry(Backend.Types.REDMINE,
                                              getIdValue(element,"id"),
                                              getIdAttribute(element,"project","id"),
                                              getIdAttribute(element,"issue","id"),
                                              getIdAttribute(element,"user","id"),
                                              getIdAttribute(element,"activity","id"),
                                              (long)(getDoubleValue(element,"hours")*60.0),
                                              getValue(element,"comments"),
                                              getSpentOn(element,"spent_on"),
                                              getDateValue(element,"created_on"),
                                              getDateValue(element,"updated_on")
                                             );
          assert(!timeEntry.id.isNone());
          assert(!timeEntry.projectId.isNone());
  Dprintf.dprintf("");
  //        assert(!timeEntry.issueId.isNone());
          entityHandler.handle(timeEntry);
          entityHandler.done(timeEntry);

          backendProgress.setSelection(index,totalCount);

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      },
      ownUserId.isSet() ? "user_id="+ownUserId.backend : null,
      "spent_on="+FILTER_GE+ISO_DATETIME_FORMAT.format(createdOnTimestamp*1000)
      );
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }
  }

  //--------------------------------------------------------------------

  /** get data
   * @param urlString URL string
   * @param name XML entry name
   * @param offset start offset
   * @param length max. number of elements to get
   * @param elementHandler element handler
   * @param additionalArguments additional arguments (optional)
   * @return parsed entity
   */
  private <T> T getData(String            urlString,
                        String            name,
                        int               offset,
                        int               length,
                        ElementHandler<T> elementHandler,
                        String...         additionalArguments
                       )
    throws BackendException, InterruptedException
  {
    T entity = null;

    if (Thread.currentThread().isInterrupted())
    {
      throw new InterruptedException();
    }

    boolean done = false;
    while (!done)
    {
      HttpURLConnection connection = null;
      try
      {
        Document document;
        NodeList nodeList;
        Node     node;

        DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
        DocumentBuilder        documentBuilder        = documentBuilderFactory.newDocumentBuilder();

        // get data from Redmine server
        String[] arguments = Arrays.copyOf(additionalArguments,additionalArguments.length+2);
        arguments[additionalArguments.length+0] = "offset="+offset;
        arguments[additionalArguments.length+1] = "limit="+((length > 0) ? Math.min(length,ENTRY_LIMIT) : ENTRY_LIMIT);
        connection = getConnection(urlString+".xml","GET",false,arguments);
        Timemine.printDebug(Settings.DebugTypes.GET,"Get '%s'",connection.getURL());

        // check result
        int responseCode = connection.getResponseCode();
        if ((responseCode == 401) || (responseCode == 403))
        {
          throw new BackendException("Authorization failure");
        }

        // parse XML
        document = documentBuilder.parse(connection.getInputStream());
        document.getDocumentElement().normalize();

        // close connectin
        connection.disconnect(); connection = null;

        // process root element, get number of total entries
        Element rootElement = (Element)document.getDocumentElement();
        elementHandler.root(rootElement);

        // get number of total entries, process data elements
        int index = offset;
        nodeList = document.getElementsByTagName(name);
        int totalCount = getIntAttribute(rootElement,"total_count");
        if (totalCount == 0) totalCount = nodeList.getLength();
        for (int i = 0; i < nodeList.getLength(); i++)
        {
          node = nodeList.item(i);

          if (node.getNodeType() == Node.ELEMENT_NODE)
          {
            entity = elementHandler.process((Element)node,index,totalCount);
          }
          index++;
          if (length > 0) length--;

          if (elementHandler.isDone()) break;
        }

        // check if done
        done =    elementHandler.isDone()
               || (length == 0)
               || (index >= totalCount)
               || (offset >= totalCount);

        // next segment
        offset += ENTRY_LIMIT;

        // throttle execution
        if (!done)
        {
          try { Thread.sleep(Settings.throttleTime); } catch (InterruptedException exception) { /* ignored */ }
        }
      }
      catch (ParserConfigurationException exception)
      {
        Timemine.printDebug("get data fail '%s': %s",connection,exception.getMessage());
        throw new BackendException("XML parse fail",exception);
      }
      catch (SAXException exception)
      {
        Timemine.printDebug("get data fail '%s': %s",connection,exception.getMessage());
        throw new BackendException("XML parse fail",exception);
      }
      catch (ProtocolException exception)
      {
        Timemine.printDebug("Get data fail '%s': %s",connection,exception.getMessage());
        throw new BackendException(exception);
      }
      catch (UnknownHostException exception)
      {
        Timemine.printDebug("Get data fail '%s': %s",connection,exception.getMessage());
        throw new BackendException("Unknown host '"+serverName+"'");
      }
      catch (IOException exception)
      {
        Timemine.printDebug("Get data fail '%s': %s",connection,exception.getMessage());
        elementHandler.error(exception);
        break;
      }
      finally
      {
        if (connection != null) connection.disconnect();
      }
    }

    return entity;
  }

  /** get all data
   * @param urlString URL string
   * @param name XML entry name
   * @param elementHandler element handler
   * @param additionalArguments additional arguments (optional)
   * @return parsed entity
   */
  private <T> T getData(String            urlString,
                        String            name,
                        ElementHandler<T> elementHandler,
                        String...         additionalArguments
                       )
    throws BackendException, InterruptedException
  {
    return getData(urlString,name,0,-1,elementHandler,additionalArguments);
  }

  /** iterate over all data
   * @param urlString URL string
   * @param name XML entry name
   * @param elementHandler element handler
   * @param additionalArguments additional arguments (optional)
   * @return result
   */
  private <T> Object iterateData(String            urlString,
                                 String            name,
                                 ElementHandler<T> elementHandler,
                                 String...         additionalArguments
                                )
    throws BackendException, InterruptedException
  {
    getData(urlString,name,elementHandler,additionalArguments);

    return elementHandler.getResult();
  }

  /** get length of data
   * @param urlString URL string
   * @param name XML entry name
   * @return length of data
   */
  private int getDataLength(String urlString, String name)
    throws BackendException
  {
    final int[] result = new int[1];

    try
    {
      getData(urlString,name,0,1,new ElementHandler()
      {
        public void root(Element element)
        {
          result[0] = getIntAttribute(element,"total_count");
        }
      });
      Timemine.printDebug(Settings.DebugTypes.GET,"Get length '"+urlString+"': "+result[0]);
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }

    return result[0];
  }

  /** add data via POST
   * @param urlString URL string
   * @param name XML entry name
   * @param entity entity for storage
   * @param createHandler create handler
   * @param additionalArguments additional arguments (optional)
   */
  private void postData(String urlString, String name, Entity entity, CreateHandler createHandler, String... additionalArguments)
    throws BackendException
  {
    if (!Settings.readOnlyFlag)
    {
      HttpURLConnection connection = null;
      try
      {
        DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
        DocumentBuilder        documentBuilder        = documentBuilderFactory.newDocumentBuilder();
        Document               document;
        TransformerFactory     transformerFactory     = TransformerFactory.newInstance();

        // create XML
        document = documentBuilder.newDocument();
        document.setXmlStandalone(true);
        Element rootElement = document.createElement(name);
        document.appendChild(rootElement);
        createHandler.data(document,rootElement,entity);

        // debug
        if (Settings.debugTypes.contains(Settings.DebugTypes.WRITE))
        {
          Transformer transformer = transformerFactory.newTransformer();
          transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION,"yes");
          StringWriter stringWriter = new StringWriter();
          transformer.transform(new DOMSource(document), new StreamResult(stringWriter));

          Dprintf.printStackTrace();
          if (!debugConfirm("POST: "+urlString+".xml\n"+
                            "XML : "+stringWriter.getBuffer().toString()
                           )
             )
          {
            return;
          }
        }

        // add data on Redmine server
        connection = getConnection(urlString+".xml","POST",true,additionalArguments);
        connection.setRequestProperty("Content-Type","application/xml; charset=utf-8");
        Timemine.printDebug(Settings.DebugTypes.PUT,"Post '"+connection.getURL()+"'");

        // output data
        Transformer transformer = transformerFactory.newTransformer();
        transformer.setOutputProperty(OutputKeys.ENCODING,"UTF-8");
        DOMSource domSource = new DOMSource(document);
        if (Settings.debugTypes.contains(Settings.DebugTypes.PUT))
        {
          Timemine.printDebug("xml data ");
          StreamResult streamResult = new StreamResult(System.err);
          transformer.transform(domSource,streamResult);
          System.err.println("");
        }
        StreamResult streamResult = new StreamResult(connection.getOutputStream());
        transformer.transform(domSource,streamResult);

        // check result
        int responseCode = connection.getResponseCode();
        if ((responseCode == 401) || (responseCode == 403))
        {
          throw new BackendException("Authorization failure");
        }
        String responseMessage = connection.getResponseMessage();
        Timemine.printDebug(Settings.DebugTypes.PUT,"Result "+responseCode+": "+((responseMessage != null) ? responseMessage : "OK"));
        if (connection.getResponseCode() != 201)
        {
          throw new BackendException("Server error "+responseCode+": "+((responseMessage != null) ? responseMessage : "unknown"));
        }

        // get entity id
        document = documentBuilder.parse(connection.getInputStream());
        document.getDocumentElement().normalize();

        NodeList nodeList = document.getElementsByTagName(name);
        for (int i = 0; i < nodeList.getLength(); i++)
        {
          Node node = nodeList.item(i);

          if (node.getNodeType() == Node.ELEMENT_NODE)
          {
            entity.id.backend = getIntValue((Element)node,"id");
            break;
          }
        }
        if (entity.id.isNone())
        {
          throw new BackendException("No id for created entity returned");
        }

        // close connectin
        connection.disconnect(); connection = null;
      }
      catch (ParserConfigurationException exception)
      {
        throw new BackendException("XML parse fail: "+exception.getMessage());
      }
      catch (SAXException exception)
      {
        throw new BackendException("XML parse fail: "+exception.getMessage());
      }
      catch (TransformerException exception)
      {
        throw new BackendException(exception);
      }
      catch (UnknownHostException exception)
      {
        throw new BackendException("Unknown host '"+serverName+"'");
      }
      catch (IOException exception)
      {
        throw new BackendException("Send data fail: "+exception.getMessage());
      }
      finally
      {
        if (connection != null) connection.disconnect();
      }
    }
  }

  /** update data via PUT
   * @param urlString URL string
   * @param name XML entry name
   * @param entity entity for storage
   * @param createHandler create handler
   * @param additionalArguments additional arguments (optional)
   */
  private void putData(String urlString, String name, Entity entity, CreateHandler createHandler, String... additionalArguments)
    throws BackendException
  {
    if (!Settings.readOnlyFlag)
    {
      HttpURLConnection connection = null;
      try
      {
        DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
        DocumentBuilder        documentBuilder        = documentBuilderFactory.newDocumentBuilder();
        Document               document;
        TransformerFactory     transformerFactory     = TransformerFactory.newInstance();

        // create XML
        document = documentBuilder.newDocument();
        document.setXmlStandalone(true);
        Element rootElement = document.createElement(name);
        document.appendChild(rootElement);
        createHandler.data(document,rootElement,entity);

        // debug
        if (Settings.debugTypes.contains(Settings.DebugTypes.WRITE))
        {
          Transformer transformer = transformerFactory.newTransformer();
          transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION,"yes");
          StringWriter stringWriter = new StringWriter();
          transformer.transform(new DOMSource(document), new StreamResult(stringWriter));
  Dprintf.dprintf("%s",stringWriter.getBuffer().toString());

          Dprintf.printStackTrace();
          if (!debugConfirm("PUT: "+urlString+".xml\n"+
                            "XML : "+stringWriter.getBuffer().toString()
                           )
             )
          {
            return;
          }
        }

        // update data on Redmine server
        connection = getConnection(urlString+".xml","PUT",true,additionalArguments);
        connection.setRequestProperty("Content-Type","application/xml; charset=utf-8");
        Timemine.printDebug(Settings.DebugTypes.PUT,"Put '"+connection.getURL()+"'");

        // output data
        Transformer transformer = transformerFactory.newTransformer();
        transformer.setOutputProperty(OutputKeys.ENCODING,"UTF-8");
        DOMSource domSource = new DOMSource(document);
        if (Settings.debugTypes.contains(Settings.DebugTypes.PUT))
        {
          Timemine.printDebug("xml data");
          StreamResult streamResult = new StreamResult(System.err);
          transformer.transform(domSource,streamResult);
          System.err.println();
        }
        StreamResult streamResult = new StreamResult(connection.getOutputStream());
        transformer.transform(domSource,streamResult);

        // check result
        int responseCode = connection.getResponseCode();
        if ((responseCode == 401) || (responseCode == 403))
        {
          throw new BackendException("Authorization failure");
        }
        String responseMessage = connection.getResponseMessage();
        Timemine.printDebug(Settings.DebugTypes.PUT,"Result "+responseCode+": "+((responseMessage != null) ? responseMessage : "OK"));
        if (responseCode != HttpURLConnection.HTTP_OK)
        {
          throw new BackendException("Server error "+responseCode+": "+((responseMessage != null) ? responseMessage : "unknown"));
        }

        // close connectin
        connection.disconnect(); connection = null;
      }
      catch (ParserConfigurationException exception)
      {
        throw new BackendException("XML parse fail: "+exception.getMessage());
      }
      catch (TransformerException exception)
      {
        throw new BackendException("XML parse fail: "+exception.getMessage());
      }
      catch (UnknownHostException exception)
      {
        throw new BackendException("Unknown host '"+serverName+"'");
      }
      catch (IOException exception)
      {
        throw new BackendException("Send data fail: "+exception.getMessage());
      }
      finally
      {
        if (connection != null) connection.disconnect();
      }
    }
  }

  /** delete data
   * @param urlString URL string
   * @param name XML entry name
   * @param additionalArguments additional arguments (optional)
   */
  private void deleteData(String urlString, String... additionalArguments)
    throws BackendException
  {
    if (!Settings.readOnlyFlag)
    {
      HttpURLConnection connection = null;
      try
      {
        // debug
        if (Settings.debugTypes.contains(Settings.DebugTypes.DELETE))
        {
          Dprintf.printStackTrace();
          if (!debugConfirm("DELETE: "+urlString+".xml"
                           )
             )
          {
            return;
          }
        }

        // delete data from Redmine server
        connection = getConnection(urlString+".xml","DELETE",true,additionalArguments);
        connection.setRequestProperty("Content-Type","application/xml; charset=utf-8");
        Timemine.printDebug(Settings.DebugTypes.DELETE,"Delete '"+connection.getURL()+"'");

        // check result
        int responseCode = connection.getResponseCode();
        if ((responseCode == 401) || (responseCode == 403))
        {
          throw new BackendException("Authorization failure");
        }
        String responseMessage = connection.getResponseMessage();
        Timemine.printDebug(Settings.DebugTypes.PUT,"Result "+responseCode+": "+((responseMessage != null) ? responseMessage : "OK"));
        if ((responseCode != HttpURLConnection.HTTP_OK) && (responseCode != HttpURLConnection.HTTP_NOT_FOUND))
        {
          throw new BackendException("Server error "+responseCode+": "+((responseMessage != null) ? responseMessage : "unknown"));
        }

        // close connectin
        connection.disconnect(); connection = null;
      }
      catch (UnknownHostException exception)
      {
        throw new BackendException("Unknown host '"+serverName+"'");
      }
      catch (IOException exception)
      {
        throw new BackendException("Send data fail: "+exception.getMessage());
      }
      finally
      {
        if (connection != null) connection.disconnect();
      }
    }
  }

  /** get value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param defaultValue default value
   * @return value
   */
  private String getValue(Element element, String tagName, String defaultValue)
  {
    String value = defaultValue;

    NodeList nodeList = element.getElementsByTagName(tagName);
    if ((nodeList != null) && (nodeList.getLength() > 0))
    {
      NodeList subNodeList = nodeList.item(0).getChildNodes();
      if (subNodeList != null)
      {
        Node node = (Node)subNodeList.item(0);
        if (node != null)
        {
          value = node.getNodeValue();
        }
      }
    }

    return value;
  }

  /** get value from XML element
   * @param element XML element
   * @param tagName tag name
   * @return value
   */
  private String getValue(Element element, String tagName)
  {
    return getValue(element,tagName,"");
  }

  /** get id value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param defaultValue default value
   * @return id value
   */
  private Id getIdValue(Element element, String tagName, int defaultValue)
  {
    Id id = new Id(defaultValue);

    try
    {
      String value = getValue(element,tagName).trim();
      if (!value.isEmpty())
      {
        id.backend = Integer.parseInt(value);
      }
    }
    catch (NumberFormatException exception)
    {
      // ignored
    }

    return id;
  }

  /** get id value from XML element
   * @param element XML element
   * @param tagName tag name
   * @return id value
   */
  private Id getIdValue(Element element, String tagName)
  {
    return getIdValue(element,tagName,Id.NONE);
  }

  /** get boolean value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param defaultValue default value
   * @return boolean value
   */
  private boolean getBooleanValue(Element element, String tagName, boolean defaultValue)
  {
    String value = getValue(element,tagName).trim();
    if (!value.isEmpty())
    {
      try
      {
        return Integer.parseInt(value) != 0;
      }
      catch (NumberFormatException exception)
      {
        return    value.equalsIgnoreCase("true")
               || value.equalsIgnoreCase("on")
               || value.equalsIgnoreCase("yes");
      }
    }
    else
    {
      return defaultValue;
    }
  }

  /** get boolean value from XML element
   * @param element XML element
   * @param tagName tag name
   * @return boolean value
   */
  private boolean getBooleanValue(Element element, String tagName)
  {
    return getBooleanValue(element,tagName,false);
  }

  /** get integer value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param defaultValue default value
   * @return integer value
   */
  private int getIntValue(Element element, String tagName, int defaultValue)
  {
    try
    {
      String value = getValue(element,tagName).trim();
      if (!value.isEmpty())
      {
        return Integer.parseInt(value);
      }
      else
      {
        return defaultValue;
      }
    }
    catch (NumberFormatException exception)
    {
      return 0;
    }
  }

  /** get integer value from XML element
   * @param element XML element
   * @param tagName tag name
   * @return integer value
   */
  private int getIntValue(Element element, String tagName)
  {
    return getIntValue(element,tagName,0);
  }

  /** get long value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param defaultValue default value
   * @return long value
   */
  private long getLongValue(Element element, String tagName, long defaultValue)
  {
    try
    {
      String value = getValue(element,tagName).trim();
      if (!value.isEmpty())
      {
        return Long.parseLong(value);
      }
      else
      {
        return defaultValue;
      }
    }
    catch (NumberFormatException exception)
    {
      return 0L;
    }
  }

  /** get double from XML element
   * @param element XML element
   * @param tagName tag name
   * @param defaultValue default value
   * @return double value
   */
  private double getDoubleValue(Element element, String tagName, double defaultValue)
  {
    try
    {
      String value = getValue(element,tagName).trim();
      if (!value.isEmpty())
      {
        return Double.parseDouble(value);
      }
      else
      {
        return defaultValue;
      }
    }
    catch (NumberFormatException exception)
    {
      return 0.0;
    }
  }

  /** get double from XML element
   * @param element XML element
   * @param tagName tag name
   * @return double value
   */
  private double getDoubleValue(Element element, String tagName)
  {
    return getDoubleValue(element,tagName,0.0);
  }

  /** get enum from XML element
   * @param element XML element
   * @param tagName tag name
   * @param defaultValue default value
   * @param clazz enum class
   * @return enum value
   */
  private <T extends Enum<T>> T getEnumValue(Element element, String tagName, T defaultValue, Class<T> clazz)
  {
    String text = getValue(element,tagName).toLowerCase().trim();

    if (!text.isEmpty())
    {
      // find enum value by name
      for (T value : clazz.getEnumConstants())
      {
        if (text.equals(value.name().toLowerCase()))
        {
          return value;
        }
      }

      // find enum value by value
      try
      {
        int n = Integer.parseInt(text);
        for (T value : clazz.getEnumConstants())
        {
          if (value.ordinal() == n)
          {
            return value;
          }
        }
      }
      catch (NumberFormatException exception)
      {
        // nothing to do
      }
    }

    return defaultValue;
  }

  /** get date value from XML element
   * @param element XML element
   * @param tagName tag name
   * @return date value
   */
  private Date getDateValue(Element element, String tagName)
  {
    return parseDate(getValue(element,tagName));
  }

  /** get spent-on value from XML element
   * @param element XML element
   * @param tagName tag name
   * @return spent-on value
   */
  private SpentOn getSpentOn(Element element, String tagName)
  {
    return new SpentOn(getDateValue(element,tagName));
  }

  /** get attribute value from XML element
   * @param element XML element
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return value
   */
  private String getAttribute(Element element, String attributeName, String defaultValue)
  {
    String value = defaultValue;

    Attr attribute = element.getAttributeNode(attributeName);
    if (attribute != null)
    {
      value = attribute.getValue();
    }

    return value;
  }

  /** get attribute value from XML element
   * @param element XML element
   * @param attributeName attribute name
   * @return value
   */
  private String getAttribute(Element element, String attributeName)
  {
    return getAttribute(element,attributeName,"");
  }

  /** get id attribute value from XML element
   * @param element XML element
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return id value
   */
  private Id getIdAttribute(Element element, String attributeName, int defaultValue)
  {
    Id id = new Id(defaultValue);

    try
    {
      String value = getAttribute(element,attributeName).trim();
      if (!value.isEmpty())
      {
        id.backend = Integer.parseInt(value);
      }
    }
    catch (NumberFormatException exception)
    {
      // ignored
    }

    return id;
  }

  /** get id attribute value from XML element
   * @param element XML element
   * @param attributeName attribute name
   * @return id value
   */
  private Id getIdAttribute(Element element, String attributeName)
  {
    return getIdAttribute(element,attributeName,Id.NONE);
  }

  /** get boolean attribute value from XML element
   * @param element XML element
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return boolean value
   */
  private boolean getBooleanAttribute(Element element, String attributeName, boolean defaultValue)
  {
    String value = getAttribute(element,attributeName).trim();
    if (!value.isEmpty())
    {
      try
      {
        return Integer.parseInt(value) != 0;
      }
      catch (NumberFormatException exception)
      {
        return    value.equalsIgnoreCase("true")
               || value.equalsIgnoreCase("on")
               || value.equalsIgnoreCase("yes");
      }
    }
    else
    {
      return defaultValue;
    }
  }

  /** get boolean attribute value from XML element
   * @param element XML element
   * @param attributeName attribute name
   * @return boolean value
   */
  private boolean getBooleanAttribute(Element element, String attributeName)
  {
    return getBooleanAttribute(element,attributeName,false);
  }

  /** get integer attribute value from XML element
   * @param element XML element
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return integer value
   */
  private int getIntAttribute(Element element, String attributeName, int defaultValue)
  {
    try
    {
      String value = getAttribute(element,attributeName).trim();
      if (!value.isEmpty())
      {
        return Integer.parseInt(value);
      }
      else
      {
        return defaultValue;
      }
    }
    catch (NumberFormatException exception)
    {
      return 0;
    }
  }

  /** get integer attribute value from XML element
   * @param element XML element
   * @param attributeName attribute name
   * @return integer value
   */
  private int getIntAttribute(Element element, String attributeName)
  {
    return getIntAttribute(element,attributeName,0);
  }

  /** get long attribute value from XML element
   * @param element XML element
   * @param attributeName tag name
   * @param defaultValue default value
   * @return long value
   */
  private long getLongAttribute(Element element, String attributeName, long defaultValue)
  {
    try
    {
      String value = getAttribute(element,attributeName).trim();
      if (!value.isEmpty())
      {
        return Long.parseLong(value);
      }
      else
      {
        return defaultValue;
      }
    }
    catch (NumberFormatException exception)
    {
      return 0L;
    }
  }

  /** get long attribute value from XML element
   * @param element XML element
   * @param attributeName tag name
   * @return long value
   */
  private long getLongAttribute(Element element, String attributeName)
  {
    return getLongAttribute(element,attributeName,0L);
  }

  /** get double attribute value from XML element
   * @param element XML element
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return double value
   */
  private double getDoubleAttribute(Element element, String attributeName, double defaultValue)
  {
    try
    {
      String value = getAttribute(element,attributeName).trim();
      if (!value.isEmpty())
      {
        return Double.parseDouble(value);
      }
      else
      {
        return defaultValue;
      }
    }
    catch (NumberFormatException exception)
    {
      return 0.0;
    }
  }

  /** get double attribute value from XML element
   * @param element XML element
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return double value
   */
  private double getDoubleAttribute(Element element, String attributeName)
  {
    return getDoubleAttribute(element,attributeName,0.0);
  }

  /** get date attribute value from XML element
   * @param element XML element
   * @param attributeName attribute name
   * @return date value
   */
  private Date getDateAttribute(Element element, String attributeName)
  {
    return parseDate(getAttribute(element,attributeName));
  }

  /** get attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return value
   */
  private String getAttribute(Element element, String tagName, String attributeName, String defaultValue)
  {
    String value = defaultValue;

    Node node = element.getElementsByTagName(tagName).item(0);
    if (node != null)
    {
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
        value = getAttribute((Element)node,attributeName);
      }
    }

    return value;
  }

  /** get id attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return id value
   */
  private Id getIdAttribute(Element element, String tagName, String attributeName, int defaultValue)
  {
    Id id = new Id(defaultValue);

    try
    {
      String value = getAttribute(element,tagName,attributeName,"" /* default */).trim();
      if (!value.isEmpty())
      {
        id.backend = Integer.parseInt(value);
      }
    }
    catch (NumberFormatException exception)
    {
      // ignored
    }

    return id;
  }

  /** get id attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return id value
   */
  private Id getIdAttribute(Element element, String tagName, String attributeName)
  {
    return getIdAttribute(element,tagName,attributeName,Id.NONE);
  }

  /** get boolean attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return boolean value
   */
  private boolean getBooleanValue(Element element, String tagName, String attributeName, boolean defaultValue)
  {
    String value = getAttribute(element,tagName,attributeName).trim();
    if (!value.isEmpty())
    {
      try
      {
        return Integer.parseInt(value) != 0;
      }
      catch (NumberFormatException exception)
      {
        return    value.equalsIgnoreCase("true")
               || value.equalsIgnoreCase("on")
               || value.equalsIgnoreCase("yes");
      }
    }
    else
    {
      return defaultValue;
    }
  }

  /** get boolean attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @return boolean value
   */
  private boolean getBooleanValue(Element element, String tagName, String attributeName)
  {
    return getBooleanValue(element,tagName,attributeName,false);
  }

  /** get integer attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return integer value
   */
  private int getIntAttribute(Element element, String tagName, String attributeName, int defaultValue)
  {
    try
    {
      String value = getAttribute(element,tagName,attributeName).trim();
      if (!value.isEmpty())
      {
        return Integer.parseInt(value);
      }
      else
      {
        return defaultValue;
      }
    }
    catch (NumberFormatException exception)
    {
      return 0;
    }
  }

  /** get integer attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return integer value
   */
  private int getIntAttribute(Element element, String tagName, String attributeName)
  {
    return getIntAttribute(element,tagName,attributeName,0);
  }

  /** get long attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return long value
   */
  private long getLongAttribute(Element element, String tagName, String attributeName, long defaultValue)
  {
    try
    {
      String value = getAttribute(element,tagName,attributeName);
      if (!value.isEmpty())
      {
        return Long.parseLong(value);
      }
      else
      {
        return defaultValue;
      }
    }
    catch (NumberFormatException exception)
    {
      return 0L;
    }
  }

  /** get long attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @return long value
   */
  private long getLongAttribute(Element element, String tagName, String attributeName)
  {
    return getLongAttribute(element,tagName,attributeName,0L);
  }

  /** get double attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @param defaultValue default value
   * @return double value
   */
  private double getDoubleAttribute(Element element, String tagName, String attributeName, double defaultValue)
  {
    try
    {
      String value = getAttribute(element,tagName,attributeName);
      if (!value.isEmpty())
      {
        return Double.parseDouble(value);
      }
      else
      {
        return defaultValue;
      }
    }
    catch (NumberFormatException exception)
    {
      return 0.0;
    }
  }

  /** get double attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @return double value
   */
  private double getDoubleAttribute(Element element, String tagName, String attributeName)
  {
    return getDoubleAttribute(element,tagName,attributeName,0.0);
  }

  /** get date attribute value from XML element
   * @param element XML element
   * @param tagName tag name
   * @param attributeName attribute name
   * @return date value
   */
  private Date getDateAttribute(Element element, String tagName, String attributeName)
  {
    return parseDate(getAttribute(element,tagName,attributeName));
  }

  //--------------------------------------------------------------------

  /** parse date
   * @param string date/time string
   * @return date/time or current date/time
   */
  private Date parseDate(String string)
  {
    Date date = null;

//string="2012-09-11T16:44:32Z";
//string="2015-03-30T15:10:23Z";
    int i = 0;
    while ((i < dateFormats.length) && (date == null))
    {
      try
      {
//Dprintf.dprintf("%s: %s",dateFormats[i],string);
        synchronized(dateFormats[i])
        {
          date = dateFormats[i].parse(string);
        }
//Dprintf.dprintf("date=%s",date);
      }
      catch (ParseException exception)
      {
        // ignored, try next format
      }
catch (Exception e)
{
e.printStackTrace();
Dprintf.dprintf("%s: %s",dateFormats[i],string);
System.exit(1);
}
      i++;
    }
//Dprintf.dprintf("date=%s",date);
//System.exit(100);

    return (date != null) ? date : new Date();
  }

  /** get list of all project ids from Redmine server
   * @param backendProgress progress handler
   * @return project backend id list
   */
  private ArrayList<Integer> getProjectIdList(final BackendProgress backendProgress)
  {
    final ArrayList<Integer> projectIdList = new ArrayList<Integer>();

    try
    {
      getData("/projects","project",new ElementHandler<Void>()
      {
        @Override
        public Void process(Element element, int index, int totalCount)
          throws BackendException
        {
          assert totalCount > 0;

          projectIdList.add(getIntValue(element,"id",Id.NONE));

          backendProgress.setSelection(index,totalCount);

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
    }
    catch (InterruptedException exception)
    {
      // ignored
    }
    catch (BackendException exception)
    {
      // ignored
    }

    return projectIdList;
  }
}

/* end of file */
