/***********************************************************************\
*
* $Revision: 658 $
* $Date: 2018-02-21 17:35:14 +0100 (Wed, 21 Feb 2018) $
* $Author: torsten $
* Contents: Timemine preferences
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Constructor;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStreamWriter;

import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

import java.text.NumberFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.AbstractMap;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.Deque;
import java.util.Timer;
import java.util.TimerTask;

// SQLite database
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

// Nebula
import org.eclipse.nebula.widgets.cdatetime.CDT;
import org.eclipse.nebula.widgets.cdatetime.CDateTime;

// graphics
import java.awt.Desktop;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.custom.TreeEditor;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.MouseWheelListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.events.ShellListener;
import org.eclipse.swt.events.TreeEvent;
import org.eclipse.swt.events.TreeListener;
import org.eclipse.swt.events.VerifyEvent;
import org.eclipse.swt.events.VerifyListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.graphics.Region;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.graphics.TextLayout;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.ColorDialog;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.DateTime;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.TabFolder;
import org.eclipse.swt.widgets.TabItem;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Tray;
import org.eclipse.swt.widgets.TrayItem;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.swt.widgets.TreeColumn;
import org.eclipse.swt.widgets.TreeItem;
import org.eclipse.swt.widgets.TypedListener;
import org.eclipse.swt.widgets.Widget;

/****************************** Classes ********************************/

/** Preferences
 */
public class Preferences
{
  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** edit preferences
   * @param shell parent shell
   */
  public Preferences(final Shell shell)
  {
    /** dialog data
     */
    class Data
    {
      Data()
      {
      }
    };

    final Data    data = new Data();
    final Display display;
    final Shell   dialog;
    TabFolder     tabFolder;
    Composite     composite,subComposite,subSubComposite;
    Separator     separator;
    Label         label;
    Combo         combo;
    Button        button;

    // init variables
    display = shell.getDisplay();

    // repository edit dialog
    dialog = Dialogs.openModal(shell,"Preferences",400,400,new double[]{1.0,0.0},1.0);

    final Button[]  widgetEnabled      = new Button[2];
    final Text[]    widgetServerName   = new Text[2];
    final Spinner[] widgetServerPort   = new Spinner[2];
    final Button[]  widgetServerUseSSL = new Button[2];
    final Text[]    widgetLoginName    = new Text[2];

    final Combo     widgetIssueUpdateStatus;
    final Combo     widgetIssueCompletedStatus;
    final Table     widgetIssueActiveStatus;
    final Combo     widgetNewIssueTrackerRedmine,widgetNewIssueTrackerJIRA;
    final Combo     widgetNewIssueStatusRedmine,widgetNewIssueStatusJIRA;

    final Table     widgetManagerRoles;
    final Table     widgetDeveloperRoles;
    final Table     widgetDesignerRoles;

    final Combo     widgetRefreshTime;
    final Combo     widgetRefreshIssuesTime;
    final Combo     widgetIncrementalRefreshIssuesTime;
    final Combo     widgetRefreshTimeEntriesTime;
    final Combo     widgetIncrementalRefreshTimeEntriesTime;
    final Combo     widgetMaxIssuesHistory;
    final Combo     widgetMaxTimeEntriesHistory;
    final Combo     widgetSyncTime;
    final Combo     widgetThrottleTime;
    final Button    widgetRefreshAtStart;
    final Button    widgetAutoRefresh;
    final DateTime  widgetAutoRefreshStartTime,widgetAutoRefreshEndTime;

    final TreeCombo widgetVacationProjects;
    final TreeCombo widgetVacationIssueSummaries;
    final Label     widgetVacationActivitiesTitle;
    final Combo     widgetVacationActivities;
    final Text      widgetVacationComment;

    final Table     widgetColors;

    final Table     widgetActionFilters;

    final Text      widgetDateFormat;
    final Spinner   widgetMinTimeDelta;
    final Spinner   widgetWorkingHoursPerDay;
    final Spinner   widgetWorkingHoursPerWeek;
    final Spinner   widgetPauseMinutesPerDay;
    final Button    widgetSnapTimeRanges;
    final Table     widgetShowFlags;

    final Button    widgetButtonSave;

    // create tab
    tabFolder = Widgets.newTabFolder(dialog);
    Widgets.layout(tabFolder,0,0,TableLayoutData.NSWE);
    {
      composite = Widgets.addTab(tabFolder,"Server");
      composite.setLayout(new TableLayout(0.0,1.0,2));
      Widgets.layout(composite,0,0,TableLayoutData.NSWE);
      {
        // --- Redmine -------------------------------------------------

        separator = Widgets.newSeparator(composite,Backend.Types.REDMINE.toString());
        separator.setLayout(new TableLayout(0.0,1.0));
        Widgets.layout(separator,0,0,TableLayoutData.WE);

        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0}));
        Widgets.layout(subComposite,1,0,TableLayoutData.WE);
        {
          widgetEnabled[0] = Widgets.newCheckbox(subComposite);
          widgetEnabled[0].setSelection(Settings.Redmine.enabled);
          Widgets.layout(widgetEnabled[0],0,0,TableLayoutData.NW);

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0,0.0,0.0,0.0}));
          Widgets.layout(subSubComposite,0,1,TableLayoutData.WE);
          {
            label = Widgets.newLabel(subSubComposite,"Server:");
            Widgets.layout(label,0,0,TableLayoutData.W);
            widgetServerName[0] = Widgets.newText(subSubComposite,SWT.LEFT|SWT.BORDER);
            if (Settings.Redmine.serverName != null) widgetServerName[0].setText(Settings.Redmine.serverName);
            widgetServerName[0].setLayoutData(new TableLayoutData(0,1,TableLayoutData.WE));

            label = Widgets.newLabel(subSubComposite,"Port:");
            Widgets.layout(label,0,2,TableLayoutData.W);
            widgetServerPort[0] = Widgets.newSpinner(subSubComposite,1,65535);
            widgetServerPort[0].setTextLimit(5);
            widgetServerPort[0].setSelection(Settings.Redmine.serverPort);
            Widgets.layout(widgetServerPort[0],0,3,TableLayoutData.WE);

            widgetServerUseSSL[0] = Widgets.newCheckbox(subSubComposite,"SSL");
            widgetServerUseSSL[0].setSelection(Settings.Redmine.serverUseSSL);
            Widgets.layout(widgetServerUseSSL[0],0,4,TableLayoutData.E);

            label = Widgets.newLabel(subSubComposite,"Name:");
            Widgets.layout(label,1,0,TableLayoutData.W);
            widgetLoginName[0] = Widgets.newText(subSubComposite);
            if (Settings.Redmine.loginName != null) widgetLoginName[0].setText(Settings.Redmine.loginName);
            Widgets.layout(widgetLoginName[0],1,1,TableLayoutData.WE,0,4);
          }
        }

        // --- JIRA ----------------------------------------------------

        separator = Widgets.newSeparator(composite,Backend.Types.JIRA.toString());
        separator.setLayout(new TableLayout(0.0,1.0));
        Widgets.layout(separator,2,0,TableLayoutData.WE);

        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0}));
        Widgets.layout(subComposite,3,0,TableLayoutData.WE);
        {
          widgetEnabled[1] = Widgets.newCheckbox(subComposite);
          widgetEnabled[1].setSelection(Settings.JIRA.enabled);
          Widgets.layout(widgetEnabled[1],0,0,TableLayoutData.NW);

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0,0.0,0.0,0.0}));
          Widgets.layout(subSubComposite,0,1,TableLayoutData.WE);
          {
            label = Widgets.newLabel(subSubComposite,"Server:");
            Widgets.layout(label,0,0,TableLayoutData.W);
            widgetServerName[1] = Widgets.newText(subSubComposite,SWT.LEFT|SWT.BORDER);
            if (Settings.JIRA.serverName != null) widgetServerName[1].setText(Settings.JIRA.serverName);
            widgetServerName[1].setLayoutData(new TableLayoutData(0,1,TableLayoutData.WE));

            label = Widgets.newLabel(subSubComposite,"Port:");
            Widgets.layout(label,0,2,TableLayoutData.W);

            widgetServerPort[1] = Widgets.newSpinner(subSubComposite,1,65535);
            widgetServerPort[1].setTextLimit(5);
            widgetServerPort[1].setSelection(Settings.JIRA.serverPort);
            Widgets.layout(widgetServerPort[1],0,3,TableLayoutData.WE);

            widgetServerUseSSL[1] = Widgets.newCheckbox(subSubComposite,"SSL");
            widgetServerUseSSL[1].setSelection(Settings.JIRA.serverUseSSL);
            Widgets.layout(widgetServerUseSSL[1],0,4,TableLayoutData.E);

            label = Widgets.newLabel(subSubComposite,"Name:");
            Widgets.layout(label,1,0,TableLayoutData.W);

            widgetLoginName[1] = Widgets.newText(subSubComposite);
            if (Settings.JIRA.loginName != null) widgetLoginName[1].setText(Settings.JIRA.loginName);
            Widgets.layout(widgetLoginName[1],1,1,TableLayoutData.WE,0,4);
          }
        }
      }
      Widgets.setNextFocus(widgetEnabled[0],
                           widgetServerName[0],
                           widgetServerPort[0],
                           widgetServerUseSSL[0],
                           widgetLoginName[0],
                           widgetEnabled[1],
                           widgetServerName[1],
                           widgetServerPort[1],
                           widgetServerUseSSL[1],
                           widgetLoginName[1]
                          );

      composite = Widgets.addTab(tabFolder,"Names");
      composite.setLayout(new TableLayout(new double[]{0.0,0.0,1.0,0.0,0.0,0.0,0.0,0.0},new double[]{0.0,1.0},2));
      Widgets.layout(composite,0,1,TableLayoutData.NSWE);
      {
        label = Widgets.newLabel(composite,"Status on update:");
        Widgets.layout(label,0,0,TableLayoutData.W);
        widgetIssueUpdateStatus = Widgets.newOptionMenu(composite);
        widgetIssueUpdateStatus.setToolTipText("Issue status to set on update of time entry.");
        Widgets.layout(widgetIssueUpdateStatus,0,1,TableLayoutData.WE);

        label = Widgets.newLabel(composite,"Status on completion:");
        Widgets.layout(label,1,0,TableLayoutData.W);
        widgetIssueCompletedStatus = Widgets.newOptionMenu(composite);
        widgetIssueCompletedStatus.setToolTipText("Issue status to set when 100% is done.");
        Widgets.layout(widgetIssueCompletedStatus,1,1,TableLayoutData.WE);

        label = Widgets.newLabel(composite,"Status active:");
        Widgets.layout(label,2,0,TableLayoutData.NW);
        widgetIssueActiveStatus = Widgets.newTable(composite,SWT.CHECK);
        widgetIssueActiveStatus.setToolTipText("Status for active issue.");
        widgetIssueActiveStatus.setHeaderVisible(false);
        Widgets.layout(widgetIssueActiveStatus,2,1,TableLayoutData.NSWE);

        label = Widgets.newLabel(composite,"New issue tracker:");
        Widgets.layout(label,3,0,TableLayoutData.NW);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0}));
        Widgets.layout(subComposite,3,1,TableLayoutData.WE);
        {
          label = Widgets.newLabel(subComposite,"Redmine");
          Widgets.layout(label,0,0,TableLayoutData.W);
          widgetNewIssueTrackerRedmine = Widgets.newOptionMenu(subComposite);
          widgetNewIssueTrackerRedmine.setToolTipText("Preset default tracker for new created Redmine issue.");
          Widgets.layout(widgetNewIssueTrackerRedmine,0,1,TableLayoutData.WE);

          label = Widgets.newLabel(subComposite,"JIRA");
          Widgets.layout(label,1,0,TableLayoutData.W);
          widgetNewIssueTrackerJIRA = Widgets.newOptionMenu(subComposite);
          widgetNewIssueTrackerJIRA.setToolTipText("Preset default tracker for new created JIRA issue.");
          Widgets.layout(widgetNewIssueTrackerJIRA,1,1,TableLayoutData.WE);
        }

        label = Widgets.newLabel(composite,"New issue status:");
        Widgets.layout(label,4,0,TableLayoutData.NW);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0}));
        Widgets.layout(subComposite,4,1,TableLayoutData.WE);
        {
          label = Widgets.newLabel(subComposite,"Redmine");
          Widgets.layout(label,0,0,TableLayoutData.W);
          widgetNewIssueStatusRedmine = Widgets.newOptionMenu(subComposite);
          widgetNewIssueStatusRedmine.setToolTipText("Preset default status for new created Redmine issue.");
          Widgets.layout(widgetNewIssueStatusRedmine,0,1,TableLayoutData.WE);

          label = Widgets.newLabel(subComposite,"JIRA");
          Widgets.layout(label,1,0,TableLayoutData.W);
          widgetNewIssueStatusJIRA = Widgets.newOptionMenu(subComposite);
          widgetNewIssueStatusJIRA.setToolTipText("Preset default status for new created JIRA issue.");
          Widgets.layout(widgetNewIssueStatusJIRA,1,1,TableLayoutData.WE);
        }
      }

      composite = Widgets.addTab(tabFolder,"Roles");
      composite.setLayout(new TableLayout(1.0,new double[]{0.0,1.0},2));
      Widgets.layout(composite,0,1,TableLayoutData.NSWE);
      {
        label = Widgets.newLabel(composite,"Manager roles:");
        Widgets.layout(label,0,0,TableLayoutData.NW);
        widgetManagerRoles = Widgets.newTable(composite,SWT.CHECK);
        widgetManagerRoles.setToolTipText("Manager roles.");
        Widgets.layout(widgetManagerRoles,0,1,TableLayoutData.NSWE);

        label = Widgets.newLabel(composite,"Developer roles:");
        Widgets.layout(label,1,0,TableLayoutData.NW);
        widgetDeveloperRoles = Widgets.newTable(composite,SWT.CHECK);
        widgetDeveloperRoles.setToolTipText("Developer roles.");
        Widgets.layout(widgetDeveloperRoles,1,1,TableLayoutData.NSWE);

        label = Widgets.newLabel(composite,"Designer roles:");
        Widgets.layout(label,2,0,TableLayoutData.NW);
        widgetDesignerRoles = Widgets.newTable(composite,SWT.CHECK);
        widgetDesignerRoles.setToolTipText("Designer roles.");
        Widgets.layout(widgetDesignerRoles,2,1,TableLayoutData.NSWE);
      }

      composite = Widgets.addTab(tabFolder,"Times");
      composite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,0.0},2));
      Widgets.layout(composite,0,1,TableLayoutData.NSWE);
      {
        label = Widgets.newLabel(composite,"Refresh (meta data):");
        Widgets.layout(label,0,0,TableLayoutData.W);
        widgetRefreshTime = Widgets.newOptionMenu(composite);
        widgetRefreshTime.setToolTipText("Time interval to auto refresh local database entries (roles, status, priorities...).\nSet to 0 to disable auto refresh.");
        Widgets.setOptionMenuItems(widgetRefreshTime,
                                   new Object[]{"disabled",   0,
                                                "30min",      30,
                                                "1h",       1*60,
                                                "2h",       2*60,
                                                "3h",       3*60,
                                                "6h",       6*60,
                                                "12h",     12*60
                                               }
                                  );
        Widgets.setSelectedOptionMenuItem(widgetRefreshTime,(Integer)Settings.refreshTime);
        Widgets.layout(widgetRefreshTime,0,1,TableLayoutData.W,0,0,0,0,100,SWT.DEFAULT);

        label = Widgets.newLabel(composite,"Refresh issues:");
        Widgets.layout(label,1,0,TableLayoutData.W);
        widgetRefreshIssuesTime = Widgets.newOptionMenu(composite);
        widgetRefreshIssuesTime.setToolTipText("Time interval to auto refresh local database issues.\nSet to 0 to disable auto refresh.");
        Widgets.setOptionMenuItems(widgetRefreshIssuesTime,
                                   new Object[]{"disabled",    0,
                                                "30min",      30,
                                                "1h",       1*60,
                                                "2h",       2*60,
                                                "3h",       3*60,
                                                "6h",       6*60,
                                                "12h",     12*60
                                               }
                                  );
        Widgets.setSelectedOptionMenuItem(widgetRefreshIssuesTime,(Integer)Settings.refreshIssuesTime);
        Widgets.layout(widgetRefreshIssuesTime,1,1,TableLayoutData.W,0,0,0,0,100,SWT.DEFAULT);

        label = Widgets.newLabel(composite,"Incremental refresh issues:");
        Widgets.layout(label,2,0,TableLayoutData.W);
        widgetIncrementalRefreshIssuesTime = Widgets.newOptionMenu(composite);
        widgetIncrementalRefreshIssuesTime.setToolTipText("Time interval for incremental auto refresh local database issues.\nSet to 0 to disable auto refresh.");
        Widgets.setOptionMenuItems(widgetIncrementalRefreshIssuesTime,
                                   new Object[]{"30s",     30,
                                                "1min",  1*60,
                                                "2min",  2*60,
                                                "5min",  5*60,
                                                "10min",10*60,
                                                "30min",30*60,
                                                "1h",   60*60
                                               }
                                  );
        Widgets.setSelectedOptionMenuItem(widgetIncrementalRefreshIssuesTime,(Integer)Settings.incrementalRefreshIssuesTime);
        Widgets.layout(widgetIncrementalRefreshIssuesTime,2,1,TableLayoutData.W,0,0,0,0,100,SWT.DEFAULT);

        label = Widgets.newLabel(composite,"Refresh time entries:");
        Widgets.layout(label,3,0,TableLayoutData.W);
        widgetRefreshTimeEntriesTime = Widgets.newOptionMenu(composite);
        widgetRefreshTimeEntriesTime.setToolTipText("Time interval to auto refresh local database time entries.\nSet to 0 to disable auto refresh.");
        Widgets.setOptionMenuItems(widgetRefreshTimeEntriesTime,
                                   new Object[]{"disabled",    0,
                                                "30min",      30,
                                                "1h",       1*60,
                                                "2h",       2*60,
                                                "3h",       3*60,
                                                "6h",       6*60,
                                                "12h",     12*60
                                               }
                                  );
        Widgets.setSelectedOptionMenuItem(widgetRefreshTimeEntriesTime,(Integer)Settings.refreshTimeEntriesTime);
        Widgets.layout(widgetRefreshTimeEntriesTime,3,1,TableLayoutData.W,0,0,0,0,100,SWT.DEFAULT);

        label = Widgets.newLabel(composite,"Incremental refresh time entries:");
        Widgets.layout(label,4,0,TableLayoutData.W);
        widgetIncrementalRefreshTimeEntriesTime = Widgets.newOptionMenu(composite);
        widgetIncrementalRefreshTimeEntriesTime.setToolTipText("Time interval for incremental auto refresh local database time entries.\nSet to 0 to disable auto refresh.");
        Widgets.setOptionMenuItems(widgetIncrementalRefreshTimeEntriesTime,
                                   new Object[]{"30s",     30,
                                                "1min",  1*60,
                                                "2min",  2*60,
                                                "5min",  5*60,
                                                "10min",10*60,
                                                "30min",30*60,
                                                "1h",   60*60
                                               }
                                  );
        Widgets.setSelectedOptionMenuItem(widgetIncrementalRefreshTimeEntriesTime,(Integer)Settings.incrementalRefreshTimeEntriesTime);
        Widgets.layout(widgetIncrementalRefreshTimeEntriesTime,4,1,TableLayoutData.W,0,0,0,0,100,SWT.DEFAULT);

        label = Widgets.newLabel(composite,"Max. refresh issues history:");
        Widgets.layout(label,5,0,TableLayoutData.W);
        widgetMaxIssuesHistory = Widgets.newOptionMenu(composite);
        widgetMaxIssuesHistory.setToolTipText("Number of days to keep time entries history.");
        Widgets.setOptionMenuItems(widgetMaxIssuesHistory,
                                   new Object[]{"7days",     7,
                                                "14days",   14,
                                                "30days",   30,
                                                "60days",   60,
                                                "90days",   90,
                                                "unlimited", 0
                                               }
                                  );
        Widgets.setSelectedOptionMenuItem(widgetMaxIssuesHistory,(Integer)Settings.maxTimeEntriesHistory);
        Widgets.layout(widgetMaxIssuesHistory,5,1,TableLayoutData.W,0,0,0,0,100,SWT.DEFAULT);

        label = Widgets.newLabel(composite,"Max. refresh time entries history:");
        Widgets.layout(label,6,0,TableLayoutData.W);
        widgetMaxTimeEntriesHistory = Widgets.newOptionMenu(composite);
        widgetMaxTimeEntriesHistory.setToolTipText("Number of days to keep time entries history.");
        Widgets.setOptionMenuItems(widgetMaxTimeEntriesHistory,
                                   new Object[]{"7days",     7,
                                                "14days",   14,
                                                "30days",   30,
                                                "60days",   60,
                                                "90days",   90,
                                                "unlimited", 0
                                               }
                                  );
        Widgets.setSelectedOptionMenuItem(widgetMaxTimeEntriesHistory,(Integer)Settings.maxTimeEntriesHistory);
        Widgets.layout(widgetMaxTimeEntriesHistory,6,1,TableLayoutData.W,0,0,0,0,100,SWT.DEFAULT);

        label = Widgets.newLabel(composite,"Sync:");
        Widgets.layout(label,7,0,TableLayoutData.W);
        widgetSyncTime = Widgets.newOptionMenu(composite);
        widgetSyncTime.setToolTipText("Time interval to synchronize local database with server.");
        Widgets.setOptionMenuItems(widgetSyncTime,
                                   new Object[]{"30s",     30,
                                                "1min",  1*60,
                                                "2min",  2*60,
                                                "5min",  5*60,
                                                "10min",10*60,
                                                "30min",30*60,
                                                "1h",   60*60
                                               }
                                  );
        Widgets.setSelectedOptionMenuItem(widgetSyncTime,(Integer)Settings.syncTime);
        Widgets.layout(widgetSyncTime,7,1,TableLayoutData.W,0,0,0,0,100,SWT.DEFAULT);

        label = Widgets.newLabel(composite,"Throttle:");
        Widgets.layout(label,8,0,TableLayoutData.W);
        widgetThrottleTime = Widgets.newOptionMenu(composite);
        widgetThrottleTime.setToolTipText("Delay time between requests to server.");
        Widgets.setOptionMenuItems(widgetThrottleTime,
                                   new Object[]{"disabled",    0,
                                                "50ms",       50,
                                                "100ms",     100,
                                                "250ms",     250,
                                                "500ms",     500,
                                                "1s",       1000,
                                                "2s",       2000,
                                                "5s",       5000,
                                                "10s",     10000
                                               }
                                  );
        Widgets.setSelectedOptionMenuItem(widgetThrottleTime,(Integer)Settings.throttleTime);
        Widgets.layout(widgetThrottleTime,8,1,TableLayoutData.W,0,0,0,0,100,SWT.DEFAULT);

        widgetRefreshAtStart = Widgets.newCheckbox(composite,"Refresh at start");
        widgetRefreshAtStart.setToolTipText("Start refresh local database at start.");
        widgetRefreshAtStart.setSelection(Settings.refreshAtStart);
        Widgets.layout(widgetRefreshAtStart,9,0,TableLayoutData.W);
        widgetRefreshAtStart.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button widget = (Button)selectionEvent.widget;

            if (!Dialogs.confirm(dialog,"Enable refresh of local database cause a heavy server load and is not recommented.\nInstead use incremental refresh or call refresh on demand in the menu.\n\nReally do always a refresh at start?"))
            {
              widget.setSelection(false);
            }
          }
        });

        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0,0.0,1.0}));
        Widgets.layout(subComposite,10,0,TableLayoutData.WE);
        {
          widgetAutoRefresh = Widgets.newCheckbox(subComposite,"Auto refresh");
          widgetAutoRefresh.setToolTipText("Auto refresh time window");
          widgetAutoRefresh.setSelection(Settings.autoRefresh);
          Widgets.layout(widgetAutoRefresh,0,0,TableLayoutData.W);

          widgetAutoRefreshStartTime = Widgets.newTime(subComposite,SWT.SHORT);
          widgetAutoRefreshStartTime.setEnabled(Settings.autoRefresh);
          widgetAutoRefreshStartTime.setToolTipText("Spent-on date.");
          widgetAutoRefreshStartTime.setHours(Settings.autoRefreshStartTime.getHours());
          widgetAutoRefreshStartTime.setMinutes(Settings.autoRefreshStartTime.getMinutes());
          Widgets.layout(widgetAutoRefreshStartTime,0,1,TableLayoutData.W);

          label = Widgets.newLabel(subComposite,"..");
          Widgets.layout(label,0,2,TableLayoutData.W);

          widgetAutoRefreshEndTime = Widgets.newTime(subComposite,SWT.SHORT);
          widgetAutoRefreshEndTime.setEnabled(Settings.autoRefresh);
          widgetAutoRefreshEndTime.setToolTipText("Spent-on date.");
          widgetAutoRefreshEndTime.setHours(Settings.autoRefreshEndTime.getHours());
          widgetAutoRefreshEndTime.setMinutes(Settings.autoRefreshEndTime.getMinutes());
          Widgets.layout(widgetAutoRefreshEndTime,0,3,TableLayoutData.W);
        }
      }

      composite = Widgets.addTab(tabFolder,"Vacation");
      composite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0},2));
      Widgets.layout(composite,0,1,TableLayoutData.NSWE);
      {
        label = Widgets.newLabel(composite,"Project:");
        Widgets.layout(label,0,0,TableLayoutData.W);
        widgetVacationProjects = new TreeCombo(composite,SWT.LEFT,SWT.NONE,100,TreeCombo.DEFAULT_HEIGHT);
        Widgets.layout(widgetVacationProjects,0,1,TableLayoutData.WE);

        label = Widgets.newLabel(composite,"Issue:");
        Widgets.layout(label,1,0,TableLayoutData.W);
        widgetVacationIssueSummaries = new TreeCombo(composite,SWT.LEFT,SWT.NONE,100,TreeCombo.DEFAULT_HEIGHT);
        widgetVacationIssueSummaries.addTreeColumn("#",     SWT.RIGHT, 80,false);
        widgetVacationIssueSummaries.addTreeColumn("Issue", SWT.LEFT,  0, true );
        widgetVacationIssueSummaries.setLayout(new double[]{0.0,1.0});
        Widgets.layout(widgetVacationIssueSummaries,1,1,TableLayoutData.WE);

        widgetVacationActivitiesTitle = Widgets.newLabel(composite,"Activity:");
        Widgets.layout(widgetVacationActivitiesTitle,2,0,TableLayoutData.W);
        widgetVacationActivities = Widgets.newOptionMenu(composite);
        Widgets.layout(widgetVacationActivities,2,1,TableLayoutData.WE);

        label = Widgets.newLabel(composite,"Comment:");
        Widgets.layout(label,3,0,TableLayoutData.W);
        widgetVacationComment = Widgets.newText(composite);
        widgetVacationComment.setToolTipText("Comment for vacation entry (can be empty).");
        widgetVacationComment.setText(Settings.vacationComment);
        Widgets.layout(widgetVacationComment,3,1,TableLayoutData.WE);
      }

      composite = Widgets.addTab(tabFolder,"Colors");
      composite.setLayout(new TableLayout(1.0,1.0,2));
      Widgets.layout(composite,0,2,TableLayoutData.NSWE);
      {
        widgetColors = Widgets.newTable(composite);
        widgetColors.setToolTipText("Colors list.");
        Widgets.layout(widgetColors,0,0,TableLayoutData.NSWE);
        Widgets.addTableColumn(widgetColors,0,"Name", SWT.LEFT,250,true);
        Widgets.addTableColumn(widgetColors,1,"Foreground",SWT.LEFT,100,true);
        Widgets.addTableColumn(widgetColors,2,"Background",SWT.LEFT,100,true);
        widgetColors.addMouseListener(new MouseListener()
        {
          @Override
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
            Table widget = (Table)mouseEvent.widget;

            int index = widget.getSelectionIndex();
            if (index >= 0)
            {
              TableItem      tableItem = widget.getItem(index);
              String         name      = tableItem.getText(0);
              Settings.Color color     = (Settings.Color)tableItem.getData();

              if (editColor(dialog,name,color))
              {
                Widgets.setTableItemColor(widgetColors,color,1,(color.foreground != null) ? new Color(null,color.foreground) : null);
                Widgets.setTableItemColor(widgetColors,color,2,(color.background != null) ? new Color(null,color.background) : null);
                widgetColors.deselectAll();
              }
            }
          }
          @Override
          public void mouseDown(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseUp(MouseEvent mouseEvent)
          {
          }
        });
        addColors(widgetColors);
      }

      composite = Widgets.addTab(tabFolder,"Action filters");
      composite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0,2));
      Widgets.layout(composite,0,2,TableLayoutData.NSWE);
      {
        widgetActionFilters = Widgets.newTable(composite);
        widgetActionFilters.setToolTipText("Action filter list.");
        Widgets.layout(widgetActionFilters,0,0,TableLayoutData.NSWE);
        widgetActionFilters.addMouseListener(new MouseListener()
        {
          @Override
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
            Table widget = (Table)mouseEvent.widget;

            int index = widget.getSelectionIndex();
            if (index >= 0)
            {
              String actionFilter = widget.getItem(index).getText(0);

              actionFilter = Dialogs.string(shell,"Edit action filter","Filter:",actionFilter,"Save","Cancel");
              if (actionFilter != null)
              {
                widget.getItem(index).setText(0,actionFilter);
              }
            }
          }
          @Override
          public void mouseDown(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseUp(MouseEvent mouseEvent)
          {
          }
        });
        for (String actionFilter : Settings.actionFilters)
        {
          Widgets.addTableItem(widgetActionFilters,
                               (Object)actionFilter,
                               actionFilter
                              );
        }

        // buttons
        subComposite = Widgets.newComposite(composite,SWT.NONE,4);
        Widgets.layout(subComposite,1,0,TableLayoutData.W);
        {
          button = Widgets.newButton(subComposite,"Add"+"\u2026");
          button.setToolTipText("Add action filter.");
          Widgets.layout(button,0,0,TableLayoutData.DEFAULT,0,0,0,0,110,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              String actionFilter = Dialogs.string(shell,"Add action filter","Filter:","","Add","Cancel");
              if (actionFilter != null)
              {
                Widgets.addTableItem(widgetActionFilters,
                                     (Object)actionFilter,
                                     actionFilter
                                    );
              }
            }
          });

          button = Widgets.newButton(subComposite,"Edit"+"\u2026");
          button.setToolTipText("Edit action filter.");
          Widgets.layout(button,0,1,TableLayoutData.DEFAULT,0,0,0,0,110,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              int index = widgetActionFilters.getSelectionIndex();
              if (index >= 0)
              {
                String actionFilter = widgetActionFilters.getItem(index).getText(0);

                actionFilter = Dialogs.string(shell,"Edit action filter","Filter:",actionFilter,"Save","Cancel");
                if (actionFilter != null)
                {
                  widgetActionFilters.getItem(index).setText(0,actionFilter);
                }
              }
            }
          });

          button = Widgets.newButton(subComposite,"Clone"+"\u2026");
          button.setToolTipText("Clone action filter.");
          Widgets.layout(button,0,2,TableLayoutData.DEFAULT,0,0,0,0,110,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              int index = widgetActionFilters.getSelectionIndex();
              if (index >= 0)
              {
                String actionFilter = widgetActionFilters.getItem(index).getText(0);

                actionFilter = Dialogs.string(shell,"Edit action filter","Filter:",actionFilter);
                if (actionFilter != null)
                {
                  Widgets.addTableItem(widgetActionFilters,
                                       (Object)actionFilter,
                                       actionFilter
                                      );
                }
              }
            }
          });

          button = Widgets.newButton(subComposite,"Remove"+"\u2026");
          button.setToolTipText("Remove action filter.");
          Widgets.layout(button,0,3,TableLayoutData.DEFAULT,0,0,0,0,110,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              int index = widgetActionFilters.getSelectionIndex();
              if (index >= 0)
              {
                String actionFilter = widgetActionFilters.getItem(index).getText(0);

                if (Dialogs.confirm(shell,"Remove action filter","Remove action filter '"+actionFilter+"'?","Remove","Cancel"))
                {
                  Widgets.removeTableItem(widgetActionFilters,index);
                }
              }
            }
          });
        }
      }
//TODO?
      ActivityList activityList = null;
      try
      {
        activityList = Backend.getActivityList();
      }
      catch (BackendException exception)
      {
        // ignored
      }
      Timemine.updateActivities(widgetVacationActivitiesTitle,widgetVacationActivities,activityList,Settings.vacationActivity);

      composite = Widgets.addTab(tabFolder,"Misc");
      composite.setLayout(new TableLayout(new double[]{0.0,0.0,0.0,0.0,0.0,1.0,0.0},new double[]{0.0,1.0},2));
      Widgets.layout(composite,0,3,TableLayoutData.NSWE);
      {
        label = Widgets.newLabel(composite,"Date format:");
        Widgets.layout(label,0,0,TableLayoutData.W);
        widgetDateFormat = Widgets.newText(composite);
        widgetDateFormat.setToolTipText("Date format.\nPatterns:\n  y - year digit\n  M - month digit\n  d - day digit\n  E - week day name");
        widgetDateFormat.setText(Settings.dateFormat);
        Widgets.layout(widgetDateFormat,0,1,TableLayoutData.WE);

        label = Widgets.newLabel(composite,"Min. time:");
        Widgets.layout(label,1,0,TableLayoutData.W);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(0.0,0.0));
        Widgets.layout(subComposite,1,1,TableLayoutData.WE);
        {
          widgetMinTimeDelta = Widgets.newSpinner(subComposite,0);
          widgetMinTimeDelta.setToolTipText("Min. time delta.");
          widgetMinTimeDelta.setTextLimit(2);
          widgetMinTimeDelta.setMinimum(5);
          widgetMinTimeDelta.setIncrement(5);
          widgetMinTimeDelta.setSelection((int)Settings.minTimeDelta);
          Widgets.layout(widgetMinTimeDelta,0,0,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
          label = Widgets.newLabel(subComposite,"min");
          Widgets.layout(label,0,1,TableLayoutData.W);
        }

        label = Widgets.newLabel(composite,"Working hours:");
        Widgets.layout(label,2,0,TableLayoutData.NW);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(0.0,0.0));
        Widgets.layout(subComposite,2,1,TableLayoutData.WE);
        {
          widgetWorkingHoursPerDay = Widgets.newSpinner(subComposite,0);
          widgetWorkingHoursPerDay.setToolTipText("Required working hours per day.");
          widgetWorkingHoursPerDay.setTextLimit(2);
          widgetWorkingHoursPerDay.setIncrement(1);
          widgetWorkingHoursPerDay.setSelection((int)Settings.workingHoursPerDay);
          Widgets.layout(widgetWorkingHoursPerDay,0,0,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
          label = Widgets.newLabel(subComposite,"h/day");
          Widgets.layout(label,0,1,TableLayoutData.W);

          widgetWorkingHoursPerWeek = Widgets.newSpinner(subComposite,0);
          widgetWorkingHoursPerWeek.setToolTipText("Required working hours per day.");
          widgetWorkingHoursPerWeek.setTextLimit(2);
          widgetWorkingHoursPerWeek.setIncrement(1);
          widgetWorkingHoursPerWeek.setSelection((int)Settings.workingHoursPerWeek);
          Widgets.layout(widgetWorkingHoursPerWeek,1,0,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
          label = Widgets.newLabel(subComposite,"h/week");
          Widgets.layout(label,1,1,TableLayoutData.W);
        }

        label = Widgets.newLabel(composite,"Pause minutes:");
        Widgets.layout(label,3,0,TableLayoutData.W);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(0.0,0.0));
        Widgets.layout(subComposite,3,1,TableLayoutData.WE);
        {
          widgetPauseMinutesPerDay = Widgets.newSpinner(subComposite,0);
          widgetPauseMinutesPerDay.setToolTipText("Required pause minutes per day.");
          widgetPauseMinutesPerDay.setTextLimit(2);
          widgetPauseMinutesPerDay.setIncrement(15);
          widgetPauseMinutesPerDay.setSelection(Settings.pauseMinutesPerDay);
          Widgets.layout(widgetPauseMinutesPerDay,0,0,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
          label = Widgets.newLabel(subComposite,"min/day");
          Widgets.layout(label,0,1,TableLayoutData.W);
        }

        label = Widgets.newLabel(composite,"Time ranges:");
        Widgets.layout(label,4,0,TableLayoutData.NW);
        widgetSnapTimeRanges = Widgets.newCheckbox(composite,"snap");
        widgetSnapTimeRanges.setSelection(Settings.snapTimeRanges);
        Widgets.layout(widgetSnapTimeRanges,4,1,TableLayoutData.W);

        label = Widgets.newLabel(composite,"Show flags:");
        Widgets.layout(label,5,0,TableLayoutData.NW);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(1.0,1.0));
        Widgets.layout(subComposite,5,1,TableLayoutData.NSWE);
        {
          widgetShowFlags = Widgets.newTable(subComposite,SWT.CHECK);
          widgetShowFlags.setToolTipText("Show dialogs flags.");
          Widgets.layout(widgetShowFlags,0,0,TableLayoutData.NSWE);
          Widgets.addTableColumn(widgetShowFlags,0,"",SWT.LEFT,500,true);
          widgetShowFlags.setLinesVisible(false);
          widgetShowFlags.setHeaderVisible(false);
          widgetShowFlags.addListener(SWT.Resize, new Listener()
          {
            @Override
            public void handleEvent(Event event)
            {
              Table       table = (Table)event.widget;
              TableColumn tableColumn = table.getColumn(0);

              tableColumn.setWidth(table.getClientArea().width);
            }
          });

          Widgets.addTableItem(widgetShowFlags,Settings.enterIssueProgress,          "enter issue progress"       ).setChecked(Settings.enterIssueProgress          );
          Widgets.addTableItem(widgetShowFlags,Settings.showSplashScreen,            "splash screen"              ).setChecked(Settings.showSplashScreen            );
          Widgets.addTableItem(widgetShowFlags,Settings.showUpdateStatusErrors,      "update status errors"       ).setChecked(Settings.showUpdateStatusErrors      );
          Widgets.addTableItem(widgetShowFlags,Settings.showUpdateIssuesWarning,     "update issues warning"      ).setChecked(Settings.showUpdateIssuesWarning     );
          Widgets.addTableItem(widgetShowFlags,Settings.showUpdateTimeEntriesWarning,"update time entries warning").setChecked(Settings.showUpdateTimeEntriesWarning);
          Widgets.addTableItem(widgetShowFlags,Settings.showFutureTimeEntryWarning,  "future time entry warning"  ).setChecked(Settings.showFutureTimeEntryWarning  );
          Widgets.addTableItem(widgetShowFlags,Settings.showNewVersionInformation,   "new version information"    ).setChecked(Settings.showNewVersionInformation   );
        }

        label = Widgets.newLabel(composite,"Miscellaneous:");
        Widgets.layout(label,6,0,TableLayoutData.W);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(0.0,0.0));
        Widgets.layout(subComposite,6,1,TableLayoutData.WE);
        {
//TODO cache expire time
//TODO requiredHoursPerDay
        }
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetButtonSave = Widgets.newButton(composite,"Save");
      Widgets.layout(widgetButtonSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetButtonSave.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Settings.Redmine.enabled                   = widgetEnabled[0].getSelection();
          Settings.Redmine.serverName                = widgetServerName[0].getText();
          Settings.Redmine.serverPort                = widgetServerPort[0].getSelection();
          Settings.Redmine.serverUseSSL              = widgetServerUseSSL[0].getSelection();
          Settings.Redmine.loginName                 = widgetLoginName[0].getText();
          Settings.JIRA.enabled                      = widgetEnabled[1].getSelection();
          Settings.JIRA.serverName                   = widgetServerName[1].getText();
          Settings.JIRA.serverPort                   = widgetServerPort[1].getSelection();
          Settings.JIRA.serverUseSSL                 = widgetServerUseSSL[1].getSelection();
          Settings.JIRA.loginName                    = widgetLoginName[1].getText();

          Settings.issueUpdateStatus                 = Widgets.getSelectedOptionMenuItem(widgetIssueUpdateStatus,"");
          Settings.issueCompletedStatus              = Widgets.getSelectedOptionMenuItem(widgetIssueCompletedStatus,"");
          Settings.issueActiveStatusSet.clear();
          for (TableItem tableItem : widgetIssueActiveStatus.getItems())
          {
            if (tableItem.getChecked())
            {
              Settings.issueActiveStatusSet.add(tableItem.getText(0));
            }
          }

          Settings.Redmine.newIssueTrackerName       = Widgets.getSelectedOptionMenuItem(widgetNewIssueTrackerRedmine,"");
          Settings.JIRA.newIssueTrackerName          = Widgets.getSelectedOptionMenuItem(widgetNewIssueTrackerJIRA,"");
          Settings.Redmine.newIssueStatusName        = Widgets.getSelectedOptionMenuItem(widgetNewIssueStatusRedmine,"");
          Settings.JIRA.newIssueStatusName           = Widgets.getSelectedOptionMenuItem(widgetNewIssueStatusJIRA,"");

          Settings.managerRoles.clear();
          for (TableItem tableItem : widgetManagerRoles.getItems())
          {
            if (tableItem.getChecked()) Settings.managerRoles.add(tableItem.getText(0));
          }
          Settings.developerRoles.clear();
          for (TableItem tableItem : widgetDeveloperRoles.getItems())
          {
            if (tableItem.getChecked()) Settings.developerRoles.add(tableItem.getText(0));
          }
          Settings.designerRoles.clear();
          for (TableItem tableItem : widgetDesignerRoles.getItems())
          {
            if (tableItem.getChecked()) Settings.designerRoles.add(tableItem.getText(0));
          }

          Project vacationProject = widgetVacationProjects.getSelection(Project.class);
          Issue   vacationIssue   = widgetVacationIssueSummaries.getSelection(Issue.class);

          Settings.vacationProject                   = (vacationProject != null) ? vacationProject.name : "";
          Settings.vacationIssue                     = (vacationIssue != null) ? vacationIssue.summary : "";
          Settings.vacationActivity                  = widgetVacationActivities.getText();
          Settings.vacationComment                   = widgetVacationComment.getText();

          Settings.refreshTime                       = Widgets.getSelectedOptionMenuItem(widgetRefreshTime,60);
          Settings.refreshIssuesTime                 = Widgets.getSelectedOptionMenuItem(widgetRefreshIssuesTime,60);
          Settings.incrementalRefreshIssuesTime      = Widgets.getSelectedOptionMenuItem(widgetIncrementalRefreshIssuesTime,60);
          Settings.refreshTimeEntriesTime            = Widgets.getSelectedOptionMenuItem(widgetRefreshTimeEntriesTime,60);
          Settings.incrementalRefreshTimeEntriesTime = Widgets.getSelectedOptionMenuItem(widgetIncrementalRefreshTimeEntriesTime,60);
          Settings.maxIssuesHistory                  = Widgets.getSelectedOptionMenuItem(widgetMaxIssuesHistory,60);
          Settings.maxTimeEntriesHistory             = Widgets.getSelectedOptionMenuItem(widgetMaxTimeEntriesHistory,60);
          Settings.syncTime                          = Widgets.getSelectedOptionMenuItem(widgetSyncTime,60);
          Settings.throttleTime                      = Widgets.getSelectedOptionMenuItem(widgetThrottleTime,1000);
          Settings.refreshAtStart                    = widgetRefreshAtStart.getSelection();
          Settings.autoRefresh                       = widgetAutoRefresh.getSelection();
          Settings.autoRefreshStartTime.setHours(widgetAutoRefreshStartTime.getHours());
          Settings.autoRefreshStartTime.setMinutes(widgetAutoRefreshStartTime.getMinutes());
          Settings.autoRefreshEndTime.setHours(widgetAutoRefreshEndTime.getHours());
          Settings.autoRefreshEndTime.setMinutes(widgetAutoRefreshEndTime.getMinutes());

          saveColors(widgetColors);

          Settings.actionFilters.clear();
          for (TableItem tableItem : widgetActionFilters.getItems())
          {
            Settings.actionFilters.add(tableItem.getText(0));
          }

          Settings.dateFormat                        = widgetDateFormat.getText().trim();
          Settings.minTimeDelta                      = widgetMinTimeDelta.getSelection();
          Settings.workingHoursPerDay                = (double)widgetWorkingHoursPerDay.getSelection();
          Settings.workingHoursPerWeek               = (double)widgetWorkingHoursPerWeek.getSelection();
          Settings.pauseMinutesPerDay                = widgetPauseMinutesPerDay.getSelection();
          Settings.snapTimeRanges                    = widgetSnapTimeRanges.getSelection();
          for (TableItem tableItem : widgetShowFlags.getItems())
          {
            Object data = tableItem.getData();
            if (data == Settings.showSplashScreen            ) Settings.showSplashScreen             = tableItem.getChecked();
            if (data == Settings.showUpdateStatusErrors      ) Settings.showUpdateStatusErrors       = tableItem.getChecked();
            if (data == Settings.showUpdateIssuesWarning     ) Settings.showUpdateIssuesWarning      = tableItem.getChecked();
            if (data == Settings.showUpdateTimeEntriesWarning) Settings.showUpdateTimeEntriesWarning = tableItem.getChecked();
            if (data == Settings.showFutureTimeEntryWarning  ) Settings.showFutureTimeEntryWarning   = tableItem.getChecked();
          }

          Settings.geometryPreferences = dialog.getSize();

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    widgetAutoRefresh.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Button widget = (Button)selectionEvent.widget;

        widgetAutoRefreshStartTime.setEnabled(widget.getSelection());
        widgetAutoRefreshEndTime.setEnabled(widget.getSelection());
      }
    });
    widgetVacationProjects.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        TreeCombo widget = (TreeCombo)selectionEvent.widget;

        Project vacationProject = widget.getSelection(Project.class);
        if (vacationProject != null)
        {
          Timemine.setSelectedProject(widgetVacationProjects,
                                      widgetVacationIssueSummaries,
                                      widgetVacationActivitiesTitle,
                                      widgetVacationActivities,
                                      vacationProject
                                     );
          Timemine.updateActivities(widgetVacationActivitiesTitle,
                                    widgetVacationActivities,
                                    vacationProject,
                                    null  // activity
                                   );
        }
      }
    });
    widgetVacationIssueSummaries.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        TreeCombo widget = (TreeCombo)selectionEvent.widget;

        Issue vacationIssue = widget.getSelection(Issue.class);
        if (vacationIssue != null)
        {
          Project  vacationProject  = vacationIssue.getProject();
          Activity vacationActivity = vacationIssue.getActivity(Settings.vacationActivity);

          Timemine.setSelectedIssue(widgetVacationIssueSummaries,
                                    vacationIssue
                                   );
          Timemine.updateActivities(widgetVacationActivitiesTitle,
                                    widgetVacationActivities,
                                    vacationProject,
                                    vacationActivity
                                   );
        }
      }
    });

    // show dialog
    Dialogs.show(dialog,Settings.geometryPreferences);

    // set data
    try
    {
      Widgets.addOptionMenuItem(widgetIssueUpdateStatus,"");
      for (Status status : Backend.getStatusList())
      {
        Widgets.addOptionMenuItem(widgetIssueUpdateStatus,status.name);
      }
      Widgets.setSelectedOptionMenuItem(widgetIssueUpdateStatus,Settings.issueUpdateStatus);

      Widgets.addOptionMenuItem(widgetIssueCompletedStatus,"");
      for (Status status : Backend.getStatusList())
      {
        Widgets.addOptionMenuItem(widgetIssueCompletedStatus,status.name);
      }
      Widgets.setSelectedOptionMenuItem(widgetIssueCompletedStatus,Settings.issueCompletedStatus);

      for (Status status : Backend.getStatusList())
      {
        TableItem tableItem = Widgets.addTableItem(widgetIssueActiveStatus,
                                                   (Object)status,
                                                   status.name
                                                  );
        tableItem.setChecked(Settings.issueActiveStatusSet.contains(status.name));
      }

//TODO
      Widgets.setSelectedOptionMenuItem(widgetNewIssueTrackerRedmine,Settings.Redmine.newIssueTrackerName);
      Widgets.addOptionMenuItem(widgetNewIssueTrackerRedmine,"");
      for (Tracker tracker : Backend.getTrackerList(Backend.TypeSet.REDMINE))
      {
        Widgets.addOptionMenuItem(widgetNewIssueTrackerRedmine,tracker.name);
      }
      Widgets.setSelectedOptionMenuItem(widgetNewIssueTrackerRedmine,Settings.Redmine.newIssueTrackerName);

      Widgets.setSelectedOptionMenuItem(widgetNewIssueTrackerJIRA,Settings.JIRA.newIssueTrackerName);
      Widgets.addOptionMenuItem(widgetNewIssueTrackerJIRA,"");
      for (Tracker tracker : Backend.getTrackerList(Backend.TypeSet.JIRA))
      {
        Widgets.addOptionMenuItem(widgetNewIssueTrackerJIRA,tracker.name);
      }
      Widgets.setSelectedOptionMenuItem(widgetNewIssueTrackerJIRA,Settings.JIRA.newIssueTrackerName);

//TODO
      Widgets.setSelectedOptionMenuItem(widgetNewIssueStatusRedmine,Settings.Redmine.newIssueStatusName);
      Widgets.addOptionMenuItem(widgetNewIssueStatusRedmine,"");
      for (Status status : Backend.getStatusList(Backend.TypeSet.REDMINE))
      {
        Widgets.addOptionMenuItem(widgetNewIssueStatusRedmine,status.name);
      }
      Widgets.setSelectedOptionMenuItem(widgetNewIssueStatusRedmine,Settings.Redmine.newIssueStatusName);

      Widgets.setSelectedOptionMenuItem(widgetNewIssueStatusJIRA,Settings.JIRA.newIssueStatusName);
      Widgets.addOptionMenuItem(widgetNewIssueStatusJIRA,"");
      for (Status status : Backend.getStatusList(Backend.TypeSet.JIRA))
      {
        Widgets.addOptionMenuItem(widgetNewIssueStatusJIRA,status.name);
      }
      Widgets.setSelectedOptionMenuItem(widgetNewIssueStatusJIRA,Settings.JIRA.newIssueStatusName);

      for (String roleName : Backend.getRoleNames())
      {
        TableItem tableItem = Widgets.addTableItem(widgetManagerRoles,roleName,roleName);
        tableItem.setChecked(Settings.managerRoles.contains(roleName));
      }
      for (String roleName : Backend.getRoleNames())
      {
        TableItem tableItem = Widgets.addTableItem(widgetDeveloperRoles,roleName,roleName);
        tableItem.setChecked(Settings.developerRoles.contains(roleName));
      }
      for (String roleName : Backend.getRoleNames())
      {
        TableItem tableItem = Widgets.addTableItem(widgetDesignerRoles,roleName,roleName);
        tableItem.setChecked(Settings.designerRoles.contains(roleName));
      }

//TODO: with or without type
      Project vacationProject = Backend.getProject(Backend.TypeSet.ALL,Settings.vacationProject);
      Timemine.updateProjects(widgetVacationProjects,
                              vacationProject,
                              true
                             );
      if (vacationProject != null)
      {
        Issue vacationIssue = vacationProject.getIssue(Settings.vacationIssue);
        Timemine.updateIssues(widgetVacationIssueSummaries,
                              vacationProject,
                              vacationIssue
                             );
        if (vacationIssue != null)
        {
          Activity vacationActivity = vacationIssue.getActivity(Settings.vacationActivity);
          Timemine.updateActivities(widgetVacationActivitiesTitle,
                                    widgetVacationActivities,
                                    vacationProject,
                                    vacationActivity
                                   );
        }
      }
    }
    catch (BackendException exception)
    {
      Timemine.printStackTrace(exception);
      Dialogs.error(dialog,"Cannot get status list! (error: "+exception.getMessage()+")");
    }

    // run dialog
    if ((Boolean)Dialogs.run(dialog,false))
    {
      boolean saveSettings = true;
      if (Settings.isExternalModified())
      {
        saveSettings = Dialogs.confirm(shell,
                                       "Confirmation",
                                       "Settings were modified externally.\nOverwrite settings?",
                                       "Overwrite",
                                       "Cancel",
                                       false
                                      );
      }
      if (saveSettings)
      {
        Settings.save();
      }

      synchronized(Settings.showRestartAfterConfigChanged)
      {
        if (Settings.showRestartAfterConfigChanged)
        {
// NYI ??? return showRestartAfterConfigChanged flag?
          if (Dialogs.confirm(shell,
                              "Confirmation",
                              "Some settings may become active only after restarting Timemine.\nRestart now?",
                              "Now",
                              "Later",
                              true
                             )
             )
          {
            Widgets.notify(shell,Timemine.USER_EVENT_QUIT,Timemine.EXITCODE_RESTART);
          }
        }
      }
    }
  }

  /** add colors to colors list
   */
  private void addColors(Table widgetColors)
  {
    try
    {
      // instantiate config adapter class
      Constructor               constructor         = Settings.SettingValueAdapterColor.class.getDeclaredConstructor(Settings.class);
      SettingUtils.ValueAdapter settingValueAdapter = (SettingUtils.ValueAdapter)constructor.newInstance(new Settings());

      // get setting classes
      Class[] settingClasses = Settings.getSettingClasses();
      for (Class clazz : settingClasses)
      {
        for (Field field : clazz.getDeclaredFields())
        {
          for (Annotation annotation : field.getDeclaredAnnotations())
          {
            if      (annotation instanceof SettingValue)
            {
              SettingValue configValue = (SettingValue)annotation;
              if (Settings.SettingValueAdapterColor.class.isAssignableFrom(configValue.type()))
              {
                // get color
                Settings.Color color = (Settings.Color)field.get(null);

                // get name
                String name = (!configValue.name().isEmpty()) ? configValue.name() : field.getName();

                // add entry
                color = color.clone();
                Widgets.addTableItem(widgetColors,color,name.substring(5));   // Note: remove prefix "Color"
                if (color.foreground != null) Widgets.setTableItemColor(widgetColors,color,1,new Color(null,color.foreground));
                if (color.background != null) Widgets.setTableItemColor(widgetColors,color,2,new Color(null,color.background));
              }
            }
          }
        }
      }
    }
    catch (Exception exception)
    {
      // cannot happen
      Timemine.printInternalError(exception);
    }
  }

  /** save colors
   * @param widgetColors color widget
   * @return
   */
  private void saveColors(Table widgetColors)
  {
    try
    {
      // instantiate config adapter class
      Constructor               constructor         = Settings.SettingValueAdapterColor.class.getDeclaredConstructor(Settings.class);
      SettingUtils.ValueAdapter settingValueAdapter = (SettingUtils.ValueAdapter)constructor.newInstance(new Settings());

      // get setting classes
      Class[] settingClasses = Settings.getSettingClasses();
      for (Class clazz : settingClasses)
      {
        for (Field field : clazz.getDeclaredFields())
        {
          for (Annotation annotation : field.getDeclaredAnnotations())
          {
            if      (annotation instanceof SettingValue)
            {
              SettingValue configValue = (SettingValue)annotation;
              if (Settings.SettingValueAdapterColor.class.isAssignableFrom(configValue.type()))
              {
                // get color
                Settings.Color color = (Settings.Color)field.get(null);

                // get name
                String name = (!configValue.name().isEmpty()) ? configValue.name() : field.getName();

                // set color
                name = name.substring(5);
                boolean found = false;
                for (TableItem tableItem : widgetColors.getItems())
                {
                  if (name.equals(tableItem.getText(0)))
                  {
                    field.set(null,tableItem.getData());
                    found = true;
                    break;
                  }
                }
                if (!found)
                {
                  Timemine.printInternalError("Color %s not found in table!",name);
                }
              }
            }
          }
        }
      }
    }
    catch (Exception exception)
    {
      // cannot happen
      Timemine.printInternalError(exception);
    }
  }

  /** edit color
   * @param shell parent shell
   * @param name name
   * @param color color
   * @return true if edit OK, false on cancel
   */
  private boolean editColor(Shell shell, String name, final Settings.Color color)
  {
    Composite composite,subComposite;
    Label     label;
    Text      text;
    Button    button;

    // add editor dialog
    final Shell dialog = Dialogs.openModal(shell,"Edit color",100,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Canvas widgetColorForeground;
    final Canvas widgetColorBackground;
    final Label  widgetValueForeground;
    final Label  widgetValueBackground;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Name:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      text = Widgets.newStringView(composite);
      text.setText(name);
      Widgets.layout(text,0,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Foreground:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,1,1,TableLayoutData.WE);
      {
        widgetColorForeground = Widgets.newCanvas(subComposite,SWT.BORDER);
        widgetColorForeground.setForeground((color.foreground != null) ? new Color(null,color.foreground) : null);
        widgetColorForeground.setBackground((color.foreground != null) ? new Color(null,color.foreground) : null);
        Widgets.layout(widgetColorForeground,0,0,TableLayoutData.WE,0,0,0,0,60,20);
        String colorName = (color.foreground != null)
                             ? String.format("#%02x%02x%02x",color.foreground.red,color.foreground.green,color.foreground.blue)
                             : "";
        widgetValueForeground = Widgets.newLabel(subComposite,colorName);
        Widgets.layout(widgetValueForeground,0,1,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
        widgetColorForeground.addMouseListener(new MouseListener()
        {
          @Override
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseDown(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseUp(MouseEvent mouseEvent)
          {
            ColorDialog colorDialog = new ColorDialog(dialog);
            colorDialog.setRGB(color.foreground);
            RGB rgb = colorDialog.open();
            if (rgb != null)
            {
              color.foreground = rgb;
              widgetColorForeground.setForeground(new Color(null,rgb));
              widgetColorForeground.setBackground(new Color(null,rgb));
              widgetValueForeground.setText(String.format("#%02x%02x%02x",color.foreground.red,color.foreground.green,color.foreground.blue));
            }
          }
        });
      }

      label = Widgets.newLabel(composite,"Background:");
      Widgets.layout(label,2,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,2,1,TableLayoutData.WE);
      {
        widgetColorBackground = Widgets.newCanvas(subComposite,SWT.BORDER);
        widgetColorBackground.setForeground((color.background != null) ? new Color(null,color.background) : null);
        widgetColorBackground.setBackground((color.background != null) ? new Color(null,color.background) : null);
        Widgets.layout(widgetColorBackground,0,0,TableLayoutData.WE,0,0,0,0,60,20);
        String colorName = (color.background != null)
                             ? String.format("#%02x%02x%02x",color.background.red,color.background.green,color.background.blue)
                             : "";
        widgetValueBackground = Widgets.newLabel(subComposite,colorName);
        Widgets.layout(widgetValueBackground,0,1,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
        widgetColorBackground.addMouseListener(new MouseListener()
        {
          @Override
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseDown(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseUp(MouseEvent mouseEvent)
          {
            ColorDialog colorDialog = new ColorDialog(dialog);
            colorDialog.setRGB(color.background);
            RGB rgb = colorDialog.open();
            if (rgb != null)
            {
              color.background = rgb;
              widgetColorBackground.setForeground(new Color(null,rgb));
              widgetColorBackground.setBackground(new Color(null,rgb));
              widgetValueBackground.setText(String.format("#%02x%02x%02x",color.background.red,color.background.green,color.background.blue));
            }
          }
        });
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      button = Widgets.newButton(composite,"Save");
      Widgets.layout(button,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Default");
      Widgets.layout(button,0,1,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          color.foreground = color.DEFAULT_FOREGROUND;
          color.background = color.DEFAULT_BACKGROUND;

          widgetColorForeground.setForeground((color.foreground != null) ? new Color(null,color.foreground) : null);
          widgetColorForeground.setBackground((color.foreground != null) ? new Color(null,color.foreground) : null);
          widgetColorBackground.setForeground((color.background != null) ? new Color(null,color.background) : null);
          widgetColorBackground.setBackground((color.background != null) ? new Color(null,color.background) : null);

          String foregroundColorName = (color.foreground != null) ? String.format("#%02x%02x%02x",color.foreground.red,color.foreground.green,color.foreground.blue) : "";
          String backgroundColorName = (color.background != null) ? String.format("#%02x%02x%02x",color.background.red,color.background.green,color.background.blue) : "";
          widgetValueForeground.setText(foregroundColorName);
          widgetValueBackground.setText(backgroundColorName);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,2,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // add listeners

    return (Boolean)Dialogs.run(dialog,false);
  }
}

/* end of file */
