;/***********************************************************************\
*
* $Revision: 878 $
* $Date: 2021-02-13 19:29:49 +0100 (Sat, 13 Feb 2021) $
* $Author: torsten $
* Contents: Sync animation widget
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.util.Timer;
import java.util.TimerTask;

// graphics
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;

/****************************** Classes ********************************/

/** Synchronize animation
 */
class SynchronizeAnimation
{
  private final Color COLOR_MODIFIED;
  private final Color COLOR_OFFLINE;
  private final Image IMAGE_SYNC_ANIMATION[];
  private final Image IMAGE_OFFLINE;

  private Button                  button;
  private Display                 display;
  private Timer                   timer;
  private Backend.ConnectionStates connectionState;
  private boolean                 modifiedFlag;
  private Integer                 counter;
  private int                     imageIndex;
  private TimerTask               timerTask;

  /** create sync animation
   */
  SynchronizeAnimation(Display display)
  {
    IMAGE_SYNC_ANIMATION = new Image[]
                           {
                             Widgets.loadImage(display,"syncAnimation0.png"),
                             Widgets.loadImage(display,"syncAnimation1.png"),
                             Widgets.loadImage(display,"syncAnimation2.png"),
                             Widgets.loadImage(display,"syncAnimation3.png")
                           };
    IMAGE_OFFLINE        = Widgets.loadImage(display,"offline.png");
    COLOR_MODIFIED       = display.getSystemColor(SWT.COLOR_YELLOW);
    COLOR_OFFLINE        = display.getSystemColor(SWT.COLOR_GRAY);

    this.display    = display;
    timer           = new Timer("SynchronizeAnimation",true);
    connectionState = Backend.ConnectionStates.OFFLINE;
    modifiedFlag    = false;
    counter         = 0;
    imageIndex      = 0;
    timerTask       = null;

    Backend.addConnectionStatesListener(new Backend.StateListener()
    {
      public void changed(Backend.ConnectionStates connectionState)
      {
        setConnectionState(connectionState);
      }
    });

    Backend.addModifyListener(new Backend.ModifyListener()
    {
      public void modified(Entity entity, Backend.ModifiedStates modifiedState)
      {
        modifiedFlag = (modifiedState != Backend.ModifiedStates.NONE);
      }
    });
  }

  /** create new sync animation button
   * @param composite parent composite
   */
  public Button newButton(Composite composite)
  {
    button = Widgets.newButton(composite,IMAGE_OFFLINE);

    return button;
  }

  /** start sync animation
   */
  public void start()
  {
    synchronized(counter)
    {
      counter++;
      if (counter == 1)
      {
        startAnimation();
      }
    }
  }

  /** stop sync animation
   */
  public void stop()
  {
    synchronized(counter)
    {
      assert counter > 0;

      counter--;
      if (counter <= 0)
      {
        stopAnimation();
      }
    }
  }

  /** start sync animation
   */
  private void startAnimation()
  {
    if ((connectionState != Backend.ConnectionStates.OFFLINE) && (timerTask == null))
    {
      imageIndex = 0;
      timerTask = new TimerTask()
      {
        public void run()
        {
          imageIndex = (imageIndex+1) % 4;
          if (!display.isDisposed())
          {
            display.syncExec(new Runnable()
            {
              public void run()
              {
                if ((button != null) && !button.isDisposed())
                {
                  button.setImage(IMAGE_SYNC_ANIMATION[imageIndex]);
                }
              }
            });
          }
        }
      };
      timer.schedule(timerTask,0,250);
    }
  }

  /** stop sync animation
   */
  private void stopAnimation()
  {
    if ((connectionState != Backend.ConnectionStates.OFFLINE) && (timerTask != null))
    {
      timerTask.cancel();
      timerTask = null;
    }
  }

  /** update sync animation
   */
  private void updateAnimation()
  {
    synchronized(counter)
    {
      if      (counter >= 1) startAnimation();
      else if (counter == 0) stopAnimation();
    }
  }

  /** set new connection state
   * @param newConnectionStates new connection state to set
   */
  private void setConnectionState(Backend.ConnectionStates newConnectionStates)
  {
    switch (newConnectionStates)
    {
      case ONLINE:
        display.syncExec(new Runnable()
        {
          public void run()
          {
            if ((button != null) && !button.isDisposed())
            {
              button.setBackground(null);
              button.setImage(IMAGE_SYNC_ANIMATION[imageIndex]);
            }
          }
        });
        break;
/*
      case MODIFIED:
        if (this.state != States.OFFLINE)
        {
          this.state = state;
          display.syncExec(new Runnable()
          {
            public void run()
            {
              if ((button != null) && !button.isDisposed())
              {
                button.setBackground(COLOR_MODIFIED);
                button.setImage(IMAGE_SYNC_ANIMATION[imageIndex]);
              }
            }
          });
        }
        modifiedFlag = true;
        break;
*/
      case OFFLINE:
        display.syncExec(new Runnable()
        {
          public void run()
          {
            if ((button != null) && !button.isDisposed())
            {
              button.setBackground(COLOR_OFFLINE);
              button.setImage(IMAGE_OFFLINE);
            }
          }
        });
        break;
    }
    connectionState = newConnectionStates;

    updateAnimation();
  }
}

/* end of file */
