/***********************************************************************\
*
* $Revision: 319 $
* $Date: 2016-05-25 08:21:52 +0200 (Wed, 25 May 2016) $
* $Author: torsten $
* Contents: Timeline widget
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.ArrayList;
import java.util.Calendar;

// graphics
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.MouseTrackListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Label;

/****************************** Classes ********************************/

/** time line widget
 */
public class Timeline extends Canvas
{
  private enum States
  {
    IDLE,
    ACTIVE
  };

  // --------------------------- constants --------------------------------
  private Color COLOR_BACKGROUND;
  private Color COLOR_INFO_BACKGROUND;

  private Color COLOR_ACTIVE;
  private Color COLOR_IDLE;

  private Color COLOR_ACTIONS[];

  private Color COLOR_RED;
  private Color COLOR_ORANGE;
  private Color COLOR_GREEN;
  private Color COLOR_BLUE;

  private final int  WIDTH           = 1024;
  private final int  HEIGHT          = 90;
  private final int  MARGIN          = 1;
  private final int  BAR_HEIGHT      = 25;
  private final int  BAR_PADDING     = 5;
  private final long SECONDS_PER_DAY = 24*60*60;

  // date/time format
  private final SimpleDateFormat TIME_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

  // --------------------------- variables --------------------------------
  private final Display     display;
  private final Shell       shell;

  private long              midnightDateTime = -1;
  private int               textHeight;
  private long              startDateTime    = -1;
  private Object            lock             = new Object();
  private ActionList        actionList       = null;
  private double            zoomFactor       = 1.0;
  private long              tagStartTime     = 0;
  private long              tagEndTime       = 0;
  private Shell             widgetToolTip    = null;
  private Text              widgetToolTipText;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create time line
   */
  public Timeline(Composite composite, int width)
  {
    super(composite,SWT.BORDER);//|SWT.H_SCROLL);
    // canvas is a composite; set default layout
    setLayout(new TableLayout(0.0,0.0,0));
    setSize((int)((double)WIDTH*zoomFactor),HEIGHT);
    // set scrolled composite content
    if (composite instanceof ScrolledComposite)
    {
      ((ScrolledComposite)composite).setContent(this);
    }

    display = composite.getDisplay();
    shell   = composite.getShell();

    // init colors
    COLOR_BACKGROUND      = composite.getDisplay().getSystemColor(SWT.COLOR_WIDGET_BACKGROUND);
    COLOR_INFO_BACKGROUND = new Color(composite.getDisplay(),224,255,255);

    COLOR_ACTIVE          = new Color(display,150,255,132);
    COLOR_IDLE            = new Color(display,255,148,148);

    COLOR_ACTIONS         = new Color[48];
    COLOR_ACTIONS[ 0]     = new Color(display,  0,  0,255);
    COLOR_ACTIONS[ 1]     = new Color(display,  0,255,  0);
    COLOR_ACTIONS[ 2]     = new Color(display,  0,255,255);
    COLOR_ACTIONS[ 3]     = new Color(display,127,127,127);
    COLOR_ACTIONS[ 4]     = new Color(display,127,255,  0);
    COLOR_ACTIONS[ 5]     = new Color(display,255,  0,  0);
    COLOR_ACTIONS[ 6]     = new Color(display,255,  0,127);
    COLOR_ACTIONS[ 7]     = new Color(display,255,  0,255);
    COLOR_ACTIONS[ 8]     = new Color(display,255,255,  0);
    COLOR_ACTIONS[ 9]     = new Color(display,255,255,255);
    COLOR_ACTIONS[10]     = new Color(display,  0,  0,204);
    COLOR_ACTIONS[11]     = new Color(display,  0,102, 51);
    COLOR_ACTIONS[12]     = new Color(display,  0,153, 51);
    COLOR_ACTIONS[13]     = new Color(display,  0,204,204);
    COLOR_ACTIONS[14]     = new Color(display,  0,204,255);
    COLOR_ACTIONS[15]     = new Color(display,  0,255, 51);
    COLOR_ACTIONS[16]     = new Color(display, 51,  0,255);
    COLOR_ACTIONS[17]     = new Color(display, 51,102,102);
    COLOR_ACTIONS[18]     = new Color(display, 51,153,102);
    COLOR_ACTIONS[19]     = new Color(display, 51,153,255);
    COLOR_ACTIONS[20]     = new Color(display, 51,204,102);
    COLOR_ACTIONS[21]     = new Color(display, 51,255,102);
    COLOR_ACTIONS[22]     = new Color(display,102,  0,204);
    COLOR_ACTIONS[23]     = new Color(display,102, 51,255);
    COLOR_ACTIONS[24]     = new Color(display,102,153,255);
    COLOR_ACTIONS[25]     = new Color(display,102,255,204);
    COLOR_ACTIONS[26]     = new Color(display,153,  0,  0);
    COLOR_ACTIONS[27]     = new Color(display,153, 51,102);
    COLOR_ACTIONS[28]     = new Color(display,153,102, 51);
    COLOR_ACTIONS[29]     = new Color(display,153,153,102);
    COLOR_ACTIONS[30]     = new Color(display,153,153,255);
    COLOR_ACTIONS[31]     = new Color(display,153,255,  0);
    COLOR_ACTIONS[32]     = new Color(display,204, 51,  0);
    COLOR_ACTIONS[33]     = new Color(display,204,153, 51);
    COLOR_ACTIONS[34]     = new Color(display,204,153,255);
    COLOR_ACTIONS[35]     = new Color(display,204,255,102);
    COLOR_ACTIONS[36]     = new Color(display,255,  0,102);
    COLOR_ACTIONS[37]     = new Color(display,255,153, 51);
    COLOR_ACTIONS[38]     = new Color(display,255,153,102);
    COLOR_ACTIONS[39]     = new Color(display,255,153,153);
    COLOR_ACTIONS[40]     = new Color(display,255,153,204);
    COLOR_ACTIONS[41]     = new Color(display,255,153,255);
    COLOR_ACTIONS[42]     = new Color(display,255,204,  0);
    COLOR_ACTIONS[43]     = new Color(display,255,204,153);
    COLOR_ACTIONS[44]     = new Color(display,255,204,204);
    COLOR_ACTIONS[45]     = new Color(display,255,204,255);
    COLOR_ACTIONS[46]     = new Color(display,255,255,153);
    COLOR_ACTIONS[47]     = new Color(display,255,255,153);

    COLOR_RED             = display.getSystemColor(SWT.COLOR_RED);
    COLOR_ORANGE          = display.getSystemColor(SWT.COLOR_YELLOW);
    COLOR_GREEN           = display.getSystemColor(SWT.COLOR_GREEN);
    COLOR_BLUE            = display.getSystemColor(SWT.COLOR_BLUE);

    // get midnight of today
    midnightDateTime = getMidnight();

/*
    addMouseListener(new MouseListener()
    {
      public void mouseDoubleClick(MouseEvent mouseEvent)
      {
      }
      public void mouseDown(MouseEvent mouseEvent)
      {
        tagStartTimestamp = toTimestamp(mouseEvent.x);
        tagEndTimestamp   = tagStartTimestamp;
        redraw();
      }
      public void mouseUp(MouseEvent mouseEvent)
      {
        tagStartTimestamp = 0;
        tagEndTimestamp   = 0;
        redraw();
      }
    });*/

    addMouseTrackListener(new MouseTrackListener()
    {
      @Override
      public void mouseEnter(MouseEvent mouseEvent)
      {
      }

      @Override
      public void mouseExit(MouseEvent mouseEvent)
      {
        Rectangle bounds = Timeline.this.getBounds();

        if ((widgetToolTip != null) && !bounds.contains(mouseEvent.x,mouseEvent.y))
        {
//          widgetToolTip.dispose();
//          widgetToolTip = null;
        }
      }

      @Override
      public void mouseHover(MouseEvent mouseEvent)
      {
        Label label;

        if (widgetToolTip != null)
        {
          widgetToolTip.dispose();
          widgetToolTip = null;
        }

        synchronized(lock)
        {
          if (   ((mouseEvent.stateMask & SWT.BUTTON_MASK) == 0)
              && (actionList != null)
              && (   ((MARGIN                               <= mouseEvent.y) && (mouseEvent.y <= MARGIN+BAR_HEIGHT                              ))
                  || ((MARGIN+BAR_HEIGHT+BAR_PADDING+MARGIN <= mouseEvent.y) && (mouseEvent.y <= MARGIN+BAR_HEIGHT+BAR_PADDING+MARGIN+BAR_HEIGHT))
                 )
             )
          {
            // find action
            Action foundAction = null;
            long   timestamp   = toTimestamp(mouseEvent.x);
            for (int i = 0; i < actionList.size(); i++)
            {
              Action action = actionList.get(i);
              if (   ((action.type == ActionTypes.WINDOW) || (action.type == ActionTypes.APPLICATION) || (action.type == ActionTypes.FILE))
                  && (timestamp >= action.timestamp)
                  && ((i+1 >= actionList.size()) || timestamp <= actionList.get(i+1).timestamp)
                 )
              {
                foundAction = action;
                break;
              }
            }

            if (foundAction != null)
            {
              // show action tooltip
              widgetToolTip = new Shell(shell,SWT.ON_TOP|SWT.NO_FOCUS|SWT.TOOL);
              widgetToolTip.setBackground(COLOR_INFO_BACKGROUND);
              widgetToolTip.setLayout(new TableLayout(1.0,new double[]{0.0,1.0},2));
              Widgets.layout(widgetToolTip,0,0,TableLayoutData.NSWE);

              label = Widgets.newLabel(widgetToolTip,String.format("%s: %s\n%s",
                                                                   TIME_FORMAT.format(new Date(foundAction.timestamp*1000L)),
                                                                   foundAction.type.toString(),
                                                                   foundAction.getText()
                                                                  )
                                      );
//              label.setForeground(COLOR_INFO_FORGROUND);
              label.setBackground(COLOR_INFO_BACKGROUND);
              Widgets.layout(label,0,0,TableLayoutData.W);

              label.addMouseTrackListener(new MouseTrackListener()
              {
                @Override
                public void mouseEnter(MouseEvent mouseEvent)
                {
                }

                @Override
                public void mouseExit(MouseEvent mouseEvent)
                {
                  if (widgetToolTip != null)
                  {
                    widgetToolTip.dispose();
                    widgetToolTip = null;
                  }
                }

                @Override
                public void mouseHover(MouseEvent mouseEvent)
                {
                }
              });

              Point point = display.getCursorLocation();
              Point size  = widgetToolTip.computeSize(SWT.DEFAULT,SWT.DEFAULT);
              widgetToolTip.setBounds(point.x,point.y,size.x,size.y);
              widgetToolTip.setVisible(true);
            }
          }
        }
      }
    });

    addMouseMoveListener(new MouseMoveListener()
    {
      public void mouseMove(MouseEvent mouseEvent)
      {
        Rectangle bounds = Timeline.this.getBounds();

        if (widgetToolTip != null)
        {
          if (bounds.contains(mouseEvent.x,mouseEvent.y))
          {
//Dprintf.dprintf("");
          }
          else
          {
Dprintf.dprintf("");
            widgetToolTip.dispose();
            widgetToolTip = null;
          }
        }
      }
    });

    addPaintListener(new PaintListener()
    {
      public void paintControl(PaintEvent paintEvent)
      {
        GC        gc   = paintEvent.gc;
        Rectangle size = Timeline.this.getClientArea();

        gc.setBackground(COLOR_BACKGROUND);
        gc.fillRectangle(0,0,size.width,size.height);

        drawTimeAxis(gc,size);
        drawActions(gc,size);
        drawTags(gc,size);
        drawStartTimeMarker(gc,size);
        drawCurrentTimeMarker(gc,size);
        drawCurrentTag(gc,size);
      }
    });
  }

  /** set date
   * @param dateTime date/time
   */
  public void setDate(long dateTime)
  {
    midnightDateTime = getMidnight(dateTime);
    if (startDateTime < midnightDateTime) startDateTime = -1;
    if (!isDisposed())
    {
      redraw();
    }
  }

  /** set start time
   * @param dateTime date/time
   */
  public void setStartTime(long dateTime)
  {
    startDateTime = dateTime;
    if (!isDisposed())
    {
      redraw();
    }
  }

  /** refresh action list and redraw
   */
  public void refresh(ActionList actionList)
  {
    synchronized(lock)
    {
      this.actionList = (ActionList)actionList.clone();
      if (!isDisposed())
      {
        redraw();
      }
    }
  }

  /** get tag start time
   * @return tag start time [s]
   */
  public long getTagStartTime()
  {
    return tagStartTime;
  }

  /** set tag start time
   * @param tag start time [s]
   */
  public void setTagStartTime(long time)
  {
    tagStartTime = time;
    tagEndTime   = time;
    redraw();
  }

  /** get tag end time
   * @return tag end time [s]
   */
  public long getTagEndTime()
  {
    return tagEndTime;
  }

  /** set tag end time
   * @param tag end time [s]
   */
  public void setTagEndTime(long time)
  {
    tagEndTime = time;
    redraw();
  }

  /** reset action list
   */
  public void clearTag()
  {
    tagStartTime = 0;
    tagEndTime   = 0;
    redraw();
  }

  /** set zoom factor
   * @param dateTime date/time
   */
  public double getZoomFactor()
  {
    return zoomFactor;
  }

  /** set zoom factor
   * @param zoomFactor zoom facator (>= 1)
   */
  public void setZoomFactor(double zoomFactor)
  {
    this.zoomFactor = zoomFactor;
    setSize((int)((double)WIDTH*zoomFactor),HEIGHT);
  }

  /** increment zoom factor
   * @param increment increment
   * @param limit max. zoom factor
   */
  public void incrementZoomFactor(double increment, double limit)
  {
    zoomFactor += increment;
    if      (zoomFactor < 1    ) zoomFactor = 1.0;
    else if (zoomFactor > limit) zoomFactor = limit;
    setSize((int)((double)WIDTH*zoomFactor),HEIGHT);
  }

  /** get timestamp from x-coordinate
   * @param x-coordinate
   * @return timestamp [s]
   */
  public long toTimestamp(int x)
  {
    long deltaSeconds = (long)(((double)x*(double)SECONDS_PER_DAY)/((double)WIDTH*zoomFactor));
    return midnightDateTime+deltaSeconds;
  }

  /** get x-coordinate for time value
   * @param dateTime date/time value
   * @return x-coordinate
   */
  public int toX(long dateTime)
  {
    long deltaSeconds = dateTime-midnightDateTime;
    return (int)(((double)deltaSeconds*(double)WIDTH*zoomFactor)/(double)SECONDS_PER_DAY);
  }

  // -------------------------------------------------------------------

  /** get midnight of date
   * @return midnight timestamp [s]
   */
  private long getMidnight(long dateTime)
  {
    Calendar calendar = Calendar.getInstance();
    calendar.setTimeInMillis(dateTime*1000);
    calendar.set(Calendar.HOUR_OF_DAY,0);
    calendar.set(Calendar.MINUTE,0);
    calendar.set(Calendar.SECOND,0);

    return calendar.getTimeInMillis()/1000;
  }

  /** get midnight of today
   * @return midnight timestamp [s]
   */
  private long getMidnight()
  {
    // get midnight of today
    Calendar calendar = Calendar.getInstance();
    calendar.set(Calendar.HOUR_OF_DAY,0);
    calendar.set(Calendar.MINUTE,0);
    calendar.set(Calendar.SECOND,0);

    return calendar.getTimeInMillis()/1000;
  }

  /** draw time axis
   * @param gc graphics context
   * @param size widget size
   */
  private boolean isAtBar0(int x, int y)
  {
    return ((MARGIN                               <= y) && (y <= MARGIN+BAR_HEIGHT                              ));
  }

  /** draw time axis
   * @param gc graphics context
   * @param size widget size
   */
  private boolean isAtBar1(int x, int y)
  {
    return ((MARGIN+BAR_HEIGHT+BAR_PADDING+MARGIN <= y) && (y <= MARGIN+BAR_HEIGHT+BAR_PADDING+MARGIN+BAR_HEIGHT));
  }

  /** draw time axis
   * @param gc graphics context
   * @param size widget size
   */
  private void drawTimeAxis(GC gc, Rectangle size)
  {
    int      x;
    String   text;
    Point    p;
    for (int h = 0; h < 24; h++)
    {
      x = toX(midnightDateTime+h*60*60);
      text = String.format("%02d:00",h);
      p = gc.textExtent(text);
      gc.setBackground(COLOR_BACKGROUND);
      gc.drawText(text,x-p.x/2,size.height-p.y-2);
      gc.drawLine(x,size.height-p.y-2-8,x,size.height-p.y-2);

      textHeight = p.y;
    }
  }

  /** draw actions
   * @param gc graphics context
   * @param size widget size
   */
  private void drawActions(GC gc, Rectangle size)
  {
    long lastStateTimestamp  = midnightDateTime;
    long lastActionTimestamp = midnightDateTime;
    int  x0,x1,x2;

    States state = States.IDLE;
    synchronized(lock)
    {
      if (actionList != null)
      {
        int i = 0;
        for (Action action : actionList)
        {
//Dprintf.dprintf("action=%s",action);
          x0 = toX(action.timestamp);

          if (action.timestamp >= lastStateTimestamp)
          {
            // draw active/inactive bar
            x1 = toX(lastStateTimestamp);

            gc.setBackground((state == States.ACTIVE) ? COLOR_ACTIVE : COLOR_IDLE);
            gc.fillRectangle(x1,MARGIN+0,x0-x1,BAR_HEIGHT-MARGIN*2);
            lastStateTimestamp = action.timestamp;
          }

          if (action.timestamp >= lastActionTimestamp)
          {
            // draw action bar
            x2 = toX(lastActionTimestamp);

            switch (action.type)
            {
              case IDLE:
                state = States.IDLE;
                break;
              case ACTIVE:
                state = States.ACTIVE;
                break;
              case WINDOW:
              case APPLICATION:
              case FILE:
                state = States.ACTIVE;
                gc.setBackground(COLOR_ACTIONS[i%COLOR_ACTIONS.length]);
                gc.fillRectangle(x2,MARGIN+BAR_HEIGHT+BAR_PADDING,x0-x2,BAR_HEIGHT-MARGIN*2);
                lastActionTimestamp = action.timestamp;
                i++;
                break;
            }
          }
        }
      }
    }

if (false){
    // draw
    x0 = toX(System.currentTimeMillis()/1000L);
    x1 = toX(lastStateTimestamp);
    gc.setBackground((state == States.ACTIVE) ? COLOR_ACTIVE : COLOR_IDLE);
    gc.fillRectangle(x1,MARGIN,x0-x1,25-MARGIN*2);
}
  }

  /** draw tags
   * @param gc graphics context
   * @param size widget size
   */
  private void drawTags(GC gc, Rectangle size)
  {
  }

  /** draw start time marker line
   * @param gc graphics context
   * @param size widget size
   */
  private void drawStartTimeMarker(GC gc, Rectangle size)
  {
    int x;

    if (startDateTime >= 0)
    {
      x = toX(startDateTime);
      gc.drawLine(x,0,x,size.height);
    }
  }

  /** draw current time marker line
   * @param gc graphics context
   * @param size widget size
   */
  private void drawCurrentTimeMarker(GC gc, Rectangle size)
  {
    int x;

    x = toX(System.currentTimeMillis()/1000L);
    gc.drawLine(x,0,x,size.height);
  }

  /** draw current tag
   * @param gc graphics context
   * @param size widget size
   */
  private void drawCurrentTag(GC gc, Rectangle size)
  {
    final int MARGIN = 0;

    long t0,t1;
    int  x0,x1;
    int  h;

    if (tagStartTime > 0)
    {
      if (tagStartTime < tagEndTime)
      {
        t0 = tagStartTime;
        t1 = tagEndTime;
      }
      else
      {
        t0 = tagEndTime;
        t1 = tagStartTime;
      }

      x0 = toX(t0);
      x1 = toX(t1);
      h  = BAR_HEIGHT+BAR_PADDING+BAR_HEIGHT-2*MARGIN;

      // draw [
      gc.drawLine(x0,  MARGIN  ,x0+0,MARGIN  );
      gc.drawLine(x0,  MARGIN  ,x0  ,MARGIN+h);
      gc.drawLine(x0,  MARGIN+h,x0+3,MARGIN+h);

      // draw ]
      gc.drawLine(x1-3,MARGIN  ,x1  ,MARGIN  );
      gc.drawLine(x1  ,MARGIN  ,x1  ,MARGIN+h);
      gc.drawLine(x1-3,MARGIN+h,x1  ,MARGIN+h);
    }
  }
}

/* end of file */
