/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** activity database id
 */
class ActivityId extends Id
{
  /** create id
   * @param backend backend id
   */
  ActivityId(int backend)
  {
    super(backend);
  }
}

/** activity
 */
class Activity extends Entity
{
  public final Id      projectId;
  public final String  name;
  public String        description;
  public final boolean isDefault;

  /** create activity
   * @param backendType backend type
   * @param id database id
   * @param projectId project id or Id.ANY
   * @param name name
   * @param description description
   * @param isDefault true iff default
   */
  Activity(Backend.Types backendType,
           Id            id,
           Id            projectId,
           String        name,
           String        description,
           boolean       isDefault
          )
  {
    super(backendType,id);
    assert backendType != Backend.Types.NONE;
    assert !id.isNone();
    assert !id.isAny();
    assert(name != null);

    this.projectId   = projectId;
    this.name        = name;
    this.description = description;
    this.isDefault   = isDefault;
  }

  /** create activity
   * @param id database id
   */
  private Activity(Id id)
  {
    super(Backend.Types.NONE,id);

    this.projectId   = new Id();
    this.name        = "";
    this.description = "";
    this.isDefault   = false;
  }

  /** create activity
   */
  private Activity()
  {
    this(new Id());
  }

  /** get database parameters
   * @returns hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("id",id);
    databaseParameters.put("projectId",projectId);
    databaseParameters.put("name",name);
    databaseParameters.put("description",description);
    databaseParameters.put("isDefault",isDefault);

    return databaseParameters;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "Activity { id="+id+", projectId="+projectId+", name="+name+", description="+description+", isDefault="+isDefault+" }";
  }
}

/** activity list
 */
class ActivityList extends ArrayList<Activity>
{
  final static ActivityList EMPTY = new ActivityList();
}

/** activity database handler
 */
class ActivityDatabaseHandler extends DataBaseHandler<Activity>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM activities"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public Activity get(ValueSet valueSet)
    throws SQLException
  {
    return new Activity(valueSet.getEnum   ("backendType",Backend.Types.class),
                        valueSet.getId     ("id","backendId"),
                        valueSet.getId     ("projectId"),
                        valueSet.getString ("name"),
                        valueSet.getString ("description"),
                        valueSet.getBoolean("isDefault")
                       );
  }
}

/* end of file */
