/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** epic name
 */
class EpicName extends Entity
{
  public final String name;

  /** create epic name
   * @param backendType backend type
   * @param id database id
   * @param name name
   */
  EpicName(Backend.Types backendType,
           Id            id,
           String        name
          )
  {
    super(backendType,id);
    assert backendType != Backend.Types.NONE;
    assert !id.isNone();
    assert !id.isAny();
    assert(name != null);

    this.name = name;
  }

  /** create epic name
   */
  private EpicName(Id id)
  {
    super(Backend.Types.NONE,id);

    this.name = "";
  }

  /** get database parameters
   * @return hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("name",name);

    return databaseParameters;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "EpicName { name="+name+" }";
  }
}

/** epic name list
 */
class EpicNameList extends ArrayList<EpicName>
{
  final static EpicNameList EMPTY = new EpicNameList();
}

/** Epic name database handler
 */
class EpicNameDatabaseHandler extends DataBaseHandler<String>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM epicNames"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public String get(ValueSet valueSet)
    throws SQLException
  {
    return valueSet.getString("name");
  }
}

/* end of file */
