/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Set;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** issue database id
 */
class IssueId extends Id
{
  /** create id
   * @param backend backend id
   */
  IssueId(int backend)
  {
    super(backend);
  }
}

/** issue
 */
class Issue extends Entity implements Cloneable
{
  public final static Issue NONE = new Issue(new Id(Id.NONE));
  public final static Issue ANY  = new Issue(new Id(Id.ANY));

  public final static long  ESTIMATED_NONE = 0L;

  public Id     parentId;
  public String shortName;
  public Id     projectId;
  public Id     trackerId;
  public String epicName;
  public Id     statusId;              // current status id
  public Id     statusTransitionId;    // status transition id or Id.NONE: new status id
  public Id     priorityId;
  public Id     authorId;
  public String summary;
  public String description;
  public Date   startDate;
  public Date   dueDate;
  public Id     assignedToMemberId;
  public int    doneRatio;
  public Id     targetVersionId;
  public long   estimatedDuration;     // estimation duration [min]
  public Date   createdOn;
  public Date   updatedOn;
  public Date   closedOn;
  public Id     newProjectId;          // new project id or Id.NONE (move operation)

  /** create issue
   * @param backendType backend type
   * @param id database id
   * @param parentId parent issue id or Id.NONE
   * @param shortName short name or null
   * @param projectId project id
   * @param trackerId tracker id
   * @param epicName epic name or null
   * @param statusId status id
   * @param statusTransitionId status transition id or Id.NONE
   * @param priorityId priority id
   * @param authorId author user id
   * @param summary summary text
   * @param description description text
   * @param startDate start date
   * @param dueDate due date
   * @param assignedToMemberId assigned to member id or Id.ANY
   * @param doneRatio done ration
   * @param targetVersionId target version id or Id.ANY
   * @param estimatedDuration estimated duration [min]
   * @param createdOn created-on date
   * @param updatedOn update-on date
   * @param closedOn closed-on date
   * @param newProjectId new project id or Id.NONE
   */
  Issue(Backend.Types backendType,
        Id            id,
        Id            parentId,
        String        shortName,
        Id            projectId,
        Id            trackerId,
        String        epicName,
        Id            statusId,
        Id            statusTransitionId,
        Id            priorityId,
        Id            authorId,
        String        summary,
        String        description,
        Date          startDate,
        Date          dueDate,
        Id            assignedToMemberId,
        int           doneRatio,
        Id            targetVersionId,
        long          estimatedDuration,
        Date          createdOn,
        Date          updatedOn,
        Date          closedOn,
        Id            newProjectId
       )
  {
    super(backendType,id);

    assert shortName != null;
    assert summary != null;
    assert description != null;

    // current values
    this.parentId                    = parentId;
    this.shortName                   = shortName;
    this.projectId                   = projectId;
    this.trackerId                   = trackerId;
    this.epicName                    = epicName;
    this.statusId                    = statusId;
    this.statusTransitionId          = statusTransitionId;
    this.priorityId                  = priorityId;
    this.authorId                    = authorId;
    this.summary                     = summary;
    this.description                 = description;
    this.startDate                   = startDate;
    this.dueDate                     = dueDate;
    this.assignedToMemberId          = assignedToMemberId;
    this.doneRatio                   = doneRatio;
    this.targetVersionId             = targetVersionId;
    this.estimatedDuration           = estimatedDuration;
    this.createdOn                   = createdOn;
    this.updatedOn                   = updatedOn;
    this.closedOn                    = closedOn;
    this.newProjectId                = newProjectId;
  }

  /** create issue
   * @param backendType backend type
   * @param id database id
   * @param parentId parent issue id or Id.ANY
   * @param shortName short name or null
   * @param projectId project id
   * @param trackerId tracker id
   * @param epicName epic name or null
   * @param statusId status id
   * @param statusTransitionId status transition id or Id.NONE
   * @param priorityId priority id
   * @param authorId author user id
   * @param summary summary text
   * @param description description text
   * @param startDate start date
   * @param dueDate due date
   * @param assignedToMemberId assigned to member id
   * @param doneRatio done ration
   * @param targetVersionId target version id
   * @param estimatedDuration estimated duration [min]
   * @param createdOn created-on date
   * @param updatedOn update-on date
   * @param closedOn closed-on date
   */
  Issue(Backend.Types backendType,
        Id            id,
        Id            parentId,
        String        shortName,
        Id            projectId,
        Id            trackerId,
        String        epicName,
        Id            statusId,
        Id            statusTransitionId,
        Id            priorityId,
        Id            authorId,
        String        summary,
        String        description,
        Date          startDate,
        Date          dueDate,
        Id            assignedToMemberId,
        int           doneRatio,
        Id            targetVersionId,
        long          estimatedDuration,
        Date          createdOn,
        Date          updatedOn,
        Date          closedOn
       )
  {
    this(backendType,
         id,
         parentId,
         shortName,
         projectId,
         trackerId,
         epicName,
         statusId,
         statusTransitionId,
         priorityId,
         authorId,
         summary,
         description,
         startDate,
         dueDate,
         assignedToMemberId,
         doneRatio,
         targetVersionId,
         estimatedDuration,
         createdOn,
         updatedOn,
         closedOn,
         new Id()
        );
  }

  /** create issue
   * @param backendType backend type
   * @param id project id
   * @param parentId parent issue id or Id.ANY
   * @param shortName short name or null
   * @param projectId project id
   * @param trackerId tracker id
   * @param epicName epic name or null
   * @param statusId status id
   * @param priorityId priority id
   * @param authorId author user id
   * @param summary summary text
   * @param description description text
   * @param startDate start date
   * @param dueDate due date
   * @param assignedToMemberId assigned to member id
   * @param doneRatio done ration
   * @param targetVersionId target version id
   * @param estimatedDuration estimated duration [min]
   * @param createdOn created-on date
   * @param updatedOn update-on date
   * @param closedOn closed-on date
   */
  Issue(Backend.Types backendType,
        Id            id,
        Id            parentId,
        String        shortName,
        Id            projectId,
        Id            trackerId,
        String        epicName,
        Id            statusId,
        Id            priorityId,
        Id            authorId,
        String        summary,
        String        description,
        Date          startDate,
        Date          dueDate,
        Id            assignedToMemberId,
        int           doneRatio,
        Id            targetVersionId,
        long          estimatedDuration,
        Date          createdOn,
        Date          updatedOn,
        Date          closedOn
       )
  {
    this(backendType,
         id,
         parentId,
         shortName,
         projectId,
         trackerId,
         epicName,
         statusId,
         new Id(),  // statusTransitionId
         priorityId,
         authorId,
         summary,
         description,
         startDate,
         dueDate,
         assignedToMemberId,
         doneRatio,
         targetVersionId,
         estimatedDuration,
         createdOn,
         updatedOn,
         closedOn
        );
  }

  /** create issue
   * @param backendType backend type
   * @param id project id
   * @param parentId parent issue id or Id.ANY
   * @param projectId project id
   * @param trackerId tracker id
   * @param statusId status id
   * @param priorityId priority id
   * @param authorId author user id
   * @param summary summary text
   * @param description description text
   * @param startDate start date
   * @param dueDate due date
   * @param assignedToMemberId assigned to member id
   * @param doneRatio done ration
   * @param targetVersionId target version id
   * @param estimatedDuration estimated duration [min]
   * @param createdOn created-on date
   * @param updatedOn update-on date
   * @param closedOn closed-on date
   */
  Issue(Backend.Types backendType,
        Id            id,
        Id            parentId,
        Id            projectId,
        Id            trackerId,
        Id            statusId,
        Id            priorityId,
        Id            authorId,
        String        summary,
        String        description,
        Date          startDate,
        Date          dueDate,
        Id            assignedToMemberId,
        int           doneRatio,
        Id            targetVersionId,
        long          estimatedDuration,
        Date          createdOn,
        Date          updatedOn,
        Date          closedOn
       )
  {
    this(backendType,
         id,
         parentId,
         (String)null,  // shortName,
         projectId,
         trackerId,
         (String)null,  // epicName,
         statusId,
         priorityId,
         authorId,
         summary,
         description,
         startDate,
         dueDate,
         assignedToMemberId,
         doneRatio,
         targetVersionId,
         estimatedDuration,
         createdOn,
         updatedOn,
         closedOn
        );
  }

  /** create issue
   * @param backendType backend type
   * @param projectId project id
   * @param issueId issue id or Id.NONE
   */
  Issue(Backend.Types backendType,
        Id            projectId,
        Id            issueId
       )
  {
    this(backendType,
         issueId,
         new Id(),                // parentId
         "",                      // shortName
         projectId,
         new Id(),                // trackerId
         null,                    // epic name
         new Id(),                // statusId
         new Id(),                // statusTransitionId
         new Id(),                // priorityId
         new Id(),                // authorId
         "",                      // summary
         "",                      // description
         new Date(),              // startDate
         new Date(),              // dueDate
         new Id(),                  // assignedToMemberId
         0,                       // doneRatio
         new Id(),                // targetVersionId
         ESTIMATED_NONE,          // estimatedDuration
         new Date(),              // createdOn
         new Date(),              // updatedOn
         new Date()               // closedOn
        );

    try
    {
      Backend.TypeSet typeSet = new Backend.TypeSet(backendType);
      this.statusId = Backend.getDefaultStatusId(typeSet);
      this.statusId = Backend.getDefaultPriorityId(typeSet);
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
      Timemine.printStackTrace(exception);
    }
  }

  /** create issue
   * @param backendType backend type
   * @param projectId project id
   */
  Issue(Backend.Types backendType,
        Id            projectId
       )
  {
    this(backendType,
         projectId,
         new Id()  // issue id
        );
  }

  /** create issue
   * @param id database id
   */
  private Issue(Id issueId)
  {
    this(Backend.Types.NONE,
         new Id(),  // project id
         issueId
        );
  }

  /** clone issue
   * @return cloned issue
   */
  public Issue clone()
  {
    return new Issue(backendType,
                     id,
                     parentId,
                     shortName,
                     projectId,
                     trackerId,
                     epicName,
                     statusId,
                     statusTransitionId,
                     priorityId,
                     authorId,
                     summary,
                     description,
                     startDate,
                     dueDate,
                     assignedToMemberId,
                     doneRatio,
                     targetVersionId,
                     estimatedDuration,
                     createdOn,
                     updatedOn,
                     closedOn,
                     newProjectId
                    );
  }

  /** set local modified state
   * @param issue issue
   * @param modifiedState modified state
   */
  @Override
  public void setModified(Backend.ModifiedStates modifiedState)
    throws BackendException, InterruptedException
  {
    assert backendType != Backend.Types.NONE;

    {
      Backend.modifyLock();
    }
    try
    {
      // get added/updated time entry modifed
      IssueModified addedIssueModified = Backend.getLocalDatabase(new IssueModifiedDatabaseHandler(),
                                                                  Backend.condition("    issuesLocal.backendType=? "+
                                                                                    "AND issuesModified.id=? "+
                                                                                    "AND issuesModified.modifiedState=?",
                                                                                    backendType.ordinal(),
                                                                                    id.modified,
                                                                                    Backend.ModifiedStates.ADDED.ordinal()
                                                                                   )
                                                                 );
      IssueModified updatedIssueModified = Backend.getLocalDatabase(new IssueModifiedDatabaseHandler(),
                                                                    Backend.condition("    issuesLocal.backendType=? "+
                                                                                      "AND issuesModified.id=? "+
                                                                                      "AND issuesModified.modifiedState=?",
                                                                                      backendType.ordinal(),
                                                                                      id.modified,
                                                                                      Backend.ModifiedStates.UPDATED.ordinal()
                                                                                     )
                                                                   );

Dprintf.printStackTrace();
//Dprintf.dprintf("_"); System.exit(1);
      switch (modifiedState)
      {
        case ADDED:
          assert(addedIssueModified == null);
          assert(updatedIssueModified == null);

          // insert entry into local database
          id.local = Backend.insertLocalDatabase("issuesLocal",
                                                 new Backend.Values
                                                 (
                                                  "backendType",       backendType,

                                                  "shortName",         shortName,
                                                  "projectId",         projectId,
                                                  "trackerId",         trackerId,
                                                  "epicName",          epicName,
                                                  "statusId",          statusId,
                                                  "priorityId",        priorityId,
                                                  "authorId",          authorId,
                                                  "summary",           summary,
                                                  "description",       description,
                                                  "startDate",         startDate,
                                                  "dueDate",           dueDate,
                                                  "doneRatio",         doneRatio,
                                                  "assignedToMemberId",assignedToMemberId,
                                                  "targetVersionId",   targetVersionId,
                                                  "estimatedDuration", estimatedDuration,
                                                  "createdOn",         createdOn,
                                                  "updatedOn",         updatedOn,
                                                  "closedOn",          closedOn
                                                 )
                                                );

          // insert "added" entry into modified database
          id.modified = Backend.insertLocalDatabase("issuesModified",
                                                    new Backend.Values
                                                    (
                                                     "issueId",           id.local,
                                                     "modifiedState",     modifiedState.ordinal(),

                                                     "shortName",         shortName,
                                                     "projectId",         projectId,
                                                     "trackerId",         trackerId,
                                                     "epicName",          epicName,
                                                     "statusId",          statusId,
                                                     "priorityId",        priorityId,
                                                     "authorId",          authorId,
                                                     "summary",           summary,
                                                     "description",       description,
                                                     "startDate",         startDate,
                                                     "dueDate",           dueDate,
                                                     "doneRatio",         doneRatio,
                                                     "assignedToMemberId",assignedToMemberId,
                                                     "targetVersionId",   targetVersionId,
                                                     "estimatedDuration", estimatedDuration,
                                                     "createdOn",         createdOn,
                                                     "updatedOn",         updatedOn,
                                                     "closedOn",          closedOn
                                                    )
                                                   );
// TODO:
Dprintf.dprintf("inserted add-entry");
          break;
        case UPDATED:
          assert((addedIssueModified == null) || addedIssueModified.id.hasLocal());
          assert((addedIssueModified == null) || !addedIssueModified.id.hasBackend());
          assert((updatedIssueModified == null) || updatedIssueModified.id.hasLocal());
          assert((updatedIssueModified == null) || updatedIssueModified.id.hasBackend());

          Backend.insertLocalDatabase("issuesModified",
                                      new Backend.Values
                                      (
                                       "backendType",       backendType,
                                       "issueId",           id.local,
                                       "modifiedState",     modifiedState.ordinal(),

                                       "shortName",         shortName,
                                       "projectId",         projectId,
                                       "trackerId",         trackerId,
                                       "epicName",          epicName,
                                       "statusId",          statusId,
                                       "priorityId",        priorityId,
                                       "authorId",          authorId,
                                       "summary",           summary,
                                       "description",       description,
                                       "startDate",         startDate,
                                       "dueDate",           dueDate,
                                       "doneRatio",         doneRatio,
                                       "assignedToMemberId",assignedToMemberId,
                                       "targetVersionId",   targetVersionId,
                                       "estimatedDuration", estimatedDuration,
                                       "createdOn",         createdOn,
                                       "updatedOn",         updatedOn,
                                       "closedOn",          closedOn
                                      )
                                     );
          break;
        case DELETED:
          assert((addedIssueModified == null) || !addedIssueModified.id.hasLocal());
          assert((addedIssueModified == null) || !addedIssueModified.id.hasBackend());
          assert((updatedIssueModified == null) || updatedIssueModified.id.hasLocal());
          assert((updatedIssueModified == null) || updatedIssueModified.id.hasBackend());

          if      (updatedIssueModified != null)
          {
            // already updated -> change to "deleted" entry
            Backend.updateLocalDatabase("issuesModified",
                                        Backend.condition("id=?",updatedIssueModified.id.modified),
                                        new Backend.Values
                                        (
                                         "modifiedState",modifiedState.ordinal()
                                        )
                                       );
          }
          else if (addedIssueModified != null)
          {
            // already added -> delete entry
            Backend.removeLocalDatabase("issuesModified",
                                        addedIssueModified.id.modified
                                       );
          }
          else
          {
            // not added/updated -> insert "deleted" entry
            id.modified = Backend.insertLocalDatabase("issuesModified",
                                                      new Backend.Values
                                                      (
                                                       "issueId",           id.local,
                                                       "modifiedState",     modifiedState.ordinal(),

                                                       "shortName",         shortName,
                                                       "projectId",         projectId,
                                                       "trackerId",         trackerId,
                                                       "epicName",          epicName,
                                                       "statusId",          statusId,
                                                       "priorityId",        priorityId,
                                                       "authorId",          authorId,
                                                       "summary",           summary,
                                                       "description",       description,
                                                       "startDate",         startDate,
                                                       "dueDate",           dueDate,
                                                       "doneRatio",         doneRatio,
                                                       "assignedToMemberId",assignedToMemberId,
                                                       "targetVersionId",   targetVersionId,
                                                       "estimatedDuration", estimatedDuration,
                                                       "createdOn",         createdOn,
                                                       "updatedOn",         updatedOn,
                                                       "closedOn",          closedOn
                                                      )
                                                     );
          }
          break;
      }
      super.setModified(modifiedState);
    }
    finally
    {
      Backend.modifyUnlock();
    }
  }

  /** load issue from backend server
   * @param projectId project backend id
   * @param issueId issue backend id
   * @return origin issue fro server
   */
  public Issue load()
    throws BackendException
  {
    return Backend.backendDrivers.get(backendType).loadIssue(id);
  }

  /** check iff epic name is supported for trackers
   * @param backendType backend type
   * @param issueId issue id
   * @return true iff epic name is supported for trackers
   */
  public boolean isEpicNameSupported()
  {
    return Backend.isEpicNameSupported(backendType,id);
  }

  /** check iff edit progress on issues is supported
   * @param backendType backend type
   * @param projectId project id
   * @return true iff setting progress on issues is supported
   */
  public boolean isEditProgressSupported()
  {
    return Backend.isEditProgressSupported(backendType,id);
  }

  /** check if issue is still active (not closed/rejected)
   * @return true iff issue is active
   */
  public boolean isActive()
  {
    boolean active = false;

    try
    {
      Status status = Backend.getStatus(new Backend.TypeSet(backendType),statusId);
      active = Settings.issueActiveStatusSet.contains(status.name);
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return active;
  }

  /** get role id list
   * @param roleIdList role id list
   * @param names role names
   */
  public void getRoleIds(final IdList roleIdList, final Set<String> names)
    throws BackendException
  {
    Backend.getRoleIds(roleIdList,new Backend.TypeSet(backendType),projectId,names);
  }

  /** get sorted tracker list
   * @return tracker list
   */
  public TrackerList getTrackerList()
    throws BackendException
  {
    return Backend.getTrackerList(new Backend.TypeSet(backendType),projectId);
  }

  /** get tracker
   * @return tracker or null
   */
  public Tracker getTracker()
  {
    Tracker tracker = null;

    try
    {
      tracker = Backend.getTracker(new Backend.TypeSet(backendType),trackerId);
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return tracker;
  }

  /** check if tracker require epic name
   * @return true iff tracker required epic name
   */
  public boolean isTrackerRequireEpicName()
  {
    boolean trackerRequireEpicName = false;

    try
    {
      Tracker tracker = Backend.getTracker(new Backend.TypeSet(backendType),trackerId);
      if (tracker != null) trackerRequireEpicName = tracker.isRequireEpicName();
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return trackerRequireEpicName;
  }

  /** get sorted status transition list
   * @return transition list
   */
  public StatusTransitionList getStatusTransitionList()
    throws BackendException
  {
    return Backend.getStatusTransitionList(new Backend.TypeSet(backendType),id);
  }

  /** get status
   * @return status or null
   */
  public Status getStatus()
  {
    Status status = null;

    try
    {
      status = Backend.getStatus(new Backend.TypeSet(backendType),statusId);
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return status;
  }


  /** get pending status transition
   * @return status transition
   */
  public StatusTransition getStatusTransition()
    throws BackendException
  {
    return Backend.getStatusTransition(new Backend.TypeSet(backendType),statusTransitionId);
  }

  /** set pending status transition id and mark modified
   * @param statusTransitionId status transition id
   */
  public void setStatusTransitionId(Id statusTransitionId)
    throws BackendException, InterruptedException
  {
    if (statusTransitionId != this.statusTransitionId)
    {
      this.statusTransitionId = statusTransitionId;
      setModified(Backend.ModifiedStates.UPDATED);
    }
  }

  /** get sorted priority list
   * @return priority list
   */
  public PriorityList getPriorityList()
    throws BackendException
  {
    return Backend.getPriorityList(new Backend.TypeSet(backendType),projectId);
  }

  /** get priority
   * @return priority or null
   */
  public Priority getPriority()
  {
    Priority priority = null;

    try
    {
      priority = Backend.getPriority(new Backend.TypeSet(backendType),priorityId);
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return priority;
  }

  /** get sorted activity list
   * @return activity list
   */
  public ActivityList getActivityList()
    throws BackendException
  {
    return Backend.getActivityList(new Backend.TypeSet(backendType),id);
  }

  /** get activity
   * @param name activity name
   * @return activity or null
   */
  public Activity getActivity(String name)
  {
    Activity activity = null;

    try
    {
      activity = Backend.getActivity(new Backend.TypeSet(backendType),name);
    }
    catch (BackendException exception)
    {
      // ignored
    }

    return activity;
  }

  /** get project
   * @param defaultProject default project
   * @return project or defaultProject
   */
  public Project getProject(Project defaultProject)
  {
    Project project = defaultProject;

    try
    {
      project = Backend.getProject(new Backend.TypeSet(backendType),projectId);
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return project;
  }

  /** get project
   * @return project or null
   */
  public Project getProject()
  {
    Project project = null;

    try
    {
      project = Backend.getProject(new Backend.TypeSet(backendType),projectId);
    }
    catch (BackendException exception)
    {
      // ignored
    }

    return project;
  }

  /** get sorted member list
   * @param roleIds role filter ids or NULL
   * @return member list
   */
  public MemberList getMemberList(List<Id> roleIds)
    throws BackendException
  {
    return Backend.getMemberList(new Backend.TypeSet(backendType),projectId,roleIds);
  }

  /** get assigned-to
   * @return member or null
   */
  public Member getAssignedToMember()
  {
    Member member = null;

    try
    {
      member = Backend.getMember(new Backend.TypeSet(backendType),assignedToMemberId);
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return member;
  }

  /** check if assigned-to current user
   * @return true iff assigned-to current user
   */
  public boolean isAssignedToCurrentUser()
  {
    return Backend.isAssignedToCurrentUser(backendType,projectId,assignedToMemberId);
  }

  /** set done-ratio and mark modified
   * @param doneRatio done-ratio
   */
  public void setDoneRatio(int doneRatio)
    throws BackendException, InterruptedException
  {
Dprintf.dprintf("%d %d",doneRatio,this.doneRatio);
    if (doneRatio != this.doneRatio)
    {
      this.doneRatio = doneRatio;
      setModified(Backend.ModifiedStates.UPDATED);
    }
  }

  /** get sorted version list
   * @return version list
   */
  public VersionList getVersionList()
    throws BackendException
  {
    return Backend.getVersionList(new Backend.TypeSet(backendType),projectId);
  }

  /** get target version
   * @return version or defaultVersion
   */
  public Version getVersion(Version defaultVersion)
  {
    Version version = defaultVersion;

    try
    {
      version = Backend.getVersion(new Backend.TypeSet(backendType),targetVersionId);
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return version;
  }

  /** get target version
   * @return version
   */
  public Version getVersion()
    throws BackendException
  {
    return Backend.getVersion(new Backend.TypeSet(backendType),targetVersionId);
  }

  /** get sorted issue list
   * @param activeOnly true for active projects, False for all projets
   * @return issue list
   */
  public IssueList getIssueList(boolean activeOnly)
    throws BackendException
  {
    return Backend.getIssueList(new Backend.TypeSet(backendType),projectId,activeOnly);
  }

  /** get sorted issue list
   * @return issue list
   */
  public IssueList getIssueList()
    throws BackendException
  {
    return getIssueList(false);
  }

  /** get sorted issue list
   * @param spentOn spent-on date
   * @return issue list
   */
  public TimeEntryList getTimeEntryList(SpentOn spentOn)
    throws BackendException
  {
    return Backend.getTimeEntryList(new Backend.TypeSet(backendType),projectId,id,spentOn);
  }

  /** get sorted issue list
   * @return issue list
   */
  public TimeEntryList getTimeEntryList()
    throws BackendException
  {
    return getTimeEntryList((SpentOn)null);
  }

  /** get duration sum of time entries
   * @param userId user id or Id.ANY
   * @return duration sum of time entries or 0 if data still not available [min]
   */
  public long getDurationSum(Id userId)
  {
    return Backend.getDurationSum(new Backend.TypeSet(backendType),projectId,id,userId);
  }

  /** set spent time
   * @param hours spent hours
   * @param minutes spent minutes
   */
  public void setEstimatedDuration(int hours, int minutes)
  {
    this.estimatedDuration = Backend.toDuration(hours,minutes);
  }

  /** get estimated hours fraction
   * @return hours fraction
   */
  public int getEstimatedHourFraction()
  {
    return (int)(estimatedDuration/60);
  }

  /** get estimated minutes fraction
   * @return minutes fraction
   */
  public int getEstimatedMinuteFraction()
  {
    return (int)(estimatedDuration%60);
  }

  /** get database parameters
   * @returns hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("id",id);
//      databaseParameters.put("name",name);

    return databaseParameters;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "Issue { id="+id+", parentId="+parentId+", projectId="+projectId+", trackerId="+trackerId+", epic name="+((epicName != null) ? epicName : "none")+", priorityId="+priorityId+", statusId="+statusId+", assignedToMemberId="+assignedToMemberId+", doneRatio="+doneRatio+", summary="+summary+", newProjectId="+newProjectId+" }";
  }
}

/** issue list
 */
class IssueList extends ArrayList<Issue>
{
  final static IssueList EMPTY = new IssueList();
}

/** issue database handler
 */
class IssueDatabaseHandler extends DataBaseHandler<Issue>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM issues"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public Issue get(ValueSet valueSet)
    throws SQLException
  {
    return new Issue(valueSet.getEnum  ("backendType",Backend.Types.class),
                     valueSet.getId    ("id","backendId","modifiedId"),
                     valueSet.getId    ("parentId"),
                     valueSet.getString("shortName"),
                     valueSet.getId    ("projectId"),
                     valueSet.getId    ("trackerId"),
                     valueSet.getString("epicName"),
                     valueSet.getId    ("statusId"),
                     valueSet.getId    ("statusTransitionId"),
                     valueSet.getId    ("priorityId"),
                     valueSet.getId    ("authorId"),
                     valueSet.getString("summary"),
                     valueSet.getString("description"),
                     valueSet.getDate  ("startDate"),
                     valueSet.getDate  ("dueDate"),
                     valueSet.getId    ("assignedToMemberId"),
                     valueSet.getInt   ("doneRatio"),
                     valueSet.getId    ("targetVersionId"),
                     valueSet.getLong  ("estimatedDuration"),
                     valueSet.getDate  ("createdOn"),
                     valueSet.getDate  ("updatedOn"),
                     valueSet.getDate  ("closedOn"),
new Id()//                     new Id(resultSet.getInt("newProjectId"))
                    );
  }
}

/** modified issue
 */
class IssueModified extends Issue
{
  public int                    modifiedId;
  public Backend.ModifiedStates modifiedState;

  /** create modified issue
   * @param backendType backend type
   * @param id database id
   * @param parentId parent issue id or Id.NONE
   * @param shortName short name or null
   * @param projectId project id
   * @param trackerId tracker id
   * @param epicName epic name or null
   * @param statusId status id
   * @param statusTransitionId status transition id or Id.NONE
   * @param priorityId priority id
   * @param authorId author user id
   * @param summary summary text
   * @param description description text
   * @param startDate start date
   * @param dueDate due date
   * @param assignedToMemberId assigned to member id or Id.ANY
   * @param doneRatio done ration
   * @param targetVersionId target version id or Id.ANY
   * @param estimatedDuration estimated duration [min]
   * @param createdOn created-on date
   * @param updatedOn update-on date
   * @param closedOn closed-on date
   * @param newProjectId new project id or Id.NONE
   * @param modifiedId modified id
   * @param modifiedState modified state
   */
  IssueModified(Backend.Types          backendType,
                Id                     id,
                Id                     parentId,
                String                 shortName,
                Id                     projectId,
                Id                     trackerId,
                String                 epicName,
                Id                     statusId,
                Id                     statusTransitionId,
                Id                     priorityId,
                Id                     authorId,
                String                 summary,
                String                 description,
                Date                   startDate,
                Date                   dueDate,
                Id                     assignedToMemberId,
                int                    doneRatio,
                Id                     targetVersionId,
                long                   estimatedDuration,
                Date                   createdOn,
                Date                   updatedOn,
                Date                   closedOn,
                Id                     newProjectId,
                int                    modifiedId,
                Backend.ModifiedStates modifiedState
               )
  {
    super(backendType,
          id,
          parentId,
          shortName,
          projectId,
          trackerId,
          epicName,
          statusId,
          statusTransitionId,
          priorityId,
          authorId,
          summary,
          description,
          startDate,
          dueDate,
          assignedToMemberId,
          doneRatio,
          targetVersionId,
          estimatedDuration,
          createdOn,
          updatedOn,
          closedOn,
          newProjectId
         );
    this.modifiedId    = modifiedId;
    this.modifiedState = modifiedState;
  }

  /** compare with other modified issue
   * @return -1/0+1 iff lower/equals/greater
   */
  public int compareTo(IssueModified otherIssueModified)
  {
    int result = 0;
    if (this != otherIssueModified)
    {
      if      (backendType != otherIssueModified.backendType)
      {
        result = -1;
      }
      else if (modifiedState != otherIssueModified.modifiedState)
      {
        result = -1;
      }
      else
      {
        result = id.compareTo(otherIssueModified.id);
      }
    }

    return result;
  }

  /** done modification
   */
  public void doneModified()
    throws BackendException, InterruptedException
  {
    {
      Backend.modifyLock();
    }
    try
    {
      Backend.removeLocalDatabase("issuesModified",
                                  modifiedId
                                 );
    }
    finally
    {
      Backend.modifyUnlock();
    }
  }

  /** update local database
   */
  protected void updateLocalDatabase()
    throws BackendException
  {
    assert(id.hasLocal());

    Backend.updateLocalDatabase("issuesLocal",
                                Backend.condition("backendType=? AND id=?",backendType,id.local),
                                new Backend.Values
                                (
                                  "parentId",          parentId,
                                  "shortName",         shortName,
                                  "projectId",         projectId,
                                  "trackerId",         trackerId,
                                  "epicName",          epicName,
                                  "statusId",          statusId,
                                  "statusTransitionId",statusTransitionId,
                                  "priorityId",        priorityId,
                                  "authorId",          authorId,
                                  "summary",           summary,
                                  "description",       description,
                                  "startDate",         startDate,
                                  "dueDate",           dueDate,
                                  "assignedToMemberId",assignedToMemberId,
                                  "doneRatio",         doneRatio,
                                  "targetVersionId",   targetVersionId,
                                  "estimatedDuration", estimatedDuration,
                                  "createdOn",         createdOn,
                                  "updatedOn",         updatedOn,
                                  "closedOn",          closedOn,
                                  "newProjectId",      newProjectId
                                )
                               );
  }

  /** remove from local database
   */
  protected void removeLocalDatabase()
    throws BackendException, InterruptedException
  {
    // Note: set as "deleted" only; issue is removed from database on synchronisation
    setModified(Backend.ModifiedStates.DELETED);
  }

  /** revert in local database
   */
  protected void revertLocalDatabase()
    throws BackendException, InterruptedException
  {
Dprintf.printStackTrace();
    clearModified();
Dprintf.dprintf("_\n"); System.exit(1);
  }

  /** store issue on backend server
   */
  public void store()
    throws BackendException
  {
    assert(id.hasLocal());

Dprintf.dprintf("_");
    Backend.backendDrivers.get(backendType).storeIssue(this);
Dprintf.dprintf("_");
    updateLocalDatabase();
Dprintf.dprintf("_");
  }

  /** update issue on backend server
   * @param originIssue origin issue
   */
  public void update(Issue originIssue)
    throws BackendException
  {
    assert id.hasLocal();
    assert id.hasBackend();

    Backend.backendDrivers.get(backendType).updateIssue(this,originIssue);
    updateLocalDatabase();
  }

  /** delete issue on backend server
   * @param timeEntry time entry to delete
   */
  public void delete()
    throws BackendException
  {
Dprintf.dprintf("delete issue: %s",this);
    if (!id.isNone()) Backend.backendDrivers.get(backendType).deleteIssue(this);
    Backend.removeLocalDatabase("issuesLocal",
                                id.local
                               );
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "IssueModified { id="+id+", modified="+modifiedState+", parentId="+parentId+", projectId="+projectId+", trackerId="+trackerId+", epic name="+((epicName != null) ? epicName : "none")+", priorityId="+priorityId+", statusId="+statusId+", assignedToMemberId="+assignedToMemberId+", doneRatio="+doneRatio+", summary="+summary+", newProjectId="+newProjectId+" }";
  }

  /** clear all local issue modifications
   * @param issue issue
   */
  private void clearModified()
    throws BackendException, InterruptedException
  {
    {
      Backend.modifyLock();
    }
    try
    {
      Backend.removeAllLocalDatabase("issuesModified",
                                     Backend.condition("issueId=?",
                                                       id.local
                                                      )
                                    );
    }
    finally
    {
      Backend.modifyUnlock();
    }
  }
};

/** modified issue database handler
 */
class IssueModifiedDatabaseHandler extends DataBaseHandler<IssueModified>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT issuesLocal.backendType,"+
           "        issuesLocal.id,"+
           "        issuesLocal.backendId,"+
           "        issuesLocal.parentId,"+
           "        issuesModified.shortName,"+
           "        issuesModified.projectId,"+
           "        issuesModified.trackerId,"+
           "        issuesModified.epicName,"+
           "        issuesModified.statusId,"+
           "        issuesModified.statusTransitionId,"+
           "        issuesModified.priorityId,"+
           "        issuesModified.authorId,"+
           "        issuesModified.summary,"+
           "        issuesModified.description,"+
           "        issuesModified.startDate,"+
           "        issuesModified.dueDate,"+
           "        issuesModified.assignedToMemberId,"+
           "        issuesModified.doneRatio,"+
           "        issuesModified.targetVersionId,"+
           "        issuesModified.estimatedDuration,"+
           "        issuesModified.createdOn,"+
           "        issuesModified.updatedOn,"+
           "        issuesModified.closedOn,"+
           "        issuesLocal.lastUpdateTimestamp,"+
           "        issuesModified.newProjectId, "+
           "        issuesModified.id AS modifiedId,"+
           "        issuesModified.modifiedState "+
           " FROM issuesModified "+
           "   LEFT JOIN issuesLocal ON issuesLocal.id=issuesModified.issueId"+
           " WHERE issuesLocal.id NOT NULL"+
           ((additionalFilter != null) ? " AND "+additionalFilter : "");
  }

  @Override
  public IssueModified get(ValueSet valueSet)
    throws SQLException
  {
    return new IssueModified(valueSet.getEnum  ("backendType",Backend.Types.class),
                             valueSet.getId    ("id","backendId","modifiedId"),
                             valueSet.getId    ("parentId"),
                             valueSet.getString("shortName"),
                             valueSet.getId    ("projectId"),
                             valueSet.getId    ("trackerId"),
                             valueSet.getString("epicName"),
                             valueSet.getId    ("statusId"),
                             valueSet.getId    ("statusTransitionId"),
                             valueSet.getId    ("priorityId"),
                             valueSet.getId    ("authorId"),
                             valueSet.getString("summary"),
                             valueSet.getString("description"),
                             valueSet.getDate  ("startDate"),
                             valueSet.getDate  ("dueDate"),
                             valueSet.getId    ("assignedToMemberId"),
                             valueSet.getInt   ("doneRatio"),
                             valueSet.getId    ("targetVersionId"),
                             valueSet.getLong  ("estimatedDuration"),
                             valueSet.getDate  ("createdOn"),
                             valueSet.getDate  ("updatedOn"),
                             valueSet.getDate  ("closedOn"),
                             valueSet.getId    ("newProjectId"),
                             valueSet.getInt   ("modifiedId"),
                             valueSet.getEnum  ("modifiedState",Backend.ModifiedStates.class)
                            );
  }
}

/* end of file */
