/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** member database id
 */
class MemberId extends Id
{
  /** create id
   * @param backend backend id
   */
  MemberId(int backend)
  {
    super(backend);
  }
}

/** member
 */
class Member extends Entity
{
  public final static Member ANY  = new Member(new Id(Id.ANY));

  public final Id     projectId;
  public final Id     userId;
  public String       firstName;        // first/given name
  public String       lastName;         // last name
  public final String name;             // user/login name
  public final Id     roleId;

  /** create Member
   * @param backendType backend type
   * @param id id
   * @param projectId project id
   * @param firstName first name
   * @param lastName last name
   * @param name name
   * @param roleId role id
   */
  Member(Backend.Types backendType,
         Id            id,
         Id            projectId,
         Id            userId,
         String        firstName,
         String        lastName,
         String        name,
         Id            roleId
        )
  {
    super(backendType,id);
    assert backendType != Backend.Types.NONE;
    assert !id.isNone();
    assert !id.isAny();
    assert(firstName != null);
    assert(lastName != null);
    assert(name != null);

    this.projectId = projectId;
    this.userId    = userId;
    this.firstName = firstName;
    this.lastName  = lastName;
    this.name      = name;
    this.roleId    = roleId;
  }

  /** create Member
   * @param id id
   */
  private Member(Id id)
  {
    super(Backend.Types.NONE,id);

    this.projectId = new Id();
    this.userId    = new Id();
    this.firstName = "";
    this.lastName  = "";
    this.name      = "";
    this.roleId    = new Id();
  }

  /** get database parameters
   * @returns hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("id",id);
    databaseParameters.put("userId",userId);
    databaseParameters.put("firstName",firstName);
    databaseParameters.put("lastName",lastName);
    databaseParameters.put("name",name);
    databaseParameters.put("roleId",roleId);

    return databaseParameters;
  }

  /** get full user name
   * @return name
   */
  public String getName()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append(lastName);
    if (!firstName.isEmpty())
    {
      buffer.append(", ");
      buffer.append(firstName);
    }

    return buffer.toString();
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "Member { id="+id+", userId="+userId+", firstName="+firstName+", lastName="+lastName+", roleId="+roleId+" }";
  }
}

/** member list
 */
class MemberList extends ArrayList<Member>
{
  final static MemberList EMPTY = new MemberList();
}

/** member database handler
 */
class MemberDatabaseHandler extends DataBaseHandler<Member>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM members"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public Member get(ValueSet valueSet)
    throws SQLException
  {
    return new Member(valueSet.getEnum  ("backendType",Backend.Types.class),
                      valueSet.getId    ("id","backendId"),
                      valueSet.getId    ("projectId"),
                      valueSet.getId    ("userId"),
                      valueSet.getString("firstName"),
                      valueSet.getString("lastName"),
                      valueSet.getString("name"),
                      valueSet.getId    ("roleId")
                     );
  }
}

/* end of file */
