/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** role database id
 */
class RoleId extends Id
{
  /** create id
   * @param backend backend id
   */
  RoleId(int backend)
  {
    super(backend);
  }
}

/** Role
 */
class Role extends Entity
{
  public final Id     projectId;
  public final String name;
  public String       description;

  /** create role
   * @param backendType backend type
   * @param id database id
   * @param projectId project id or Id.ANY
   * @param name name
   * @param description description
   */
  Role(Backend.Types backendType,
       Id            id,
       Id            projectId,
       String        name,
       String        description
      )
  {
    super(backendType,id);
    assert backendType != Backend.Types.NONE;
    assert !id.isNone();
    assert !id.isAny();
    assert(name != null);
    assert(description != null);

    this.projectId   = projectId;
    this.name        = name;
    this.description = description;
  }

  /** create role
   * @param id database id
   */
  private Role(Id id)
  {
    super(Backend.Types.NONE,id);

    this.projectId   = new Id();
    this.name        = "";
    this.description = "";
  }

  /** create role
   */
  private Role()
  {
    this(new Id());
  }

  /** get database parameters
   * @returns hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("id",id);
    databaseParameters.put("projectId",projectId);
    databaseParameters.put("name",name);
    databaseParameters.put("description",description);

    return databaseParameters;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "Role { id="+id+", projectId="+projectId+", name="+name+", description="+description+" }";
  }
}

/** role list
 */
class RoleList extends ArrayList<Role>
{
  final static RoleList EMPTY = new RoleList();
}

/** role database handler
 */
class RoleDatabaseHandler extends DataBaseHandler<Role>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM roles"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public Role get(ValueSet valueSet)
    throws SQLException
  {
    return new Role(valueSet.getEnum  ("backendType",Backend.Types.class),
                    valueSet.getId    ("id","backendId"),
                    valueSet.getId    ("projectId"),
                    valueSet.getString("name"),
                    valueSet.getString("description")
                   );
  }
}

/* end of file */
