/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** status database id
 */
class StatusId extends Id
{
  /** create id
   * @param backend backend id
   */
  StatusId(int backend)
  {
    super(backend);
  }
}

/** status
 */
class Status extends Entity
{
  public final Id      projectId;
  public final String  name;
  public String        description;
  public final boolean isDefault;

  /** create status
   * @param backendType backend type
   * @param id database id
   * @param projectId project id or Id.ANY
   * @param name name
   * @param description description
   * @param isDefault true iff default
   */
  Status(Backend.Types backendType,
         Id            id,
         Id            projectId,
         String        name,
         String        description,
         boolean       isDefault
        )
  {
    super(backendType,id);
    assert backendType != Backend.Types.NONE;
    assert !id.isNone();
    assert !id.isAny();
    assert(name != null);
    assert(description != null);

    this.projectId   = projectId;
    this.name        = name;
    this.description = description;
    this.isDefault   = isDefault;
  }

  /** create status
   * @param backendType backend type
   * @param id backend id
   * @param name name
   * @param description description
   */
  Status(Backend.Types backendType,
         Id            id,
         Id            projectId,
         String        name,
         String        description
        )
  {
    this(backendType,
         id,
         projectId,
         name,
         description,
         false
        );
  }

  /** create status
   * @param id backend id
   */
  private Status(Id id)
  {
    super(Backend.Types.NONE,id);

    this.projectId   = new Id();
    this.name        = "";
    this.description = "";
    this.isDefault   = false;
  }

  /** get database parameters
   * @returns hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("id",id);
    databaseParameters.put("projectId",projectId);
    databaseParameters.put("name",name);
    databaseParameters.put("description",description);
    databaseParameters.put("isDefault",isDefault);

    return databaseParameters;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "Status { id="+id+", projectId="+projectId+", name="+name+", description="+description+", isDefault="+isDefault+" }";
  }
}

/** status list
 */
class StatusList extends ArrayList<Status>
{
  final static StatusList EMPTY = new StatusList();
}

/** status database handler
 */
class StatusDatabaseHandler extends DataBaseHandler<Status>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM status"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public Status get(ValueSet valueSet)
    throws SQLException
  {
    return new Status(valueSet.getEnum   ("backendType",Backend.Types.class),
                      valueSet.getId     ("id","backendId"),
                      valueSet.getId     ("projectId"),
                      valueSet.getString ("name"),
                      valueSet.getString ("description"),
                      valueSet.getBoolean("isDefault")
                     );
  }
}

/* end of file */
