/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** status transition database id
 */
class StatusTransitionId extends Id
{
  /** create id
   * @param backend backend id
   */
  StatusTransitionId(int backend)
  {
    super(backend);
  }
}

/** status transition
 */
class StatusTransition extends Entity
{
  public final static StatusTransition NONE = new StatusTransition();

  public final Id     issueId;
  public final String name;
  public String       description;
  public final Id     statusId;

  /** create status transition
   * @param backendType backend type
   * @param id database id
   * @param name name
   * @param description description
   * @param statusId status id
   */
  StatusTransition(Backend.Types backendType,
                   Id            id,
                   Id            issueId,
                   String        name,
                   String        description,
                   Id            statusId
                  )
  {
    super(backendType,id);
    assert backendType != Backend.Types.NONE;
    assert !id.isNone();
    assert !id.isAny();
    assert(name != null);
    assert(description != null);

    this.issueId     = issueId;
    this.name        = name;
    this.description = description;
    this.statusId    = statusId;
  }

  /** create status transition
   * @param id database id
   */
  private StatusTransition(Id id)
  {
    super(Backend.Types.NONE,id);

    this.issueId     = new Id();
    this.name        = "";
    this.description = "";
    this.statusId    = new Id();
  }

  /** create status transition
   */
  private StatusTransition()
  {
    this(new Id());
  }

  /** get database parameters
   * @returns hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("id",id);
    databaseParameters.put("issueId",issueId);
    databaseParameters.put("name",name);
    databaseParameters.put("description",description);
    databaseParameters.put("statusId",statusId);

    return databaseParameters;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "Status transition { id="+id+", issueId="+issueId+", name="+name+", description="+description+", statusId="+statusId+" }";
  }
}

/** status transition list
 */
class StatusTransitionList extends ArrayList<StatusTransition>
{
  final static StatusTransitionList EMPTY = new StatusTransitionList();
}

/** status transition database handler: map status -> status transition
 */
class StatusTransitionDatabaseHandler extends DataBaseHandler<StatusTransition>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM statusTransitions"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public StatusTransition get(ValueSet valueSet)
    throws SQLException
  {
    return new StatusTransition(valueSet.getEnum  ("backendType",Backend.Types.class),
                                valueSet.getId    ("id","backendId"),
                                valueSet.getId    ("issueId"),
                                valueSet.getString("name"),
                                valueSet.getString("description"),
                                valueSet.getId    ("statusId")
                               );
  }
}

/* end of file */
