/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.text.SimpleDateFormat;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** spent-on date
 */
class SpentOn extends Object implements Cloneable
{
  private final static SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd");

  private Calendar calendar;

  /** create spent-on date
   * @param year year
   * @param month month
   * @param day day
   */
  SpentOn(int year, int month, int day)
  {
    this();
    calendar.set(Calendar.YEAR,        year );
    calendar.set(Calendar.MONTH,       month);
    calendar.set(Calendar.DAY_OF_MONTH,day  );
  }

  /** create spent-on date
   * @param date date
   */
  SpentOn(Date date)
  {
    this();
    Calendar fromCalendar = Calendar.getInstance(); fromCalendar.setTime(date);
    calendar.set(Calendar.YEAR,        fromCalendar.get(Calendar.YEAR        ));
    calendar.set(Calendar.MONTH,       fromCalendar.get(Calendar.MONTH       ));
    calendar.set(Calendar.DAY_OF_MONTH,fromCalendar.get(Calendar.DAY_OF_MONTH));
  }

  /** create spent-on date
   * @param dayOffset day offset from today
   */
  SpentOn(int dayOffset)
  {
    this();
    Calendar fromCalendar = Calendar.getInstance();
    calendar.add(Calendar.DAY_OF_MONTH,dayOffset);
  }

  /** create spent-on date
   */
  SpentOn()
  {
    this.calendar = Calendar.getInstance();
  }

  /** clone object
   * @return cloned object
   */
  public SpentOn clone()
  {
    return new SpentOn(calendar.getTime());
  }

  /** get spent-on today
   * @return spent-on today
   */
  public static SpentOn today()
  {
    return new SpentOn();
  }

  /** add/subtract days
   * @param days number of days
   */
  public void addDays(int days)
  {
    calendar.add(Calendar.DAY_OF_MONTH,days);
  }

  /** get year
   * @return year
   */
  int getYear()
  {
    return calendar.get(Calendar.YEAR);
  }

  /** get month
   * @return month
   */
  int getMonth()
  {
    return calendar.get(Calendar.MONTH);
  }

  /** get day
   * @return day
   */
  int getDay()
  {
    return calendar.get(Calendar.DAY_OF_MONTH);
  }

  /** get weekday
   * @return weekday
   */
  int getWeekday()
  {
    return calendar.get(Calendar.DAY_OF_WEEK);
  }

  /** get date
   * @param addDays days to add
   * @return date
   */
  Date getDate(int addDays)
  {
    Calendar date = (Calendar)calendar.clone();
    date.add(Calendar.DATE,addDays);
    return date.getTime();
  }

  /** get date
   * @return date
   */
  Date getDate()
  {
    Calendar date = Calendar.getInstance();
    date.set(calendar.get(Calendar.YEAR),calendar.get(Calendar.MONTH),calendar.get(Calendar.DAY_OF_MONTH),0,0,0);
    date.set(Calendar.MILLISECOND,0);
    return date.getTime();
  }

  /** get time from Epoch
   * @return time [ms]
   */
  long getTime()
  {
    return calendar.getTimeInMillis();
  }

  /** check if spent-on date is today
   * @return true iff today
   */
  public boolean isToday()
  {
    Calendar todayCalendar = Calendar.getInstance();

    return    (calendar.get(Calendar.YEAR        ) == todayCalendar.get(Calendar.YEAR        ))
           && (calendar.get(Calendar.MONTH       ) == todayCalendar.get(Calendar.MONTH       ))
           && (calendar.get(Calendar.DAY_OF_MONTH) == todayCalendar.get(Calendar.DAY_OF_MONTH));
  }

  /** check if spent-in dates are equal
   * @param spentOn spent-on date
   * @return true iff equal
   */
  public boolean equals(SpentOn spentOn)
  {
    return    (calendar.get(Calendar.YEAR) == spentOn.calendar.get(Calendar.YEAR))
           && (calendar.get(Calendar.MONTH) == spentOn.calendar.get(Calendar.MONTH))
           && (calendar.get(Calendar.DAY_OF_MONTH) == spentOn.calendar.get(Calendar.DAY_OF_MONTH));
  }

  /** check if spent-on is before date
   * @param spentOn spent-on date
   * @return true iff before date
   */
  public boolean isBefore(SpentOn spentOn)
  {
    return    (   (calendar.get(Calendar.YEAR) < spentOn.calendar.get(Calendar.YEAR))
              )
           || (   (calendar.get(Calendar.YEAR) == spentOn.calendar.get(Calendar.YEAR))
               && (calendar.get(Calendar.MONTH) < spentOn.calendar.get(Calendar.MONTH))
              )
           || (   (calendar.get(Calendar.YEAR) == spentOn.calendar.get(Calendar.YEAR))
               && (calendar.get(Calendar.MONTH) == spentOn.calendar.get(Calendar.MONTH))
               && (calendar.get(Calendar.DAY_OF_MONTH) < spentOn.calendar.get(Calendar.DAY_OF_MONTH))
              );
  }

  /** check if spent-on is before date
   * @param spentOn spent-on date
   * @return true iff before date
   */
  public boolean isBefore(Date date)
  {
    return isBefore(new SpentOn(date));
  }

  /** check if spent-on is after date
   * @param spentOn spent-on date
   * @return true iff after date
   */
  public boolean isAfter(SpentOn spentOn)
  {
    return    (   (calendar.get(Calendar.YEAR) > spentOn.calendar.get(Calendar.YEAR))
              )
           || (   (calendar.get(Calendar.YEAR) == spentOn.calendar.get(Calendar.YEAR))
               && (calendar.get(Calendar.MONTH) > spentOn.calendar.get(Calendar.MONTH))
              )
           || (   (calendar.get(Calendar.YEAR) == spentOn.calendar.get(Calendar.YEAR))
               && (calendar.get(Calendar.MONTH) == spentOn.calendar.get(Calendar.MONTH))
               && (calendar.get(Calendar.DAY_OF_MONTH) > spentOn.calendar.get(Calendar.DAY_OF_MONTH))
              );
  }

  /** check if spent-on is after date
   * @param spentOn spent-on date
   * @return true iff after date
   */
  public boolean isAfter(Date date)
  {
    return isAfter(new SpentOn(date));
  }

  /** compare spent-on dates
   * @param spentOn spent-on date
   * @return -1,0,1 if before/equal/after
   */
  public int compareTo(SpentOn spentOn)
  {
    return calendar.compareTo(spentOn.calendar);
  }

  /** get hash code: day+month+year only
   * @return hash code
   */
  public int hashCode()
  {
    return   ((calendar.get(Calendar.DAY_OF_MONTH) & 0x0000001F) << 0)
           | ((calendar.get(Calendar.MONTH       ) & 0x0000000F) << 5)
           | ((calendar.get(Calendar.YEAR        ) & 0x007FFFFF) << 9);
  }

  /** convert to string
   * @return string
   */
  public String toString(SimpleDateFormat simpleDateFormat)
  {
    return simpleDateFormat.format(calendar.getTime());
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return toString(DATE_FORMAT);
  }
}

/** time entry database id
 */
class TimeEntryId extends Id
{
  /** create id
   * @param backend backend id
   */
  TimeEntryId(int backend)
  {
    super(backend);
  }
}

/** time entry
 */
class TimeEntry extends Entity implements Cloneable
{
//TODO: required?
  public Id      projectId;
  public Id      issueId;
//TODO: required?
  public Id      userId;
  public Id      activityId;
  public long    duration;         // [min]
  public String  comments;
  public SpentOn spentOn;
  public Date    createdOn;
  public Date    updatedOn;

  /** create time entry
   * @param backendType backend type
   * @param timeEntryId time entry id
   * @param modifyState modified state
   * @param projectId project id
   * @param issueId issue id
   * @param userId user id
   * @param activityId activity id
   * @param duration spent duration [min]
   * @param comments comments
   * @param spentOn spent-on date
   * @param createdOn created-on date
   * @param updatedOn update-on date
   */
  TimeEntry(Backend.Types backendType,
            Id            timeEntryId,
            Id            projectId,
            Id            issueId,
            Id            userId,
            Id            activityId,
            long          duration,
            String        comments,
            SpentOn       spentOn,
            Date          createdOn,
            Date          updatedOn
           )
  {
    super(backendType,timeEntryId);

    assert backendType != Backend.Types.NONE;
    assert projectId.isSet();
    assert comments != null;

    this.projectId     = projectId;
    this.issueId       = issueId;
    this.userId        = userId;
    this.activityId    = activityId;
    this.duration      = duration;
    this.comments      = comments;
    this.spentOn       = spentOn;
    this.createdOn     = createdOn;
    this.updatedOn     = updatedOn;
  }

  /** create time entry
   * @param backendType backend type
   * @param timeEntryId time entry id
   * @param projectId project id
   * @param issueId issue id
   * @param activityId activity id
   * @param duration spent duration [min]
   * @param comments comments
   * @param spentOn spent-on date
   */
  TimeEntry(Backend.Types backendType,
            Id            timeEntryId,
            Id            projectId,
            Id            issueId,
            Id            activityId,
            long          duration,
            String        comments,
            SpentOn       spentOn
           )
  {
    this(backendType,
         timeEntryId,
         projectId,
         issueId,
//TODO
new Id(),//           ownUser.id,
         activityId,
         duration,
         comments,
         spentOn,
         new Date(),
         new Date()
        );
  }

  /** create time entry
   * @param backendType backend type
   * @param timeEntryId time entry id
   * @param projectId project id
   * @param issueId issue id
   * @param activityId activity id
   * @param duration spent duration [min]
   * @param comments comments
   */
  TimeEntry(Backend.Types backendType,
            Id            timeEntryId,
            Id            projectId,
            Id            issueId
           )
  {
    this(backendType,
         timeEntryId,
         projectId,
         issueId,
         new Id(),
         0,
         "",
         SpentOn.today()
        );
  }

  /** clone time entry
   * @return cloned time entry
   */
  public TimeEntry clone()
  {
    return new TimeEntry(backendType,
                         id,
                         projectId,
                         issueId,
                         userId,
                         activityId,
                         duration,
                         comments,
                         spentOn,
                         createdOn,
                         updatedOn
                        );
  }

  /** load time entry from backend server
   * @param timeEntry time entry
   * @return orgin time entry from server
   */
  public TimeEntry load()
    throws BackendException
  {
    return Backend.backendDrivers.get(backendType).loadTimeEntry(projectId,issueId,id);
  }

  /** set local modified state
   * @param modifiedState modified state
   */
  public void setModified(Backend.ModifiedStates modifiedState)
    throws BackendException, InterruptedException
  {
    assert backendType != Backend.Types.NONE;

    {
      Backend.modifyLock();
    }
    try
    {
//Dprintf.dprintf("id=%s: %s",id,modifiedState);Dprintf.printStackTrace();

      // get added/updated time entry modifed
      TimeEntryModified addedTimeEntryModified = Backend.getLocalDatabase(new TimeEntryModifiedDatabaseHandler(),
                                                                          Backend.condition("    timeEntriesLocal.backendType=? "+
                                                                                            "AND timeEntriesModified.id=? "+
                                                                                            "AND timeEntriesModified.modifiedState=?",
                                                                                            backendType.ordinal(),
                                                                                            id.modified,
                                                                                            Backend.ModifiedStates.ADDED.ordinal()
                                                                                           )
                                                                         );
      TimeEntryModified updatedTimeEntryModified = Backend.getLocalDatabase(new TimeEntryModifiedDatabaseHandler(),
                                                                            Backend.condition("    timeEntriesLocal.backendType=? "+
                                                                                              "AND timeEntriesModified.id=? "+
                                                                                              "AND timeEntriesModified.modifiedState=?",
                                                                                              backendType.ordinal(),
                                                                                              id.modified,
                                                                                              Backend.ModifiedStates.UPDATED.ordinal()
                                                                                             )
                                                                           );

Dprintf.dprintf("id=%s modifiedState=%s",id,modifiedState);
Dprintf.dprintf("addedTimeEntryModified=%s",addedTimeEntryModified);
Dprintf.dprintf("updatedTimeEntryModified=%s",updatedTimeEntryModified);
      switch (modifiedState)
      {
        case ADDED:
          assert(addedTimeEntryModified == null);
          assert(updatedTimeEntryModified == null);

          // insert entry into local database
          id.local = Backend.insertLocalDatabase("timeEntriesLocal",
                                                 new Backend.Values
                                                 (
                                                  "backendType",  backendType,

                                                  "projectId",    projectId,
                                                  "issueId",      issueId,
                                                  "userId",       userId,
                                                  "activityId",   activityId,
                                                  "duration",     duration,
                                                  "comments",     comments,
                                                  "spentOn",      spentOn,
                                                  "createdOn",    createdOn,
                                                  "updatedOn",    updatedOn
                                                 )
                                                );

          // insert "added" entry into modified database
          id.modified = Backend.insertLocalDatabase("timeEntriesModified",
                                                    new Backend.Values
                                                    (
                                                     "timeEntryId",  id.local,
                                                     "modifiedState",modifiedState.ordinal(),

                                                     "projectId",    projectId,
                                                     "issueId",      issueId,
                                                     "userId",       userId,
                                                     "activityId",   activityId,
                                                     "duration",     duration,
                                                     "comments",     comments,
                                                     "spentOn",      spentOn,
                                                     "createdOn",    createdOn,
                                                     "updatedOn",    updatedOn
                                                    )
                                                   );

// TODO: remove timeEntries table and rename timeEntriesLocal -> timeEntries
Dprintf.dprintf("inserted add-entry");
          break;
        case UPDATED:
          assert((addedTimeEntryModified == null) || addedTimeEntryModified.id.hasLocal());
          assert((addedTimeEntryModified == null) || !addedTimeEntryModified.id.hasBackend());
          assert((updatedTimeEntryModified == null) || updatedTimeEntryModified.id.hasLocal());
          assert((updatedTimeEntryModified == null) || updatedTimeEntryModified.id.hasBackend());
// TODO:remove
/*
          Backend.removeAllLocalDatabase("timeEntriesModified",
                                         Backend.condition("id=? AND modifiedState=?",
                                                           id.local,
                                                           Backend.ModifiedStates.UPDATED
                                                          )
                                        );
*/
          if      (updatedTimeEntryModified != null)
          {
            // already updated -> update "updated" entry
            Backend.updateLocalDatabase("timeEntriesModified",
                                        Backend.condition("id=?",updatedTimeEntryModified.id.modified),
                                        new Backend.Values
                                        (
                                         "projectId",    projectId,
                                         "issueId",      issueId,
                                         "userId",       userId,
                                         "activityId",   activityId,
                                         "duration",     duration,
                                         "comments",     comments,
                                         "spentOn",      spentOn,
                                         "createdOn",    createdOn,
                                         "updatedOn",    updatedOn
                                        )
                                       );
Dprintf.dprintf("updated modify update-entry");
          }
          else if (addedTimeEntryModified != null)
          {
            // already added -> update "added" entry
            Backend.updateLocalDatabase("timeEntriesModified",
                                        Backend.condition("id=?",addedTimeEntryModified.id.modified),
                                        new Backend.Values
                                        (
                                         "projectId",    projectId,
                                         "issueId",      issueId,
                                         "userId",       userId,
                                         "activityId",   activityId,
                                         "duration",     duration,
                                         "comments",     comments,
                                         "spentOn",      spentOn,
                                         "createdOn",    createdOn,
                                         "updatedOn",    updatedOn
                                        )
                                       );
Dprintf.dprintf("updated modify add-entry");
          }
          else
          {
            // not addded/updated -> insert new "updated" entry
            id.modified = Backend.insertLocalDatabase("timeEntriesModified",
                                                      new Backend.Values
                                                      (
                                                       "timeEntryId",  id.local,
                                                       "modifiedState",modifiedState.ordinal(),

                                                       "projectId",    projectId,
                                                       "issueId",      issueId,
                                                       "userId",       userId,
                                                       "activityId",   activityId,
                                                       "duration",     duration,
                                                       "comments",     comments,
                                                       "spentOn",      spentOn,
                                                       "createdOn",    createdOn,
                                                       "updatedOn",    updatedOn
                                                      )
                                                     );
Dprintf.dprintf("inserted modify update-entry+update entry");
          }
          break;
        case DELETED:
          assert((addedTimeEntryModified == null) || !addedTimeEntryModified.id.hasLocal());
          assert((addedTimeEntryModified == null) || !addedTimeEntryModified.id.hasBackend());
          assert((updatedTimeEntryModified == null) || updatedTimeEntryModified.id.hasLocal());
          assert((updatedTimeEntryModified == null) || updatedTimeEntryModified.id.hasBackend());
/*
          Backend.removeAllLocalDatabase("timeEntriesModified",
                                         Backend.condition("modifiedId=?",
                                                           modifiedId
                                                          )
                                        );*/
          if      (updatedTimeEntryModified != null)
          {
            // already updated -> change to "deleted" entry
            Backend.updateLocalDatabase("timeEntriesModified",
                                        Backend.condition("id=?",updatedTimeEntryModified.id.modified),
                                        new Backend.Values
                                        (
                                         "modifiedState",modifiedState.ordinal()
                                        )
                                       );
Dprintf.dprintf("changed modify update-entry -> modify delete-entry");
          }
          else if (addedTimeEntryModified != null)
          {
            // already added -> delete entry
            Backend.removeLocalDatabase("timeEntriesModified",
                                        addedTimeEntryModified.id.modified
                                       );
Dprintf.dprintf("removed modify delete-entry");
          }
          else
          {
            // not added/updated -> insert "deleted" entry
            id.modified = Backend.insertLocalDatabase("timeEntriesModified",
                                                      new Backend.Values
                                                      (
                                                       "timeEntryId",  id.local,
                                                       "modifiedState",modifiedState.ordinal(),

                                                       "projectId",    projectId,
                                                       "issueId",      issueId,
                                                       "userId",       userId,
                                                       "activityId",   activityId,
                                                       "duration",     duration,
                                                       "comments",     comments,
                                                       "spentOn",      spentOn,
                                                       "createdOn",    createdOn,
                                                       "updatedOn",    updatedOn
                                                      )
                                                     );
Dprintf.dprintf("inserted modify delete-entry");
          }
          break;
      }
      super.setModified(modifiedState);
    }
/*
    catch (InterruptedException exception)
    {
      // ignored
    }*/
    finally
    {
      Backend.modifyUnlock();
    }
  }

  /** check iff activity in time entries is supported for project
   * @return true iff activity in time entries is supported
   */
  public boolean isActivitySupported()
  {
    return Backend.backendDrivers.get(backendType).isActivitySupported();
  }

  /** get sorted tracker list
   * @return tracker list
   */
  public TrackerList getTrackerList()
    throws BackendException
  {
    return Backend.getTrackerList(new Backend.TypeSet(backendType),projectId);
  }

  /** get sorted status transition list
   * @return transition list
   */
  public StatusTransitionList getStatusTransitionList()
    throws BackendException
  {
    return Backend.getStatusTransitionList(new Backend.TypeSet(backendType),id);
  }

  /** get sorted activity list
   * @return activity list
   */
  public ActivityList getActivityList()
    throws BackendException
  {
    return Backend.getActivityList(new Backend.TypeSet(backendType),projectId);
  }

  /** get sorted priority list
   * @return priority list
   */
  public PriorityList getPriorityList()
    throws BackendException
  {
    return Backend.getPriorityList(new Backend.TypeSet(backendType),projectId);
  }

  /** get sorted member list
   * @param roleIds role filter ids or NULL
   * @return member list
   */
  public MemberList getMemberList(List<Id> roleIds)
    throws BackendException
  {
    return Backend.getMemberList(new Backend.TypeSet(backendType),projectId,roleIds);
  }

  /** get sorted version list
   * @return version list
   */
  public VersionList getVersionList()
    throws BackendException
  {
    return Backend.getVersionList(new Backend.TypeSet(backendType),projectId);
  }

  /** get sorted issue list
   * @param activeOnly true for active projects, False for all projets
   * @return issue list
   */
  public IssueList getIssueList(boolean activeOnly)
    throws BackendException
  {
    return Backend.getIssueList(new Backend.TypeSet(backendType),projectId,activeOnly);
  }

  /** get sorted issue list
   * @return issue list
   */
  public IssueList getIssueList()
    throws BackendException
  {
    return getIssueList(false);
  }

  /** get activity
   * @param defaultActivity default activity
   * @return activity or defaultActivity
   */
  public Activity getActivity()
  {
    Activity activity = null;

    try
    {
      activity = Backend.getActivity(new Backend.TypeSet(backendType),activityId);
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return activity;
  }

  /** get project
   * @return project or null
   */
  public Project getProject()
  {
    Project project = null;

    try
    {
      project = Backend.getProject(new Backend.TypeSet(backendType),projectId);
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return project;
  }

  /** get issue
   * @return issue or null
   */
  public Issue getIssue()
  {
    Issue issue = null;

// TODO remove try+catch?
    try
    {
      if (!issueId.isNone())
      {
        issue = Backend.getIssue(new Backend.TypeSet(backendType),issueId);
      }
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return issue;
  }

  /** get spent hours fraction
   * @return hours fraction
   */
  public int getHourFraction()
  {
    return (int)(duration/60);
  }

  /** get spent minutes fraction
   * @return minutes fraction
   */
  public int getMinuteFraction()
  {
    return (int)(duration%60);
  }

  /** get database parameters
   * @returns hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("id",id);
//      databaseParameters.put("name",name);

    return databaseParameters;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "TimeEntry { backend="+backendType+", id="+id+", projectId="+projectId+", issueId="+issueId+", activityId="+activityId+", duration="+duration+", spentOn="+spentOn+", '"+comments+"' }";
  }
}

/** time entry list
 */
class TimeEntryList extends ArrayList<TimeEntry>
{
  final static TimeEntryList EMPTY = new TimeEntryList();
}

/** time entry comparator
 */
class TimeEntryComparator implements Comparator<TimeEntry>
{
  // --------------------------- constants --------------------------------
  enum SortModes
  {
    SPENT_ON,
    ID,
    HOURS,
    ACTIVITY,
    PROJECT,
    ISSUE,
    COMMENTS
  };

  // --------------------------- variables --------------------------------
  private SortModes sortMode;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create time entry comparator
   */
  TimeEntryComparator()
  {
    this.sortMode = SortModes.SPENT_ON;
  }

  /** get sort mode
   * @return sort mode
   */
  public SortModes getSortMode()
  {
    return sortMode;
  }

  /** set sort mode
   * @param sortMode sort mode
   */
  public void setSortMode(SortModes sortMode)
  {
    this.sortMode = sortMode;
  }

  /** compare time entries
   * @param timeEntry0, timeEntry1 file data to compare
   * @return -1 iff timeEntry0 < timeEntry1,
              0 iff timeEntry0 = timeEntry1,
              1 iff timeEntry0 > timeEntry1
   */
  public int compare(TimeEntry timeEntry0, TimeEntry timeEntry1)
  {
    if      (timeEntry0 == null)
    {
      return  1;
    }
    else if (timeEntry1 == null)
    {
      return -1;
    }
    else
    {
      switch (sortMode)
      {
        case SPENT_ON:
          return timeEntry0.spentOn.compareTo(timeEntry1.spentOn);
        case HOURS:
          if      (timeEntry0.duration < timeEntry1.duration) return -1;
          else if (timeEntry0.duration > timeEntry1.duration) return  1;
          else                                                return  0;
        case ACTIVITY:
          Activity activity0 = timeEntry0.getActivity();
          Activity activity1 = timeEntry1.getActivity();
          if ((activity0 != null) && (activity1 != null))
            return activity0.name.compareToIgnoreCase(activity1.name);
          else if (activity0 != null)
            return -1;
          else if (activity1 != null)
            return 1;
          else
            return 0;
        case PROJECT:
          Project project0 = timeEntry0.getProject();
          Project project1 = timeEntry1.getProject();
          if ((project0 != null) && (project1 != null))
            return project0.name.compareToIgnoreCase(project1.name);
          else if (project0 != null)
            return -1;
          else if (project1 != null)
            return 1;
          else
            return 0;
        case ISSUE:
          Issue issue0 = timeEntry0.getIssue();
          Issue issue1 = timeEntry1.getIssue();
          if ((issue0 != null) && (issue1 != null))
            return issue0.summary.compareToIgnoreCase(issue1.summary);
          else if (issue0 != null)
            return -1;
          else if (issue1 != null)
            return 1;
          else
            return 0;
        case COMMENTS:
          return timeEntry0.comments.compareToIgnoreCase(timeEntry1.comments);
        default:
          return 0;
      }
    }
  }
}

/** time entry database handler
 */
class TimeEntryDatabaseHandler extends DataBaseHandler<TimeEntry>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM timeEntries"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public TimeEntry get(ValueSet valueSet)
    throws SQLException
  {
    return new TimeEntry(valueSet.getEnum   ("backendType",Backend.Types.class),
                         valueSet.getId     ("id","backendId","modifiedId"),
                         valueSet.getId     ("projectId"),
                         valueSet.getId     ("issueId"),
                         valueSet.getId     ("userId"),
                         valueSet.getId     ("activityId"),
                         valueSet.getLong   ("duration"),
                         valueSet.getString ("comments"),
                         valueSet.getSpentOn("spentOn"),
                         valueSet.getDate   ("createdOn"),
                         valueSet.getDate   ("updatedOn")
                        );
  }
}

/** modified time entry info
 */
class TimeEntryModified extends TimeEntry
{
  public Backend.ModifiedStates modifiedState;

  /** create modifed time entry
   * @param backendType backend type
   * @param id database id
   * @param projectId project id
   * @param issueId issue id
   * @param userId user id
   * @param activityId activity id
   * @param duration spent duration [min]
   * @param comments comments
   * @param spentOn spent-on date
   * @param createdOn created-on date
   * @param updatedOn update-on date
   * @param modifiedState modified state
   */
  TimeEntryModified(Backend.Types          backendType,
                    Id                     id,
                    Id                     projectId,
                    Id                     issueId,
                    Id                     userId,
                    Id                     activityId,
                    long                   duration,
                    String                 comments,
                    SpentOn                spentOn,
                    Date                   createdOn,
                    Date                   updatedOn,
                    Backend.ModifiedStates modifiedState
                   )
  {
    super(backendType,
          id,
          projectId,
          issueId,
          userId,
          activityId,
          duration,
          comments,
          spentOn,
          createdOn,
          updatedOn
         );
    this.modifiedState = modifiedState;
  }

  /** compare with other modified time entry
   * @return -1/0+1 iff lower/equals/greater
   */
  public int compareTo(TimeEntryModified otherTimeEntryModified)
  {
    int result = 0;
    if (this != otherTimeEntryModified)
    {
      if      (backendType != otherTimeEntryModified.backendType)
      {
        result = -1;
      }
      else if (modifiedState != otherTimeEntryModified.modifiedState)
      {
        result = -1;
      }
      else
      {
        result = id.compareTo(otherTimeEntryModified.id);
      }
    }

    return result;
  }

  /** done modification
   */
  public void doneModified()
    throws BackendException
  {
    Backend.removeLocalDatabase("timeEntriesModified",
                                id.modified
                               );
  }

  /** update local database
   */
  protected void updateLocalDatabase()
    throws BackendException
  {
    assert(id.hasLocal());

    Backend.updateLocalDatabase("timeEntriesLocal",
                                Backend.condition("backendType=? AND id=?",backendType,id.local),
                                new Backend.Values
                                (
                                  "backendId", id,
                                  "projectId", projectId,
                                  "issueId",   issueId,
                                  "userId",    userId,
                                  "activityId",activityId,
                                  "duration",  duration,
                                  "comments",  comments,
                                  "spentOn",   spentOn,
                                  "createdOn", createdOn,
                                  "updatedOn", updatedOn
                                )
                               );
  }

  /** remove from local database
   */
  protected void removeLocalDatabase()
    throws BackendException, InterruptedException
  {
    // Note: set as "deleted" only; time entry is removed from database on synchronisation
    setModified(Backend.ModifiedStates.DELETED);
  }

  /** revert in local database
   */
  protected void revertLocalDatabase()
    throws BackendException, InterruptedException
  {
    clearModified();
  }

  /** store time entry on backend server
   */
  public void store()
    throws BackendException
  {
    assert(id.hasLocal());

    Backend.backendDrivers.get(backendType).storeTimeEntry(this);
    updateLocalDatabase();
  }

  /** update time entry on backend server
   * @param originTimeEntry origin time entry
   */
  public void update(TimeEntry originTimeEntry)
    throws BackendException
  {
    assert id.hasLocal();
    assert id.hasBackend();

    // update on server
    Backend.backendDrivers.get(backendType).updateTimeEntry(this,originTimeEntry);

    // update local database
    Backend.updateLocalDatabase("timeEntriesLocal",
                                backendType,
                                id,
                                new Backend.Values
                                (
                                  "backendId", id,
                                  "projectId", projectId,
                                  "issueId",   issueId,
                                  "userId",    userId,
                                  "activityId",activityId,
                                  "duration",  duration,
                                  "comments",  comments,
                                  "spentOn",   spentOn,
                                  "createdOn", createdOn,
                                  "updatedOn", updatedOn
                                )
                               );
  }

  /** delete time entry on backend server
   * @param timeEntry time entry to delete
   */
  public void delete()
    throws BackendException
  {
    assert(id.hasLocal());

Dprintf.dprintf("xxxxxxxxxxxxxxxxxxx delete timeEntry: %s",this);
Dprintf.printStackTrace();
// TODO: remove?
//    assert id.hasBackend();

    // delete on server
    if (id.hasBackend())
    {
      Backend.backendDrivers.get(backendType).deleteTimeEntry(this);
Dprintf.dprintf("backendType=%s id=%s",backendType,id);
    }

    // remove from local database
    Backend.removeLocalDatabase("timeEntriesLocal",
                                Backend.condition("backendType=? AND id=?",backendType,id.local)
                               );
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "TimeEntryModified { backend="+backendType+", id="+id+", modifiedState="+modifiedState+", projectId="+projectId+", issueId="+issueId+", activityId="+activityId+", duration="+duration+", spentOn="+spentOn+", '"+comments+"' }";
  }

  /** clear all local time entry modifications
   * @param timeEntry time entry
   */
  private void clearModified()
    throws BackendException, InterruptedException
  {
    assert(id.hasLocal());

    {
      Backend.modifyLock();
    }
    try
    {
      Backend.removeAllLocalDatabase("timeEntriesModified",
                                     Backend.condition("timeEntryId=?",
                                                       id.local
                                                      )
                                    );
    }
    finally
    {
      Backend.modifyUnlock();
    }
  }
};

/** time entry database handler
 */
class TimeEntryModifiedDatabaseHandler extends DataBaseHandler<TimeEntryModified>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT "+
           "   timeEntriesLocal.backendType,"+
           "   timeEntriesLocal.id,"+
           "   timeEntriesLocal.backendId,"+
           "   timeEntriesModified.id AS modifiedId,"+
           "   timeEntriesModified.modifiedState,"+
           "   timeEntriesModified.projectId,"+
           "   timeEntriesModified.issueId,"+
           "   timeEntriesModified.userId,"+
           "   timeEntriesModified.activityId,"+
           "   timeEntriesModified.duration,"+
           "   timeEntriesModified.comments,"+
           "   timeEntriesModified.spentOn,"+
           "   timeEntriesModified.createdOn,"+
           "   timeEntriesModified.updatedOn,"+
           "   timeEntriesLocal.lastUpdateTimestamp"+
           " FROM timeEntriesModified"+
           "   LEFT JOIN timeEntriesLocal ON timeEntriesLocal.id=timeEntriesModified.timeEntryId"+
           " WHERE timeEntriesLocal.id NOT NULL"+
           ((additionalFilter != null) ? " AND "+additionalFilter : "");
  }

  @Override
  public TimeEntryModified get(ValueSet valueSet)
    throws SQLException
  {
    return new TimeEntryModified(valueSet.getEnum   ("backendType",Backend.Types.class),
                                 valueSet.getId     ("id","backendId","modifiedId"),
                                 valueSet.getId     ("projectId"),
                                 valueSet.getId     ("issueId"),
                                 valueSet.getId     ("userId"),
                                 valueSet.getId     ("activityId"),
                                 valueSet.getLong   ("duration"),
                                 valueSet.getString ("comments"),
                                 valueSet.getSpentOn("spentOn"),
                                 valueSet.getDate   ("createdOn"),
                                 valueSet.getDate   ("updatedOn"),
                                 valueSet.getEnum   ("modifiedState",Backend.ModifiedStates.class)
                                );
  }
}

/* end of file */
