/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** version database id
 */
class VersionId extends Id
{
  /** create id
   * @param backend backend id
   */
  VersionId(int backend)
  {
    super(backend);
  }
}

/** version
 */
class Version extends Entity
{
  public final static Version NONE = new Version(new Id(Id.NONE));
  public final static Version ANY  = new Version(new Id(Id.ANY));

  public Id         projectId;
  public String     name;
  public String     description;
  public String     status;
  public Date       dueDate;
  public final Date createdOn;
  public Date       updatedOn;

  /** create version
   * @param backendType backend type
   * @param id database id
   * @param projectId project id
   * @param name name of version
   * @param description description
   * @param status status
   * @param dueDate due date
   * @param createdOn created-on date
   * @param updatedOn update-on date
   */
  Version(Backend.Types backendType,
          Id            id,
          Id            projectId,
          String        name,
          String        description,
          String        status,
          Date          dueDate,
          Date          createdOn,
          Date          updatedOn
         )
  {
    super(backendType,id);
    assert backendType != Backend.Types.NONE;
    assert !id.isNone();
    assert !id.isAny();
    assert(name != null);
    assert(description != null);

    this.projectId   = projectId;
    this.name        = name;
    this.description = description;
    this.status      = status;
    this.dueDate     = dueDate;
    this.createdOn   = createdOn;
    this.updatedOn   = updatedOn;
  }

  /** create version
   * @param backendType backend type
   * @param id database id
   * @param projectId project id
   * @param name name of version
   * @param description description
   * @param status status
   */
  Version(Backend.Types backendType,
          Id            id,
          Id            projectId,
          String        name,
          String        description,
          String        status
         )
  {
    this(backendType,
         id,
         projectId,
         name,
         description,
         status,
         new Date(),
         new Date(),
         new Date()
        );
  }

  /** create version
   * @param id database id
   */
  private Version(Id id)
  {
    super(Backend.Types.NONE,id);

    this.projectId   = new Id();
    this.name        = "";
    this.description = "";
    this.status      = "";
    this.dueDate     = new Date();
    this.createdOn   = new Date();
    this.updatedOn   = new Date();
  }

  /** get database parameters
   * @returns hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("id",id);
    databaseParameters.put("name",name);

    return databaseParameters;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "Version { id="+id+", name="+name+" }";
  }
}

/** version list
 */
class VersionList extends ArrayList<Version>
{
  final static VersionList EMPTY = new VersionList();
}

/** version database handler
 */
class VersionDatabaseHandler extends DataBaseHandler<Version>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM versions"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public Version get(ValueSet valueSet)
    throws SQLException
  {
    return new Version(valueSet.getEnum  ("backendType",Backend.Types.class),
                       valueSet.getId    ("id","backendId"),
                       valueSet.getId    ("projectId"),
                       valueSet.getString("name"),
                       valueSet.getString("description"),
                       valueSet.getString("status")
                      );
  }
}

/* end of file */
