/***********************************************************************\
*
* $Revision: 239 $
* $Date: 2015-09-04 17:26:34 +0200 (Fri, 04 Sep 2015) $
* $Author: torsten $
* Contents: Table combo
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.util.ArrayList;
import java.util.Comparator;
import java.util.EnumSet;

// graphics
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.TypedListener;

/****************************** Classes ********************************/

/** table combo widget
*/
class TableCombo extends Composite
{
  /** table combo item
   */
  class Item<T>
  {
    TableItem tableItem;
    T         data;
    Image     image;
    boolean   checked;
    Object    values[];

    /** create item
     * @param data data
     * @param parentItem parent item
     * @param column values
     * @param image image or null
     * @param values values
     */
    public Item(T data, Item<T> parentItem, Image image, Object values[])
    {
      this.tableItem = null;
      this.data      = data;
      this.image     = image;
      this.checked   = false;
      this.values    = values;
    }

    /** create item
     * @param data data
     * @param image image or null
     * @param values values
     */
    public Item(T data, Image image, Object values[])
    {
      this(data,(Item<T>)null,image,values);
    }

    /** get data
     * @return data
     */
    T getData()
    {
      return data;
    }

    /** get checked
     * @return true iff checked
     */
    boolean getChecked()
    {
      return checked;
    }

    /** set checked
     * @param checked true iff checked
     */
    void setChecked(boolean checked)
    {
      this.checked = checked;
    }

    /** check if item is visible
     * @param text text
     * @return true if text match to item data or values
     */
    boolean isVisible(String filters[])
    {
      if (filters.length > 0)
      {
        boolean visibleFlag;

        if (filterTypeSet.contains(FilterTypes.DATA))
        {
          // check if data matches
          visibleFlag = true;
          String text = (data != null) ? getFilterText(data.toString()) : "";
          for (String filter : filters)
          {
            if (text.indexOf(filter) < 0)
            {
              visibleFlag = false;
              break;
            }
          }
          if (visibleFlag) return true;
        }

        if (filterTypeSet.contains(FilterTypes.TEXT))
        {
          // check if text matches
          visibleFlag = true;
          for (String filter : filters)
          {
            visibleFlag = false;
            for (int i = 0; i < values.length; i++)
            {
              if (values[i] != null)
              {
                if      (values[i] instanceof String)
                {
                  if (getFilterText((String)values[i]).indexOf(filter) >= 0)
                  {
                    visibleFlag = true;
                    break;
                  }
                }
                else if (values[i] instanceof Image)
                {
                  // nothing to do
                }
                else
                {
                  if (getFilterText(values[i].toString()).indexOf(filter) >= 0)
                  {
                    visibleFlag = true;
                    break;
                  }
                }
              }
            }
            if (!visibleFlag) break;
          }
          if (visibleFlag) return true;
        }

        return false;
      }
      else
      {
        return true;
      }
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      StringBuilder buffer = new StringBuilder();
      for (int i = 0; i < values.length; i++)
      {
        if (values[i] != null)
        {
          buffer.append(' ');
          if      (values[i] instanceof String)
          {
            buffer.append((String)values[i]);
          }
          else if (values[i] instanceof Image)
          {
          }
          else
          {
            buffer.append(values[i].toString());
          }
        }
      }

      return "Item { data="+data.toString()+" }";
    }
  };

  /** table combo item list
   */
  class ItemList extends ArrayList<Item>
  {
  };

  /** table combo item filter types
   */
  static enum FilterTypes
  {
    DATA,
    TEXT
  };

  private static Color     COLOR_BACKGROUND_ODD;
  private static Color     COLOR_BACKGROUND_EVEN;

  private final static int DEFAULT_HEIGHT = 200;

  private final int                  additionalWidth;
  private final int                  height;
  private final EnumSet<FilterTypes> filterTypeSet;
  private ItemList                   itemList         = new ItemList();
  private ItemList                   filteredItemList = new ItemList();
  private Button                     widgetText;
  private Shell                      dialog;
  private Table                      widgetTable;
  private Text                       widgetFilter;
  private Object                     data = null;

  /** create table combo
   * @param composite parent composite
   * @param style SWT style
   * @param rightComposite additional (optional) widget on the right
   * @param tableStyle table SWT style
   * @param additionalWidth additional width
   * @param height height
   * @param filterTypeSet data view filter set
   */
  public TableCombo(Composite composite, int style, Composite rightComposite, int tableStyle, final int additionalWidth, final int height, EnumSet<FilterTypes> filterTypeSet)
  {
    super(composite,SWT.NONE);

    Composite subComposite,subSubComposite;
    Label     label;

    this.additionalWidth = additionalWidth;
    this.height          = height;
    this.filterTypeSet   = filterTypeSet;

    final Display display = composite.getDisplay();

    COLOR_BACKGROUND_ODD  = new Color(display,0xF8,0xF7,0xF6);
    COLOR_BACKGROUND_EVEN = display.getSystemColor(SWT.COLOR_WHITE);

    // layout
    FillLayout fillLayout = new FillLayout();
    fillLayout.type = SWT.HORIZONTAL;
    this.setLayout(fillLayout);

    // combo text
    widgetText = new Button(this,style|SWT.PUSH);

    // combo dialog
    dialog = new Shell(getShell(),SWT.TOOL);
    dialog.setLayout(new TableLayout(1.0,new double[]{1.0,0.0}));
    dialog.setVisible(false);

    subComposite = Widgets.newComposite(dialog);
    subComposite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));
    Widgets.layout(subComposite,0,0,TableLayoutData.NSWE,0,0,2,2);
    {
      widgetTable = Widgets.newTable(subComposite,tableStyle);
      widgetTable.setHeaderVisible(false);
      widgetTable.setLayout(new TableLayout(1.0,new double[]{0.0,1.0}));
      Widgets.layout(widgetTable,0,0,TableLayoutData.NSWE);

      subSubComposite = Widgets.newComposite(subComposite);
      subSubComposite.setLayout(new TableLayout(1.0,new double[]{0.0,1.0}));
      Widgets.layout(subSubComposite,1,0,TableLayoutData.NSWE,0,0,2,2);
      {
        label = Widgets.newLabel(subSubComposite,"Filter:");
        Widgets.layout(label,0,0,TableLayoutData.W);

        widgetFilter = Widgets.newText(subSubComposite,SWT.SEARCH|SWT.ICON_CANCEL);
        Widgets.layout(widgetFilter,0,1,TableLayoutData.NSWE);
      }
    }
    if (rightComposite != null)
    {
      rightComposite.setParent(dialog);
      Widgets.layout(rightComposite,0,1,TableLayoutData.NS,0,0,2,2);
    }

    // add listeners
    getShell().addListener(SWT.Move, new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        if (!dialog.isDisposed())
        {
          if (dialog.isVisible())
          {
            Rectangle bounds = widgetText.getBounds();

            Point p = toDisplay(bounds.x,bounds.y);
            dialog.setLocation(p.x,p.y+bounds.height);
          }
        }
      }
    });
    getShell().addListener(SWT.Resize, new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        if (!dialog.isDisposed())
        {
          if (dialog.isVisible())
          {
            Rectangle bounds = widgetText.getBounds();

            dialog.setSize(bounds.width+additionalWidth,height);

            Point p = toDisplay(bounds.x,bounds.y);
            dialog.setLocation(p.x,p.y+bounds.height);
          }
        }
      }
    });

    dialog.addListener(SWT.Close,new Listener()
    {
      public void handleEvent(Event event)
      {
        close();
        event.doit = false;
      }
    });

    widgetText.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        if (!dialog.isDisposed())
        {
          if (dialog.isVisible())
          {
            close();
          }
          else
          {
            open();
          }
        }
      }
    });

    widgetTable.addListener(SWT.EraseItem, new Listener()
    {
      public void handleEvent(Event event)
      {
        TableItem tableItem  = (TableItem)event.item;
        int       index      = widgetTable.indexOf(tableItem);

        Rectangle clientArea = widgetTable.getBounds();
        Rectangle bounds     = event.getBounds();
        GC        gc         = event.gc;

/*
        // enable alpha blending if possible
        gc.setAdvanced(true);
        if (gc.getAdvanced()) gc.setAlpha(127);
*/
        // draw background
        Color background = gc.getBackground();
        gc.setBackground(((index % 2)==0) ? COLOR_BACKGROUND_EVEN : COLOR_BACKGROUND_ODD);
        gc.fillRectangle(bounds.x, bounds.y, clientArea.width, bounds.height);
        gc.setBackground(background);

        event.detail &= ~SWT.BACKGROUND;
      }
    });
    widgetTable.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Table widget = (Table)selectionEvent.widget;

        TableItem selectedTableItems[] = widget.getSelection();
        if (selectedTableItems.length > 0)
        {
          // get selected item
          Item item = (Item)selectedTableItems[0].getData();

          // set checked
          item.setChecked(selectedTableItems[0].getChecked());

          // get data
          data = item.data;

          // trigger selection listeners
          Event event = new Event();
          event.item  = selectedTableItems[0];
          notifyListeners(SWT.Selection,event);
        }

        if ((widget.getStyle() & SWT.CHECK) == 0) close();
      }
    });

    widgetFilter.addKeyListener(new KeyListener()
    {
      @Override
      public void keyPressed(KeyEvent keyEvent)
      {
        if      (Widgets.isAccelerator(keyEvent,SWT.ARROW_UP))
        {
          TableItem selectedTableItem = getSelectedTableItem();

          // find next selectable item
          TableItem newSelectedTableItem = null;
          if (selectedTableItem != null)
          {
            // check if previous item
            TableItem tableItems[] = widgetTable.getItems();
            for (int i = 1; i < tableItems.length; i++)
            {
              if (tableItems[i] == selectedTableItem)
              {
                newSelectedTableItem = tableItems[i-1];
                break;
              }
            }
          }

          // select new item
          if (newSelectedTableItem != null)
          {
            widgetTable.setSelection(newSelectedTableItem);

            // trigger modify listeners
            Event event = new Event();
            event.widget = widgetText;
            notifyListeners(SWT.Modify,event);
          }
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.ARROW_DOWN))
        {
          TableItem selectedTableItem = getSelectedTableItem();

          // find previous selectable item
          TableItem newSelectedTableItem = null;
          if (selectedTableItem != null)
          {
            // check if next item
            TableItem tableItems[] = widgetTable.getItems();
            for (int i = 0; i < tableItems.length-1; i++)
            {
              if (tableItems[i] == selectedTableItem)
              {
                newSelectedTableItem = tableItems[i+1];
                break;
              }
            }
          }

          // select new item
          if (newSelectedTableItem != null)
          {
            widgetTable.setSelection(newSelectedTableItem);

            // trigger modify listeners
            Event event = new Event();
            event.widget = widgetText;
            notifyListeners(SWT.Modify,event);
          }
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'x'))
        {
          widgetFilter.setText("");
        }
      }
      @Override
      public void keyReleased(KeyEvent keyEvent)
      {
      }
    });
    widgetFilter.addModifyListener(new ModifyListener()
    {
      @Override
      public void modifyText(ModifyEvent modifyEvent)
      {
        filterItems();
      }
    });
    widgetFilter.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        TableItem selectedTableItem = getSelectedTableItem();
        if (selectedTableItem != null)
        {
          Event event;

          // get selected item data
          data = (selectedTableItem.getData() != null) ? ((Item)selectedTableItem.getData()).data : null;

          // trigger selection listeners
          event = new Event();
          event.widget = widgetText;
          event.item = selectedTableItem;
          notifyListeners(SWT.Selection,event);

          // trigger modify listeners
          event = new Event();
          event.widget = widgetText;
          notifyListeners(SWT.Modify,event);
        }

        close();
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
      }
    });

    final Listener closeListener = new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        if (event.widget instanceof Control)
        {
          Point p = ((Control)event.widget).toDisplay(event.x,event.y);
          Point q = widgetText.toControl(p);
//Dprintf.dprintf("w=%s p=%s q=%s",event.widget,p,q);
//Dprintf.dprintf("widgetText=%s",widgetText.getBounds());
//Dprintf.dprintf("dialog=%s",dialog.getBounds());
          if (   (!widgetText.isVisible() || !widgetText.getBounds().contains(q))
              && (!dialog.isVisible()     || !dialog.getBounds().contains(p)    )
             )
          {
            close();
          }
        }
      }
    };
    getDisplay().addFilter(SWT.MouseDown,closeListener);
    addDisposeListener(new DisposeListener()
    {
      public void widgetDisposed(DisposeEvent disposeEvent)
      {
        getDisplay().removeFilter(SWT.MouseDown,closeListener);
      }
    });
  }

  /** create table combo
   * @param composite parent composite
   * @param style SWT style
   * @param rightComposite additional (optional) widget on the right
   * @param tableStyle table SWT style
   * @param additionalWidth additional width
   * @param height height
   */
  public TableCombo(Composite composite, int style, Composite rightComposite, int tableStyle, int additionalWidth, int height)  {
    this(composite,style,rightComposite,tableStyle,additionalWidth,height,EnumSet.allOf(FilterTypes.class));
  }

  /** create table combo
   * @param composite parent composite
   * @param style SWT style
   * @param rightComposite additional (optional) widget on the right
   * @param tableStyle table SWT style
   * @param additionalWidth additional width
   */
  public TableCombo(Composite composite, int style, Composite rightComposite, int tableStyle, int additionalWidth)
  {
    this(composite,style,rightComposite,tableStyle,additionalWidth,DEFAULT_HEIGHT);
  }

  /** create table combo
   * @param composite parent composite
   * @param style SWT style
   * @param tableStyle table SWT style
   * @param additionalWidth additional width
   * @param filterTypeSet data view filter set
   */
  public TableCombo(Composite composite, int style, int tableStyle, int additionalWidth, EnumSet<FilterTypes> filterTypeSet)
  {
    this(composite,style,(Composite)null,tableStyle,additionalWidth,DEFAULT_HEIGHT,filterTypeSet);
  }

  /** create table combo
   * @param composite parent composite
   * @param style SWT style
   * @param tableStyle table SWT style
   * @param additionalWidth additional width
   */
  public TableCombo(Composite composite, int style, int tableStyle, int additionalWidth)
  {
    this(composite,style,tableStyle,additionalWidth,EnumSet.allOf(FilterTypes.class));
  }

  /** create table combo
   * @param composite parent composite
   * @param style SWT style
   * @param rightComposite additional (optional) widget on the right
   * @param tableStyle table SWT style
   */
  public TableCombo(Composite composite, int style, Composite rightComposite, int tableStyle)
  {
    this(composite,style,rightComposite,tableStyle,0);
  }

  /** create table combo
   * @param composite parent composite
   * @param style SWT style
   * @param tableStyle table SWT style
   */
  public TableCombo(Composite composite, int style, int tableStyle)
  {
    this(composite,style,(Composite)null,tableStyle);
  }

  /** create table combo
   * @param composite parent composite
   */
  public TableCombo(Composite composite)
  {
    this(composite,SWT.NONE,SWT.NONE);
  }

  /** add column to table widget
   * @param columnNb column number
   * @param title column title
   * @param style style
   * @param width width of column
   * @param resizable TRUE iff resizable column
   * @return new table column
   */
  public void addTableColumn(int columnNb, String title, int style, int width, boolean resizable)
  {
    if (title != null) widgetTable.setHeaderVisible(true);
    Widgets.addTableColumn(widgetTable,columnNb,title,style,width,resizable);
  }

  /** add column to table widget
   * @param columnNb column number
   * @param style style
   * @param width width of column
   * @param resizable TRUE iff resizable column
   * @return new table column
   */
  public void addTableColumn(int columnNb, int style, int width, boolean resizable)
  {
    addTableColumn(columnNb,(String)null,style,width,resizable);
  }

  /** set button image
   * @param image image
   */
  public void setImage(Image image)
  {
    widgetText.setImage(image);
  }

  /** set table combo layout
   * @param columnWidth column width
   */
  public void setLayout(double columnWeight)
  {
    widgetTable.setLayout(new TableLayout(1.0,columnWeight));
  }

  /** set table combo layout
   * @param columnWidths column widths
   */
  public void setLayout(double columnWeights[])
  {
    widgetTable.setLayout(new TableLayout(1.0,columnWeights));
  }

  /** set enabled
   */
  public void setEnabled(boolean enabled)
  {
    widgetText.setEnabled(enabled);
    if (!enabled && !dialog.isDisposed() && dialog.isVisible())
    {
      close();
    }
  }

  /** add table item
   * @param data data
   * @param image image or null
   * @param values table column values
   * @return table item
   */
  public <T> Item addTableItem(T data, Image image, Object... values)
  {
    Item<T> item = new Item<T>(data,image,values);
    itemList.add(item);

    return item;
  }

  /** add table item
   * @param data data
   * @param values table column values
   * @return table item
   */
  public <T> Item addTableItem(T data, Object... values)
  {
    return addTableItem(data,(Image)null,values);
  }

  /** insert table item
   * @param index index [0..n-1]
   * @param data data
   * @param image image or null
   * @param values table column values
   * @return table item
   */
  public <T> Item insertTableItem(int index, T data, Image image, Object... values)
  {
    Item<T> item = new Item<T>(data,image,values);
    itemList.add(index,item);

    return item;
  }

  /** insert table item
   * @param index index [0..n-1]
   * @param data data
   * @param values table column values
   * @return table item
   */
  public <T> Item insertTableItem(int index, T data, Object... values)
  {
    return insertTableItem(index,data,(Image)null,values);
  }

  /** update or insert table item
   * @param index index [0..n-1]
   * @param data data
   * @param image image or null
   * @param values table column values
   * @return table item
   */
  public <T> Item updateInsertTableItem(Comparator<T> comparator, T data, Image image, Object... values)
  {
    int index = 0;
    while (index < itemList.size())
    {
      Item item = itemList.get(index);
      switch (comparator.compare((T)item.data,data))
      {
        case -1:
          return insertTableItem(index,data,image,values);
        case  0:
          item.image  = image;
          item.values = values;
          return item;
        case  1:
          index++;
          break;
      }
    }

    return addTableItem(data,image,values);
  }

  /** update or insert table item
   * @param index index [0..n-1]
   * @param data data
   * @param values table column values
   * @return table item
   */
  public <T> Item updateInsertTableItem(Comparator<T> comparator, T data, Object... values)
  {
    return updateInsertTableItem(comparator,data,(Image)null,values);
  }

  /** remove table item
   * @param index index [0..n-1]
   */
  public void removeTableItem(int index)
  {
    Item item = itemList.get(index);

    itemList.remove(index);
//    Widgets.removeTableItem(widgetTable,item);
  }

  /** remove table item
   * @param data data
   */
  public <T> void removeTableItem(T data)
  {
    for (int i = 0; i < itemList.size(); i++)
    {
      Item item = itemList.get(i);

      if (item.data.equals(data))
      {
        itemList.remove(i);
//        Widgets.removeTableItem(widgetTable,item);
      }
    }
  }

  /** remove all table items
   */
  public void removeAllTableItems()
  {
    itemList.clear();
    filteredItemList.clear();
    widgetText.setText("");
    filterItems();
  }

  /** find table item
   * @param data data
   * @return table item or null
   */
  public <T> Item findTableItem(T data)
  {
    for (Item item : itemList)
    {
      if (item.data.equals(data))
      {
        return item;
      }
    }

    return null;
  }

  /** get filter text
   * @param text text
   * @return filtered text
   */
  private String getFilterText(String text)
  {
    return text.replaceAll("[^A-Za-z0-9 ]","").toLowerCase().trim();
  }

  /** filter visible items
   * @param itemList item list
   * @param filteredItemList filtered item list
   * @param filters filters
   */
  private void filterItems(ItemList itemList, ItemList filteredItemList, String[] filters)
  {
    filteredItemList.clear();
    for (final Item item : itemList)
    {
      if (item.isVisible(filters))
      {
        filteredItemList.add(item);
      }
    }
  }

  /** filter visible items
   */
  public void filterItems()
  {
    // save selected item
    Item selectedItem = getSelectedItem();

    // clear
    Widgets.removeAllTableItems(widgetTable);

    // filter items
    String filters[] = getFilterText(widgetFilter.getText()).split("\\s+");
    filterItems(itemList,filteredItemList,filters);

    // refresh
    Widgets.removeAllTableItems(widgetTable);
    for (Item item : filteredItemList)
    {
      TableItem tableItem = Widgets.addTableItem(widgetTable,item,item.image,item.values);
      tableItem.setChecked(item.getChecked());
    }

    // find selected table item
    TableItem tableItems[] = widgetTable.getItems();
    TableItem selectedTableItem = null;
    for (final TableItem tableItem : tableItems)
    {
      Item item = (Item)tableItem.getData();
      if (item == selectedItem)
      {
        selectedTableItem = tableItem;
        break;
      }
    }

    // set selected item
    if (selectedTableItem != null)
    {
      // restore selected item
      widgetTable.setSelection(selectedTableItem);
    }
    else
    {
      // select first item
      if (tableItems.length > 0)
      {
        widgetTable.setSelection(tableItems[0]);
      }
    }
  }

  /** add modify listener
   * @param modifyListener modify listener to add
   */
  public void addModifyListener(ModifyListener modifyListener)
  {
    TypedListener typedListener = new TypedListener(modifyListener);
    addListener(SWT.Modify,typedListener);
  }

  /** add selection listener
   * @param selectionListener selection listener to add
   */
  public void addSelectionListener(SelectionListener selectionListener)
  {
    TypedListener typedListener = new TypedListener(selectionListener);
    addListener(SWT.Selection, typedListener);
    addListener(SWT.DefaultSelection, typedListener);
  }

  /** set focus
   */
  public boolean setFocus()
  {
    return widgetText.setFocus();
  }

  /** get selection
   * @return selected value text
   */
  public String getText()
  {
    return widgetText.getText();
  }

  /** set selection text
   * @param text value text
   */
  public void setText(String value)
  {
    widgetText.setText(value.replaceAll("&","&&"));
  }

  /** get item
   * @param index index [0..n-1]
   * @return item
   */
  public Item getItem(int index)
  {
    return itemList.get(index);
  }

  /** get items
   * @return items
   */
  public Item[] getItems()
  {
    return itemList.toArray(new Item[itemList.size()]);
  }

  /** get item count
   * @return item count
   */
  public int getItemCount()
  {
    return itemList.size();
  }

  /** get selectedion
   * @param defaultValue default value
   * @return selected value
   */
  public <T> T getSelection(T defaultValue)
  {
    return (data != null) ? (T)data : defaultValue;
  }

  /** get selectedion
   * @param type return type
   * @return selected value
   */
  public <T> T getSelection(Class<T> type)
  {
    return getSelection((T)null);
  }

  /** get selectedion
   * @return selected value
   */
  public <T> T getSelection()
  {
    return getSelection((T)null);
  }

  /** select table combo item
   * @param data data
   * @param text text format string
   * @param arguments optional arguments for text
   */
  public <T> void setSelection(T data, String text, Object... arguments)
  {
    setText(String.format(text,arguments));
    this.data = data;

    if (data != null)
    {
      widgetTable.deselectAll();
      for (TableItem tableItem : widgetTable.getItems())
      {
        if (tableItem.getData().equals(data))
        {
          widgetTable.setSelection(tableItem);
          break;
        }
      }
      widgetTable.showSelection();
    }
  }

  /** select table combo item
   * @param text text format string
   * @param arguments optional arguments for text
   */
  public <T> void setSelection(String text, Object... arguments)
  {
    setSelection((T)null,text,arguments);
  }

  /** get items
   * @return items
   */
  public <T> ArrayList<T> getChecked(Class<T> clazz)
  {
    ArrayList<T> data = new ArrayList<T>();

    for (Item item : itemList)
    {
      if (item.checked) data.add((T)item.data);
    }

    return data;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "TableCombo";
  }

  // -----------------------------------------------------------------


  /** open table combo
   */
  private void open()
  {
    Rectangle bounds = widgetText.getBounds();

    // set size
    dialog.setSize(bounds.width+additionalWidth,height);

    // show
    Point p = toDisplay(bounds.x,bounds.y);
    dialog.setLocation(p.x,p.y+bounds.height);
    dialog.setVisible(true);

    // set focus
    widgetFilter.setFocus();
    widgetFilter.setText("");

    // refresh
    Widgets.removeAllTableItems(widgetTable);
    for (Item item : filteredItemList)
    {
      TableItem tableItem = Widgets.addTableItem(widgetTable,item,item.image,item.values);
      tableItem.setChecked(item.getChecked());
    }

    // scroll to and show item
    show(filteredItemList,data);
  }

  /** close table combo
   */
  private void close()
  {
    if (!dialog.isDisposed())
    {
      if (dialog.isVisible())
      {
        dialog.setVisible(false);
      }
    }
  }

  /** show item
   * @param tableItem table item
   * @param filteredItemList filterted item list
   * @param data item data
   * @return true iff found
   */
  private <T> boolean show(TableItem tableItem, ItemList filteredItemList, T data)
  {
    int i = 0;
    for (Item item : filteredItemList)
    {
      if (item.data.equals(data))
      {
        widgetTable.setTopIndex(i);
        return true;
      }

      i++;
    }

    return false;
  }

  /** show item
   * @param filteredItemList filterted item list
   * @param data item data
   * @return true iff found
   */
  private <T> boolean show(ItemList filteredItemList, T data)
  {
    int i = 0;
    for (Item item : filteredItemList)
    {
      if (item.data.equals(data))
      {
        widgetTable.setTopIndex(i);
        return true;
      }

      i++;
    }

    return false;
  }

  /** get selected item
   * @return selected item or null
   */
  private Item getSelectedItem()
  {
    TableItem selectedTableItems[] = widgetTable.getSelection();
    return (selectedTableItems.length > 0) ? (Item)selectedTableItems[0].getData() : null;
  }

  /** get last table item
   * @param tableItem table item
   * @return last table item
   */
  private TableItem getSelectedTableItem()
  {
    TableItem selectedTableItems[] = widgetTable.getSelection();
    return (selectedTableItems.length > 0) ? selectedTableItems[0] : null;
  }
}

/* end of file */
